/*
 *  joint.cpp
 *  BioMove3D
 *
 *  Created by Jim Mainprice on 12/05/10.
 *  Copyright 2010 __MyCompanyName__. All rights reserved.
 *
 */

#include "API/Device/robot.hpp"

#include <Eigen/Core>
#define EIGEN_USE_NEW_STDVECTOR
#include <Eigen/StdVector>
 
#include <Eigen/LU>

#include <string>

#include "P3d-pkg.h"

using namespace std;

// import most common Eigen types 
//using namespace Eigen;
using namespace Eigen;

Joint::Joint(Robot *R, p3d_jnt* jntPt, int id, bool copy ) : 
  m_Robot(R),
  m_id(id)
{
	m_Joint = jntPt;
  m_Name = jntPt->name;
}

Vector3d Joint::getVectorPos() const
{
	Vector3d v;
	
	v(0) = m_Joint->abs_pos[0][3];
	v(1) = m_Joint->abs_pos[1][3];
	v(2) = m_Joint->abs_pos[2][3];
	
	return v;
}

Affine3d Joint::getMatrixPos() const
{
    Affine3d t;
	
	t(0,0) = m_Joint->abs_pos[0][0];
	t(1,0) = m_Joint->abs_pos[1][0];
	t(2,0) = m_Joint->abs_pos[2][0];
	t(3,0) = 0.0;
	
	t(0,1) = m_Joint->abs_pos[0][1];
	t(1,1) = m_Joint->abs_pos[1][1];
	t(2,1) = m_Joint->abs_pos[2][1];
	t(3,1) = 0.0;
	
	t(0,2) = m_Joint->abs_pos[0][2];
	t(1,2) = m_Joint->abs_pos[1][2];
	t(2,2) = m_Joint->abs_pos[2][2];
	t(3,2) = 0.0;
	
	t(0,3) = m_Joint->abs_pos[0][3];
	t(1,3) = m_Joint->abs_pos[1][3];
	t(2,3) = m_Joint->abs_pos[2][3];
	t(3,3) = 1.0;
	
	return t;
}

void Joint::getAbsVector(p3d_vector3 r)
{

    r[0] = m_Joint->abs_pos[0][3];
    r[1] = m_Joint->abs_pos[1][3];
    r[2] = m_Joint->abs_pos[2][3];
}

p3d_matrix4* Joint::getAbsPos()
{ 
	return &(m_Joint->abs_pos); 
}

void Joint::shoot(Configuration& q,bool sample_passive)
{
	for(int j=0; j<m_Joint->dof_equiv_nbr; j++) 
	{
		int k = m_Joint->index_dof + j;
	
		if ( 
			sample_passive ||
			((p3d_jnt_get_dof_is_user(m_Joint, j) && p3d_jnt_get_dof_is_active_for_planner(m_Joint,j)) &&
			(q.getRobot()->getRobotStruct()->cntrt_manager->in_cntrt[k] != 2) ) )
		{
			double vmin,vmax;
			p3d_jnt_get_dof_rand_bounds(m_Joint, j, &vmin, &vmax);
			q[k] = p3d_random(vmin, vmax);
			//std::cout << "Sample Passive = "<<sample_passive<<" , Sampling q["<<k<<"] = "<<q[k]<< std::endl;
		} 
		else
		{ 
			q[k] = p3d_jnt_get_dof(m_Joint, j); 
		}
	}
}

double Joint::getJointDof(int ithDoF) const
{
	return p3d_jnt_get_dof(m_Joint,ithDoF);
}

void Joint::setJointDof(int ithDoF, double value)
{
	p3d_jnt_set_dof(m_Joint,ithDoF,value);
}

bool Joint::setFreeFlyerFromMatrix( const Eigen::Affine3d& T )
{
  if( m_Joint->type == P3D_FREEFLYER || m_Joint->type == P3D_QUADROTOR )
  {
    Eigen::Vector3d trans = T.translation();
    p3d_jnt_set_dof(m_Joint,0,trans[0]);
    p3d_jnt_set_dof(m_Joint,1,trans[1]);
    p3d_jnt_set_dof(m_Joint,2,trans[2]);
    
    cout << "trans : " << endl << trans << endl;
    
    Eigen::Vector3d rot = T.linear().eulerAngles(0, 1, 2);
    p3d_jnt_set_dof(m_Joint,3,rot[0]);
    p3d_jnt_set_dof(m_Joint,4,rot[1]);
    p3d_jnt_set_dof(m_Joint,5,rot[2]);
    return true;
  }
  else
    return false;
}

bool Joint::isJointDofUser(int ithDoF) const
{  
  return p3d_jnt_get_dof_is_user(m_Joint,ithDoF);
}

void Joint::getDofBounds(int ithDoF, double& vmin, double& vmax) const
{
	vmin = m_Joint->dof_data[ithDoF].vmin;
	vmax = m_Joint->dof_data[ithDoF].vmax;
}

unsigned int Joint::getNumberOfDof() const
{
	return m_Joint->dof_equiv_nbr;
}

unsigned int Joint::getIndexOfFirstDof() const
{
	return m_Joint->index_dof;
}

void Joint::setConfigFromDofValues(Configuration& q)
{
	for(int j=0; j<m_Joint->dof_equiv_nbr; j++) 
	{
		int k = m_Joint->index_dof + j;
		q[k] = getJointDof(j); 
	}
}

Joint* Joint::getPreviousJoint()
{
  Joint* prevJnt=NULL; int found=0;
  
  for (unsigned int i=0; i<m_Robot->getNumberOfJoints(); i++ ) 
  {
    Joint* jnt = m_Robot->getJoint(i);
    if (m_Joint->prev_jnt == jnt->m_Joint ) 
    {
      found++;
      prevJnt = jnt;
    }
  }
  
  if (found == 1) {
    return prevJnt;
  }
  else if (found > 1 ) {
    cout << "Found : " << found << " prev. joints!!!" << endl; 
  }

  return NULL;
}

std::vector<Joint*> Joint::getAllPrevJoints()
{
  Joint* jnt( getPreviousJoint() );
 
  std::vector<Joint*> prevJoints;
  prevJoints.clear();

  cout << "Prev Joints ---------" << endl; 
  while (jnt != NULL)
  {
    cout << "jnt->m_id = " << jnt->m_id << endl;
    prevJoints.push_back( jnt );
    jnt = jnt->getPreviousJoint();
  }
  
  return prevJoints;
}


/**
 * Compute the dimension of the space swept by the joint.
 */
unsigned Joint::computeDimension()
{
	unsigned dimension(0);

	// prismatic joint (1D)
	if (m_Joint->type == P3D_TRANSLATE)
	{
		if (m_Joint->dof_data[0].vmax != m_Joint->dof_data[0].vmin)
			++dimension;
	}

	// revolute joint (1D)
	else if (m_Joint->type == P3D_ROTATE)
	{
		if (m_Joint->dof_data[0].vmax != m_Joint->dof_data[0].vmin)
			++dimension;
	}

	// planar joint (2D)
	else if (m_Joint->type == P3D_PLAN)
	{
		for (unsigned i(0); i < 2; ++i)
		{
			if (m_Joint->dof_data[i].vmax != m_Joint->dof_data[i].vmin)
				++dimension;
		}
	}

	// freefly joint (6D)
	else if (m_Joint->type == P3D_FREEFLYER)
	{
		for (unsigned i(0); i < 6; ++i)
		{
			if (m_Joint->dof_data[i].vmax != m_Joint->dof_data[i].vmin)
				++dimension;
		}
	}

	return dimension;
}


/**
 * Compute the volume of the space swept by the joint.
 */
double Joint::computeSpaceVolume()
{
	double volumeOfSpace(1.);

	// prismatic joint (1D)
	if (m_Joint->type == P3D_TRANSLATE)
	{
		double l = fabs((m_Joint->dof_data[0].vmax - m_Joint->dof_data[0].vmin));
		if (l > 0.)
			volumeOfSpace *= l;
	}

	// revolute joint (1D)
	else if (m_Joint->type == P3D_ROTATE)
	{
		double l = m_Joint->dist * fabs((m_Joint->dof_data[0].vmax - m_Joint->dof_data[0].vmin));
		if (l > 0.)
			volumeOfSpace *= l;
	}

	// planar joint (2D)
	else if (m_Joint->type == P3D_PLAN)
	{
		for (unsigned i(0); i < 2; ++i)
		{
			double l = fabs((m_Joint->dof_data[i].vmax - m_Joint->dof_data[i].vmin));
			if (l > 0.)
				volumeOfSpace *= l;
		}
	}

	// freefly joint (6D)
	else if (m_Joint->type == P3D_FREEFLYER)
	{
		for (unsigned i(0); i < 3; ++i)
		{
			double l = fabs((m_Joint->dof_data[i].vmax - m_Joint->dof_data[i].vmin));
			if (l > 0.)
				volumeOfSpace *= l;
		}
		for (unsigned i(3); i < 6; ++i)
		{
			double l = m_Joint->dist * fabs((m_Joint->dof_data[i].vmax - m_Joint->dof_data[i].vmin));
			if (l > 0.)
				volumeOfSpace *= l;
		}
	}

	return volumeOfSpace;
}

