#include "objectrob.hpp"
#include "VimanTags.h"

using namespace std;

INIT_MOVE3D_STATIC_LOGGER(ObjectRob, "planners.devices.ObjectRob");

ObjectRob::ObjectRob():
    _hasVerticalityConstraint(false)
{
}

ObjectRob::~ObjectRob()
{
    for(unsigned int i = 0; i < _supportList.size(); i++)
    {
        delete _supportList.at(i);
    }
    _supportList.clear();
    for(unsigned int i = 0; i < _currentSupportList.size(); i++)
    {
        delete _currentSupportList.at(i);
    }
    _currentSupportList.clear();
}


ObjectRob::ObjectRob(Robot* rob)
{
    _robot = rob;

    string s = "";
    _isObject = true;
    _manipulable = false;
    _support = false;
    _isVirtual = false;
    _hasVerticalityConstraint = false;
}

vector<GeometricForm*> ObjectRob::getSupports()
{
    p3d_matrix4 m;
    p3d_mat4Copy(_robot->getJoint(0)->getJointStruct()->abs_pos,m);
    _currentSupportList.clear();

    for(unsigned int i = 0; i < _supportList.size(); i++)
    {
        GeometricForm* gf = _supportList.at(i);
        GeometricForm* gfTmp = gf->getTransformedForm(m);
        p3d_vector3 znorm, res, norm;
        znorm[0] = 0; znorm[1] = 0; znorm[2] = 1;
        gfTmp->getNorm(norm);
        p3d_vectXprod(norm,znorm,res);
        M3D_TRACE("vectNorm = (" <<norm[0] << " , " << norm[1] << " , " << norm[2] << ")");
        M3D_TRACE("norm = " << p3d_vectNorm(res));
        if (fabs(p3d_vectNorm(res)) <= EPS1 )
            _currentSupportList.push_back(gfTmp);
    }
    return _currentSupportList;
}

void ObjectRob::drawSupports(bool val)
{
    PlanEnv->setBool(PlanParam::drawSurfaces,val);
    if (val)
    {
        vector<GeometricForm*> vgf = getSupports();
        for (unsigned int i = 0; i < vgf.size();i++)
        {
            formToDraw.push_back(vgf.at(i));
        }
    }
    else
    {
        formToDraw.clear();
    }
}


void ObjectRob::cleanSupportList()
{
    formToDraw.clear();
}


//return -1 if not in stable rotation
int ObjectRob::isInStableRotation()
{
    configPt q = _robot->getCurrentPos()->getConfigStruct();
    for (unsigned int i = 0; i < _stableRotations.size(); i++)
    {
        configPt tmp = _stableRotations.at(i);
        if (fabs(q[9] - tmp[9]) < 0.01 && fabs(q[10] - tmp[10]) < 0.01)
            return 11;
        else if (fabs(q[10] - tmp[10]) < 0.01 && fabs(q[11] - tmp[11]) < 0.01)
            return 9;
        else if (fabs(q[9] - tmp[9]) < 0.01 && fabs(q[11] - tmp[11]) < 0.01)
            return 10;
    }
    return -1;
}

double ObjectRob::distToPoint(p3d_point p)
{
    double dist = 1000;

    double tmp;
    getSupports();
    for (unsigned int i = 0; i < _currentSupportList.size(); i++)
    {
        GeometricForm* prec= _currentSupportList.at(i);
        tmp = prec->getDistanceFromPoint(p);
        if (tmp < dist)
            dist = tmp;
    }
    return dist;


}

double ObjectRob::distToPointZ(p3d_point p)
{
    double dist = 1000;

    double tmp;
    for (unsigned int i = 0; i < _supportList.size(); i++)
    {
        GeometricForm* prec= _supportList.at(i);
        tmp = prec->getPlannarDistanceFromPoint(p);
        if (tmp < dist)
            dist = tmp;
    }
    return dist;
}

configPt ObjectRob::getStableRotation(unsigned int id, double rot)
{

    p3d_matrix4 K, M, res;

    if (_stableRotations.size() <= id || id < 0)
        return NULL;

    configPt q = p3d_copy_config(_robot->getRobotStruct(),_stableRotations.at(id));

    p3d_vector3 v;
    for (int i = 0; i<=2; i++)
    {
        if (q[i+9] == OPEN_VALUE)
        {
            v[i] = 1;
            q[i+9] = 0;
        }
        else
            v[i] = 0;
    }


    p3d_set_and_update_this_robot_conf(_robot->getRobotStruct(), q);

    p3d_mat4Copy(_robot->getJoint(0)->getJointStruct()->abs_pos, K);

//    p3d_mat4Print(K, "K");

    v[0] = 0; v[1] = 0; v[2] = 1;

    p3d_mat4Rot(M,v,rot);
//    p3d_mat4Print(M, "rotation matric");

    p3d_mat4Mult(M, K, res);
    //p3d_mat4Print(res, "res");


    double Tx, Ty, Tz;
    double Rx, Ry, Rz;
    p3d_mat4ExtractPosReverseOrder(res, &Tx, &Ty, &Tz,
                                   &Rx, &Ry, &Rz);

    q[6] = Tx;
    q[7] = Ty;
    q[8] = Tz;
    q[9] = Rx;
    q[10] = Ry;
    q[11] = Rz;

    p3d_set_and_update_this_robot_conf(_robot->getRobotStruct(), q);


    return q;
}

void ObjectRob::setToVirtual(bool val)
{
    if (!val)
        p3d_set_col_on(_robot->getRobotStruct());
    else
        p3d_set_col_off(_robot->getRobotStruct());
}
bool ObjectRob::hasVerticalityConstraint() const
{
    return _hasVerticalityConstraint;
}

void ObjectRob::setHasVerticalityConstraint(bool hasVerticalityConstraint)
{
    _hasVerticalityConstraint = hasVerticalityConstraint;
}
Eigen::Vector3d ObjectRob::keepVerticalAxis() const
{
    return _keepVerticalAxis;
}

void ObjectRob::setKeepVerticalAxis(const Eigen::Vector3d &keepVerticalAxis)
{
    _keepVerticalAxis = keepVerticalAxis;
}
double ObjectRob::maxVerticalTolerance() const
{
    return _maxVerticalTolerance;
}

void ObjectRob::setMaxVerticalTolerance(double angle)
{
    _maxVerticalTolerance = angle;
}

