/*
 * This file is intended to test the HPP Wrapper for Romeo robot
 *
 * author : Renaud Viry
 * 			renaud.viry@laas.fr
 * Copyright 2014 LAAS/CNRS. All rights reserved.
 */

#include "romeo_test.hpp"

	RomeoWrap *romeo_;

void init(Robot *romeo)
{
	romeo_ = new RomeoWrap(romeo);
	std::vector<double> bounds;
	bounds.push_back(3);
	bounds.push_back(6);
	bounds.push_back(5);
	bounds.push_back(10);
	bounds.push_back(0);
	bounds.push_back(2);
	romeo_->setJointBound("base_joint_xyz", bounds);

	romeo_->setCurrentConfig(romeo_->createHppConfiguration(romeo->getCurrentPos()));
}

void result(int testNumber, double stabilityScore, double overallTime)
{
	std::cout << "Measures gives, over " << testNumber << " tests : "
			  << 100*stabilityScore/testNumber << "% configuration stable "
			  << overallTime/testNumber << "s mean time for computing stability"
			  << std::endl;
}

void test_pseudo_random_stability(int testNum)
{

	std::cout << "-------------------------------------------------------------" << std::endl
			  << "-------------------- Pseudo Random (" << testNum << ") --------------------" << std::endl;
	double stability = 0;
	double timeOut = 0;
	struct timeval start, end;

	hpp::core::ConfigurationPtr_t config = romeo_->getCurrentConfig();

	gettimeofday(&start, NULL);
	bool stable = romeo_->isPositionStable(romeo_->createConfiguration(config));
	gettimeofday(&end, NULL);
	double timer = (double)(((end.tv_sec - start.tv_sec) * 1000. + (end.tv_usec - start.tv_usec) / 1000.) / 1000.);
	std::cout << "Initial position is " << (stable ? "stable" : "not stable") << " (in " << timer << "s)" << std::endl;

	std::map<string, bool> randomMap;
	randomMap["leftArm"] = true;
	randomMap["rightArm"] = true;
	randomMap["leftLeg"] = false;
	randomMap["rightLeg"] = false;
	randomMap["torso"] = true;
	randomMap["neck"] = true;
	randomMap["head"] = true;
	std::cout << "Random map : " << "leftArm  : " << randomMap["leftArm"]  << std::endl
			  << "             " << "rightArm : " << randomMap["rightArm"] << std::endl
			  << "             " << "leftLeg  : " << randomMap["leftLeg"]  << std::endl
			  << "             " << "rightLeg : " << randomMap["rightLeg"] << std::endl
			  << "             " << "torso    : " << randomMap["torso"]    << std::endl
			  << "             " << "neck     : " << randomMap["neck"]     << std::endl
			  << "             " << "head     : " << randomMap["head"]     << std::endl
			  << std::endl;

	for (int i = 0; i < testNum; i++)
	{
		std::cout << "Test pseudo random #" << i << std::endl;
		romeo_->randomizeConfig(config, randomMap);
		gettimeofday(&start, NULL);
		bool stable = romeo_->isPositionStable(romeo_->createConfiguration(config));
		gettimeofday(&end, NULL);
		double timer = (double)(((end.tv_sec - start.tv_sec) * 1000. + (end.tv_usec - start.tv_usec) / 1000.) / 1000.);
		timeOut += timer;
		if (stable)
		{
			std::cout << "     is stable !" << std::endl;
			stability ++;
			// Calculate mean, min, max and standard deviation of distance ?
			// store (index, config, distance, time) to replay it later ?
		}
	}

	result(testNum, stability, timeOut);

	std::cout << "-------------------------------------------------------------" << std::endl;
}

void test_random_stability(int testNum)
{
	std::cout << "-------------------------------------------------------------" << std::endl
			  << "-------------------- Fully  Random (" << testNum << ") --------------------" << std::endl;
	double stability = 0;
	double timeOut = 0;
	struct timeval start, end;

	hpp::core::ConfigurationPtr_t config = romeo_->getCurrentConfig();

	gettimeofday(&start, NULL);
	bool stable = romeo_->isPositionStable(romeo_->createConfiguration(config));
	gettimeofday(&end, NULL);
	double timer = (double)(((end.tv_sec - start.tv_sec) * 1000. + (end.tv_usec - start.tv_usec) / 1000.) / 1000.);
	std::cout << "Initial position is " << (stable ? "stable" : "not stable") << " (in " << timer << "s)" << std::endl;

	for (int i = 0; i < testNum; i++)
	{
		std::cout << "Test fully random #" << i << std::endl;
		romeo_->randomizeFullConfig(config);
		gettimeofday(&start, NULL);
		bool stable = romeo_->isPositionStable(romeo_->createConfiguration(config));
		gettimeofday(&end, NULL);
		double timer = (double)(((end.tv_sec - start.tv_sec) * 1000. + (end.tv_usec - start.tv_usec) / 1000.) / 1000.);
		timeOut += timer;
		if (stable)
		{
			std::cout << "     is stable !" << std::endl;
			stability ++;
			// Calculate mean, min, max and standard deviation of distance ?
			// store (index, config, distance, time) to replay it later ?
		}
	}

	result(testNum, stability, timeOut);

	std::cout << "-------------------------------------------------------------" << std::endl;
}

void test_stability(int testNum)
{
	std::cout << "-------------------------------------------------------------" << std::endl
			  << "------------------- Projected stable (" << testNum << ") ------------------" << std::endl;
	double stability = 0;
	double timeOut = 0;
	struct timeval start, end;
	double *csTimer;
	csTimer = new double(0);

	hpp::core::ConfigurationPtr_t config = romeo_->getCurrentConfig();

	gettimeofday(&start, NULL);
	bool stable = romeo_->isPositionStable(romeo_->createConfiguration(config));
	gettimeofday(&end, NULL);
	double timer = (double)(((end.tv_sec - start.tv_sec) * 1000. + (end.tv_usec - start.tv_usec) / 1000.) / 1000.);
	std::cout << "Initial position is " << (stable ? "stable" : "not stable") << " (in " << timer << "s)" << std::endl;

	for (int i = 0; i < testNum; i++)
	{
		std::cout << "Test random stabilized #" << i << std::endl;
		romeo_->randomizeFullConfig(config);
		romeo_->applyConstraints(config, csTimer);
		std::cout << "    applyConstraints returned in " << csTimer[0] << "s" << std::endl;
		gettimeofday(&start, NULL);
		bool stable = romeo_->isPositionStable(romeo_->createConfiguration(config));
		gettimeofday(&end, NULL);
		double timer = (double)(((end.tv_sec - start.tv_sec) * 1000. + (end.tv_usec - start.tv_usec) / 1000.) / 1000.);
		timeOut += timer;
		if (stable)
		{
			std::cout << "     is stable !" << std::endl;
			stability ++;
			// Calculate mean, min, max and standard deviation of distance ?
			// store (index, config, distance, time) to replay it later ?
		}
	}

	result(testNum, stability, timeOut);

	std::cout << "-------------------------------------------------------------" << std::endl;
}


void test(Robot *romeo)
{
	std::cout << std::endl
			  << " !!! Test stability criterion on Romeo + config randomization !!! " << std::endl
			  << std::endl;
	init(romeo);

	std::cout << std::endl;
	test_pseudo_random_stability(25);
	std::cout << std::endl;
	test_random_stability(25);
	std::cout << std::endl;
	test_stability(25);
	std::cout << std::endl;
}