#include "HRICS_MultiAgentGrid.hpp"

#include <sstream>
#include "P3d-pkg.h"
#include "Collision-pkg.h"
#include "Graphic-pkg.h"
#include "planEnvironment.hpp"
#include <time.h>
#include "time.h"
#include "Chronometer.hpp"

#include "Logging/Logger.h"

#include <boost/random/normal_distribution.hpp>
#include <boost/random/uniform_int.hpp>
#include <boost/random/mersenne_twister.hpp>
#include <boost/random/variate_generator.hpp>

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif

using namespace std;
using namespace tr1;
using namespace HRICS;

INIT_MOVE3D_STATIC_LOGGER(MultiAgentGrid,"planners.hrics.mho.grid");


// import most common Eigen types 
//using namespace Eigen;
using namespace Eigen;


//uses monotonic clock
//set duration to time ellapsed since last call with duration=0
//when called with duration=0 (default), set internal values only
//returns the value returned by subsequent call of clock_gettime(CLOCK_MONOTONIC,...)
int exec_time(timespec *duration=0){
    static timespec ts_start;
    if(duration==0){
        return clock_gettime(CLOCK_MONOTONIC,&ts_start);
    }else{
        timespec ts_end;
        if(clock_gettime(CLOCK_MONOTONIC,&ts_end)==0){
            duration->tv_sec=ts_end.tv_sec-ts_start.tv_sec;
            duration->tv_nsec=ts_end.tv_nsec-ts_start.tv_nsec;
        }else{return -1;}
    }
    return 0;
}

class MACellDistComp
{
public:

        bool operator()(pair<double,MultiAgentCell*> first, pair<double,MultiAgentCell*> second)
        {
                return ( first.first < second.first );
        }

} MACellDistCompObject;

char MultiAgentGrid::MetaAgentInGrid::_letter_id_cnt = 'A';
MultiAgentGrid::MetaAgentInGrid::MetaAgentInGrid(unsigned int agent_number): letter_id(_letter_id_cnt++){
    this->_realAgents.clear();
    this->_cylinder=0;
    //this->_frontierFullyChecked.assign(agent_number,false);
}

MultiAgentGrid::MetaAgentInGrid::s_ptr MultiAgentGrid::MetaAgentInGrid::shared_clone(){
    s_ptr ptr(new MetaAgentInGrid(this->_frontierFullyChecked.size()));
    ptr->_realAgents.clear();
    ptr->setCylinder(this->getCylinder());
    return ptr;
}

void MultiAgentGrid::MetaAgentInGrid::addAgent(AgentInGrid *a, MetaAgentInGrid::s_ptr s_ptr_this){
    if(a->getMetaAgent() && a->getMetaAgent()!=s_ptr_this){
        a->getMetaAgent()->removeAgent(a);
    }
    if(std::find(_realAgents.begin(),_realAgents.end(),a) == _realAgents.end()){
        this->_realAgents.push_back(a);
    }
    a->setMetaAgent(s_ptr_this);
}
void MultiAgentGrid::MetaAgentInGrid::removeAgent(AgentInGrid *a){
    vector<AgentInGrid*>::iterator it = std::find(_realAgents.begin(),_realAgents.end(),a);
    if(it!=_realAgents.end()){
        _realAgents.erase(it);
    }
}

void MultiAgentGrid::MetaAgentInGrid::setCylinder(Robot *c){
    _cylinder=c;
    if(c){
        double x_min,x_max;
        x_min=c->getRobotStruct()->BB.xmin;
        x_max=c->getRobotStruct()->BB.xmax;
        _cyl_radius=(x_max-x_min)/2;
        assert(_cyl_radius>0);
    }
}


//---------------------------------------------------------------------------
// Grid
//---------------------------------------------------------------------------
MultiAgentGrid::MultiAgentGrid() :
        API::TwoDGrid(),
        mAgents(0),
        _use_reachable_for_frontier_search(false),
        _save_log_frontier(false),
        _drawAgentReachGrid(false),
        _agent_index_to_draw_grid(0),
        _drawLog(false)
{
    resetData();
}

MultiAgentGrid::MultiAgentGrid(double pace, vector<double> envSize) :
    API::TwoDGrid(pace,envSize),
    mAgents(0),
    _use_reachable_for_frontier_search(false),
    _save_log_frontier(false),
    _drawAgentReachGrid(false),
    _agent_index_to_draw_grid(0),
    _drawLog(false)
{
    resetData();
}

MultiAgentGrid::MultiAgentGrid(double pace, vector<double> envSize, vector<Robot *> &agents) :
    API::TwoDGrid(pace,envSize),
    _use_reachable_for_frontier_search(false),
    _save_log_frontier(false),
    _drawAgentReachGrid(false),
    _agent_index_to_draw_grid(0),
    _drawLog(false)
{
    setAgents(agents);
    resetData();
    createAllCells();
    cout << "Number total of cells = " << _nbCellsX*_nbCellsY << endl;
    ENV.setDouble(Env::optimalDistFactor,0.5);
    ENV.setDouble(Env::robotMaximalDistFactor,0);
    ENV.setDouble(Env::gazeAngleFactor,1.0);
}

MultiAgentGrid::~MultiAgentGrid(){
    for (unsigned int i=0;i<mAgentsInGrid.size();i++){
        delete getAgentStructAt(i);
    }
    mAgentsInGrid.clear();
}

void MultiAgentGrid::resetData(){
    _agent_index_to_draw_grid=0;
    _draw_a_star=false;
#ifdef DRAW_OPTIMIZE_FRONTIER
    _draw_grid_for_optim=false;
#endif
#if defined(DRAW_FRONTIER_HAND_IN_HAND) || defined(DRAW_OPTIMIZE_FRONTIER)
    _draw_frontier_search=false;
#endif

    this->_max_alternative_distances_for_cells=16;
    //TODO add data to reset/initialize here
}

void MultiAgentGrid::init(/*pair<double,double> minMax*/)
{
    _showText = PlanEnv->getBool(PlanParam::env_showText);
    if (_showText)
    {
        cout << "in: MultiAgentGrid::init(pair<double,double> minMax)\n" << endl;
    }
    //vector<pair<Robot*,shared_ptr<Configuration> > > initConfiguration;

    setAgents(mAgents);
    initCylinders();
    for(int i=0;i<getNumberOfCells();++i){
        dynamic_cast<MultiAgentCell*>(getCell(i))->resetData();
    }

    if (_showText)
    {
        cout << "MultiAgentGrid::init(pair<double,double> minMax) end" << endl;
    }
}

void MultiAgentGrid::setAgents(std::vector<Robot*> &agents) {
    mAgents = agents;
    mAgentsInGrid.clear();
    MetaAgentInGrid::s_ptr meta(new MetaAgentInGrid(mAgents.size()));
    //meta->frontierFullyChecked().assign(mAgents.size(),false);
    for (unsigned int i=0;i<mAgents.size();++i){
        AgentInGrid * aig=new AgentInGrid;
        aig->setRobot(mAgents[i]);
        aig->setMaxDist(0);
        aig->setSpeed(1);//TODO change this default speed value to a real one
        aig->setDistanceComputed(false);
        mAgentsInGrid.push_back(aig);
        meta->addAgent(aig,meta);
    }

}

void MultiAgentGrid::setAgentTarget(unsigned int a){
    M3D_DEBUG("MHO grid: set agent target in");
    Eigen::Vector2d pos;
    confPtr_t q=getAgentStructAt(a)->getRobot()->getGoTo();
    pos.x()=q->at(6+0);
    pos.y()=q->at(6+1);
    MultiAgentCell* cell= dynamic_cast<MultiAgentCell*>(this->getCell(pos));
    if (cell){
        getAgentStructAt(a)->setTarget(cell);
    }else{
        cerr << "MHO grid ERROR: taget conf out of grid" << endl;
        throw std::exception();
    }
    cout<< "MHO grid: set agent target out"<<endl;
}

//void MultiAgentGrid::computeAccessibility(){
//
//    if (_showText)
//    {
//        cout << "MultiAgentGrid::computeAccessibility() : Compute Accessibility of each cell for each agent" << endl;
//    }
//
//    exec_time();
//
//    resetAllAccessibility();
//
//    if (_showText)
//    {
//        cout << "MultiAgentGrid::computeAccessibility() : Accessibility reset for all agents" << endl;
//    }
//
//    //TODO deactivate collisions all with manipulable objects
//    for (unsigned int i=0;i<mAgentsInGrid.size();++i){
//        p3d_col_deactivate_robot(getAgentStructAt(i)->robot->getRobotStruct());
//        for (unsigned int j=0;j<mAgentsInGrid.size();++j){
//            //deactivates collisions between cylinders and robots
//            p3d_col_deactivate_rob_rob(getAgentStructAt(i)->robot->getRobotStruct(),getAgentStructAt(j)->cylinder->getRobotStruct());
//        }
//    }
//
//
//    if (_showText)
//    {
//        cout << "MultiAgentGrid::computeAccessibility() : agents colisions deactivated" << endl;
//    }
//
//    for (unsigned int x =0;x<_nbCellsX;++x)
//    {
//        for (unsigned int y =0;y<_nbCellsY;++y)
//        {
//            MultiAgentCell* Cell = dynamic_cast<MultiAgentCell*>(getCell(x,y));
//
//            for (unsigned int i_ag=0;i_ag<mAgents.size();++i_ag){
//                Cell->computeReach(i_ag);
//                if (Cell->isAccessible(i_ag))
//                {
//                    getAgentStructAt(i_ag)->accessible.push_back(Cell);
//                }
//            }
//        }
//    }
//
//    if (_showText)
//    {
//        cout << "MultiAgentGrid::computeAccessibility() : Accessibility computed" << endl;
//    }
//
//    for (unsigned int i=0;i<mAgentsInGrid.size();++i){
//        for (unsigned int j=0;j<mAgentsInGrid.size();++j){
//            p3d_col_activate_rob_rob(getAgentStructAt(i)->robot->getRobotStruct(),getAgentStructAt(j)->cylinder->getRobotStruct());
//        }
//    }
//
//    exec_time(&_computation_stat.access_total_time);
//
//    if (_showText){
//        cout << "MultiAgentGrid::computeAccessibility() : agents collisions reactivated" << endl;
//        cout << "MultiAgentGrid::computeAccessibility() : done.\n" << endl;
//    }
//
//}


void MultiAgentGrid::computeAccessAndDistCombined(){
    exec_time();

    resetAllAccessibility();
    resetDistances();

    for (unsigned int i=0;i<mAgentsInGrid.size();++i){
//        p3d_col_deactivate_robot(getAgentStructAt(i)->robot->getRobotStruct());
        for (unsigned int j=0;j<mAgentsInGrid.size();++j){
            p3d_col_deactivate_rob_rob(getAgentStructAt(i)->getRobot()->getRobotStruct(),getAgentStructAt(j)->getCylinder()->getRobotStruct());
            if(getAgentStructAt(i)->getCylinder()->getRobotStruct()!=getAgentStructAt(j)->getCylinder()->getRobotStruct()){
                p3d_col_deactivate_rob_rob(getAgentStructAt(i)->getCylinder()->getRobotStruct(),getAgentStructAt(j)->getCylinder()->getRobotStruct());
            }
        }
    }

    for(unsigned int i=0;i<mAgentsInGrid.size();++i){
        initAllCellState(); //reset exploration status for all cells

        AgentInGrid * agent= getAgentStructAt(i);
        Robot * rob=agent->getRobot();
        shared_ptr<Configuration> q_agent_cur = rob->getCurrentPos();

        //TODO test with baseJnt for human too (have to implement/wait for  it)
        int firstIndexOfBaseDof=6;
        if(agent->isHuman()){
            firstIndexOfBaseDof=rob->getJoint("Pelvis")->getIndexOfFirstDof();
        }else{
            firstIndexOfBaseDof= dynamic_cast<p3d_jnt*>(rob->getRobotStruct()->baseJnt)->user_dof_equiv_nbr;
        }

        this->setAgentTarget(i);
        Vector2d pos;
        pos[0] = (*q_agent_cur)[firstIndexOfBaseDof + 0];
        pos[1] = (*q_agent_cur)[firstIndexOfBaseDof + 1];

        MultiAgentCell* cell=dynamic_cast<MultiAgentCell*>(getCell(pos));
        if(cell){
            agent->setOrigin(cell);
            this->computeDistancesForAgent(cell,i);
        }else{
            cerr<< "MHO Grid accessibility computation : ERROR agent "<<getAgents()[i]->getName()<< " (i="<<i<<") is not in a valid position"<<endl;
        }
    }


    for (unsigned int i=0;i<mAgentsInGrid.size();++i){
        for (unsigned int j=0;j<mAgentsInGrid.size();++j){
            p3d_col_activate_rob_rob(getAgentStructAt(i)->getRobot()->getRobotStruct(),getAgentStructAt(j)->getCylinder()->getRobotStruct());
            if(getAgentStructAt(i)->getCylinder()->getRobotStruct()!=getAgentStructAt(j)->getCylinder()->getRobotStruct()){
                p3d_col_activate_rob_rob(getAgentStructAt(i)->getCylinder()->getRobotStruct(),getAgentStructAt(j)->getCylinder()->getRobotStruct());
            }
        }
    }

    exec_time(&_computation_stat.access_total_time);


}

void MultiAgentGrid::initCylinders(){
    //TODO CYLINDERS ??
    humCyl=0;
    robotCyl=0;

    if (mAgents.size()==0){return;}
    for (int i=0; i<XYZ_ENV->nr; i++)
    {
        string name(XYZ_ENV->robot[i]->name);
        if(name.find("HUMCYLINDER") != string::npos )
        {
            humCyl = new Robot(XYZ_ENV->robot[i]);
        }else if (name.find("PR_2CYLINDER") != string::npos ){
            robotCyl = new Robot(XYZ_ENV->robot[i]);
        }
        if(humCyl && robotCyl){break;}//stop loop when both are found
    }
    if(!humCyl || !robotCyl){
        cerr<<"MultiAgentGrid::initCylinders() : error : missing cylinder"<<endl
              <<"MultiAgentGrid::initCylinders() : ..... : no Robot named ";
        if(!humCyl){
            cerr<<"\"HUMCYLINDER\" found"<<endl;
        }else{
            cerr<<"\"PR_2CYLINDER\" found"<<endl;
        }
    }
    //associate cylinders to agents (human or robots)
    MetaAgentInGrid::s_ptr meta_robot = getAgentStructAt(0)->getMetaAgent();
    MetaAgentInGrid::s_ptr meta_human = meta_robot->shared_clone();
    meta_robot->setCylinder(robotCyl);
    meta_human->setCylinder(humCyl);
    for (unsigned int i=0;i<mAgentsInGrid.size();++i){
        assert(getAgentStructAt(i)->getMetaAgent() == meta_robot);
        if(getAgentStructAt(i)->isHuman()){
            meta_human->addAgent(getAgentStructAt(i),meta_human);
        }
    }
}

API::TwoDCell* MultiAgentGrid::createNewCell(unsigned int index,unsigned  int x,unsigned  int y )
{
    Vector2i coord;

    coord[0] = x;
    coord[1] = y;

    if (index == 0)
    {
        return new MultiAgentCell( 0, coord, _originCorner , this );
    }
    API::TwoDCell* newCell = new MultiAgentCell( index, coord, computeCellCorner(x,y) , this );
    return newCell;
}

void MultiAgentGrid::_drawCell(MultiAgentCell *cell,double *color){
    if(cell && color){
        double depth;
        Vector2d center = cell->getCenter();
        depth =0.1;
        glColor3d(color[0],color[1],color[2]);
        glVertex3d( (double)(center[0] - _cellSize[0]/2) + 0.01 , (double)(center[1] - _cellSize[1]/2) + 0.01, depth);
        glVertex3d( (double)(center[0] + _cellSize[0]/2) - 0.01 , (double)(center[1] - _cellSize[1]/2) + 0.01, depth);
        glVertex3d( (double)(center[0] + _cellSize[0]/2) - 0.01 , (double)(center[1] + _cellSize[1]/2) - 0.01, depth);
        glVertex3d( (double)(center[0] - _cellSize[0]/2) + 0.01 , (double)(center[1] + _cellSize[1]/2) - 0.01, depth);
    }

}

void MultiAgentGrid::draw()
{
    /*if( mRobot == 0x00 )
    {
        std::cout << "Error : PlanGrid::draw() => No Robot "  << std::endl;
    }*/

#ifdef LIGHT_PLANNER
    deactivateCcCntrts(mAgents[0]->getRobotStruct(),-1);
#else
    cout << "Warning: Light Planner not compiled" << endl;
#endif
	
//	double colorvector[4];

//    colorvector[0] = 1.0;       //red
//    colorvector[1] = 0.5;       //green
//    colorvector[2] = 0.0;       //blue
//    colorvector[3] = 0.05;       //transparency


    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    //    glBlendFunc(GL_SRC_ALPHA, GL_ONE);

    glDisable(GL_LIGHTING);
    glDisable(GL_LIGHT0);

    glEnable(GL_CULL_FACE);
    glBegin(GL_QUADS);


    double color[3];


    /*std::vector<std::pair<double,MultiAgentCell*> > sortedCells;
    std::vector<double> cellToDraw;
    std::vector<double> cellsCost;
    double maxCost = 0;
    double minCost = 100;
    if (PlanEnv->getBool(PlanParam::env_drawDistGrid))
    {
        sortedCells = getSortedGrid();
        for (unsigned int k = 0; k < sortedCells.size(); k++)
        {
            cellToDraw.push_back(sortedCells[k].second->getIndex());
            cellsCost.push_back(sortedCells[k].first);
            if(sortedCells[k].first > maxCost)
            {
                maxCost = sortedCells[k].first;
            }
            if (sortedCells[k].first < minCost)
            {
                minCost = sortedCells[k].first;
            }
       }
    }*/

#if defined(DRAW_FRONTIER_HAND_IN_HAND) || defined(DRAW_OPTIMIZE_FRONTIER)
    if(this->_draw_frontier_search){
        if(_draw_frontier_search_is_pair_ok){
            color[1]=1;//green
        }else{
            color[0]=1;//red
        }
        _drawCell(_draw_frontier_search_cells.first,color);
        _drawCell(_draw_frontier_search_cells.second,color);
        this->_drawFrontiers=false;
        this->_drawAgentReachGrid=false;
    }
#endif
#ifdef DRAW_OPTIMIZE_FRONTIER
    if(this->_draw_grid_for_optim){
        double max_dist=0;
        for(unsigned int j=0;j<_draw_grid_for_optim_cells_agents.size();++j){
            max_dist+=getAgentStructAt(_draw_grid_for_optim_cells_agents[j].second)->getMaxDist();
        }
        for (unsigned int i=0;i<this->_nbCellsX*this->_nbCellsY;++i){
            MultiAgentCell * c=dynamic_cast<MultiAgentCell*>(this->getCell(i));
            double dist=0;
            for(unsigned int j=0;j<_draw_grid_for_optim_cells_agents.size();++j){
                unsigned int a=_draw_grid_for_optim_cells_agents[j].second;
                MultiAgentCell* o=_draw_grid_for_optim_cells_agents[j].first;
                dist+=c->getAltDistance(a,o);
            }
            if(dist<numeric_limits<double>::infinity()){
                double colorRation=0;
                colorRation = dist/max_dist;
                GroundColorMixGreenToRed(color,colorRation);
                this->_drawCell(c,color);
            }
        }
    }
#endif
#ifdef DRAW_FRONTIER_STEPS
    if (_draw_frontier_steps){
        for (unsigned int i=0;i<_draw_frontier_steps_cells.size();++i){
            _drawCell(_draw_frontier_steps_cells[i],color);
        }
    }else
#endif
    if(this->_drawAgentReachGrid){
        unsigned int index_to_draw_begin=this->_agent_index_to_draw_grid;
        unsigned int index_to_draw_end=index_to_draw_begin+1;
        if(index_to_draw_begin==mAgentsInGrid.size()){
            index_to_draw_begin=0;
            index_to_draw_end=mAgentsInGrid.size();
        }
        for (unsigned int index_to_draw=index_to_draw_begin;index_to_draw<index_to_draw_end;++index_to_draw){
            vector<MultiAgentCell*> * ptr_to_cells=&getAgentStructAt(index_to_draw)->reachable();
            if(this->_drawBoundariesOnly){
                ptr_to_cells=&getAgentStructAt(index_to_draw)->boundaries();
            }


            for (vector<MultiAgentCell*>::iterator it=ptr_to_cells->begin();it!=ptr_to_cells->end();++it){
                MultiAgentCell* Cell=*it;
                if(Cell->isReachable(index_to_draw)){
                    double colorRation=0;
                    colorRation = Cell->getDistance(index_to_draw) / getAgentStructAt(index_to_draw)->getMaxDist();
                    GroundColorMixGreenToRed(color,colorRation);
                    _drawCell(Cell,color);
                }
            }
        }
    }else if (this->_drawFrontiers){
        if(_drawLog && _save_log_frontier){
            unsigned int a1=this->_drawFrontierIndices.first,a2=this->_drawFrontierIndices.second;
            if(a1<mAgentsInGrid.size() && a2<mAgentsInGrid.size()){
                unsigned int ind=getFrontierIndex(a1,a2);
                if(_log_frontier_seeking.size()>ind && _log_frontier_seeking[ind].size()>0){
                    unsigned int window_size=min<unsigned int>(_log_draw_window_size,_log_frontier_seeking[ind].size());
                    if(_log_draw_window+window_size>=this->_log_frontier_seeking[ind].size()){
                        _log_draw_window=_log_frontier_seeking[ind].size()-window_size;
                    }
                    //unsigned int i=_log_draw_window;
                    for(unsigned int i=_log_draw_window;i<_log_draw_window+window_size;++i){
                        double d_i=i-_log_draw_window;
                        double colorRationA = d_i/window_size;
                        d_i=i;
                        double colorRationB = d_i/_log_frontier_seeking[ind].size();
                        double colorRation = 0.7*colorRationB+0.3*colorRationA;
                        GroundColorMixGreenToRed(color,colorRation);
                        _drawCell(_log_frontier_seeking[ind][i].second.first,color);
                        _drawCell(_log_frontier_seeking[ind][i].second.second,color);
                    }
                }
            }
        }else{
            color[1]=1;
            unsigned int first_start=_drawFrontierIndices.first,first_end=_drawFrontierIndices.first+1;
            unsigned int second_start=_drawFrontierIndices.second,second_end=_drawFrontierIndices.second+1;
            if(_drawFrontierIndices.first==mAgentsInGrid.size()){
                //draw all frontiers for agent _drawFrontierIndices.first
                first_start=0;first_end=mAgentsInGrid.size();
            }//else draw only one
            if(_drawFrontierIndices.second==mAgentsInGrid.size()){
                //draw all frontiers for agent _drawFrontierIndices.first
                second_start=0;second_end=mAgentsInGrid.size();
            }//else draw only one
            for(unsigned int first=first_start;first<first_end;++first){
                for(unsigned int second=second_start;second<second_end;++second){
                    vector<MultiAgentCell*> const & cells=getFrontierCells(first,second);
                    for (unsigned int i=0;i<cells.size();++i){
                        //_drawCell(cells[i],color);
                        if(cells[i]->getMetaFrontier(first,second).size()){
                            MetaFrontierPtr mf=cells[i]->getMetaFrontier(first,second)[0];
                            if(mf->getCheckedOk()){
                                color[0]=0;
                                color[1]=1;
                            }else{
                                color[0]=1;
                                color[1]=0;
                            }
                            this->_drawCell(mf->cellOf(getAgentStructAt(first)->getMetaAgent()),color);
                        }
                    }
                    //vector<MultiAgentCell*> const & cells2=getFrontierCells(second,first);
                    //color[0]=0;
                    //color[1]=0;
                    //color[2]=1;
                    //for (unsigned int i=0;i<cells2.size();++i){
                    //    _drawCell(cells2[i],color);
                        //if(cells2[i]->getMetaFrontier(second,first).size()>0){
                        //    this->_drawCell(cells2[i]->getMetaFrontier(second,first)[0]->cellOf(getAgentStructAt(second)->getMetaAgent()),color);
                        //}
                    //}

                    //if(_frontiers.size()>getFrontierIndex(first,second)){
                    //    for (unsigned int i=0;i<this->getFrontier(first,second).size();++i){
                    //        this->_drawCell(this->getFrontier(first,second)[i]->cellA,color);
                    //        this->_drawCell(this->getFrontier(first,second)[i]->cellB,color);
                    //    }
                    //}
                }
            }
        }
    }else if ( this->_draw_a_star){
        vector<MultiAgentCell*> * ptr_to_cells=&getAgentStructAt(this->_agent_index_to_draw_grid)->reachable();

        for (vector<MultiAgentCell*>::iterator it=ptr_to_cells->begin();it!=ptr_to_cells->end();++it){
            MultiAgentCell* Cell=*it;
            if(Cell->getClosed()){
                double colorRation=0;
                colorRation = Cell->getTempDist() / getAgentStructAt(this->_agent_index_to_draw_grid)->getMaxDist();
                GroundColorMixGreenToRed(color,colorRation);
                _drawCell(Cell,color);
            }else if(Cell->getOpen()){
                color[0]=1;color[1]=1;color[2]=1;
                _drawCell(Cell,color);
            }
        }
    }else
    if(this->_draw_extra){
        for(std::vector<MultiAgentCell*>::iterator it=_draw_extra_cells.first;it!=_draw_extra_cells.second;++it){
            MultiAgentCell* Cell=*it;
            double tc=Cell->getTempColor();
            color[0]=0;color[1]=0;color[2]=0;//black

            if(tc==-2){
                //tmp cost
                //color[0]=1;color[1]=0.5;color[2]=0;//orange
                color[0]=color[1]=color[2]=0.6;//grey
            }else if(tc == -3 ){
                //invalid cost
                //color[0]=1;color[1]=0;color[2]=0;//red
                color[0]=color[1]=color[2]=0.2;//dark grey
            }else if(tc==-4){
                //traj to draw
                color[0]=1;color[1]=0;color[2]=1;//purple
            }else if(tc == -1){
                //unasigned
                color[0]=0;color[1]=0;color[2]=0;//black
            }else{
                GroundColorMixGreenToRed(color,Cell->getTempColor()/this->_temp_color_max_ratio);
            }
            _drawCell(Cell,color);

            /*if(Cell->getClosed()){
                if(Cell->getTempDist()==-1){
                    color[0]=1;color[1]=0;color[2]=1;//purple
                }else if(Cell->getTempDist()==-2){
                    color[0]=1;color[1]=0.5;color[2]=0;//orange
                }else if(Cell->getTempDist()<numeric_limits<double>::infinity()){
                    color[0]=0;color[1]=1;color[2]=0;//green
                }else {
                    color[0]=1;color[1]=0;color[2]=0;//red
                }
                _drawCell(Cell,color);
            }else if(Cell->getOpen()){
                color[0]=0;color[1]=0;color[2]=0.5;//dark blue
                _drawCell(Cell,color);
            }else{
                color[0]=0;color[1]=0;color[2]=0;//black
                _drawCell(Cell,color);
            }*/
        }
    }

    glEnd();

    glDisable(GL_CULL_FACE);
    glDisable(GL_BLEND);

    //    glEnable(GL_LIGHTING);
    //    glEnable(GL_LIGHT0);
}

unsigned int MultiAgentGrid::getFrontierSearchLogSize(){
    if(_save_log_frontier && _log_frontier_seeking.size()>getFrontierIndex(_drawFrontierIndices) ){
        return _log_frontier_seeking[getFrontierIndex(_drawFrontierIndices)].size();
    }else{
        return 0;
    }
}

void MultiAgentGrid::setCellsToblankCost()
{
    for (int i = 0; i < getNumberOfCells(); i++)
    {
        dynamic_cast<MultiAgentCell*>(this->getCell(i))->setBlankCost();
    }

}



void MultiAgentGrid::initAllCellState()
{
    for (int i = 0; i < getNumberOfCells(); i++)
    {
        dynamic_cast<MultiAgentCell*>(this->getCell(i))->resetExplorationStatus();
    }
}

void MultiAgentGrid::initAllTrajs()
{
    for (int i = 0; i < getNumberOfCells(); i++)
    {
        dynamic_cast<MultiAgentCell*>(this->getCell(i))->resetTraj();
    }
}

void MultiAgentGrid::resetAllAccessibility()
{
    for (int i = 0; i < getNumberOfCells(); i++)
    {
        dynamic_cast<MultiAgentCell*>(this->getCell(i))->resetAccessibility();
        dynamic_cast<MultiAgentCell*>(this->getCell(i))->resetAllDistances();
    }
    //for(unsigned int i;i<mAgentsInGrid.size();++i){
    //    getAgentStructAt(i)->accessible.clear();
    //}
}

void MultiAgentGrid::resetDistances(){
    for(unsigned int i=0;i<mAgentsInGrid.size();++i){
        for (unsigned int reach_i=0;reach_i<getAgentStructAt(i)->reachable().size();++reach_i){
            getAgentStructAt(i)->reachable()[reach_i]->resetDistances(i);
        }
        getAgentStructAt(i)->reachable().clear();
        getAgentStructAt(i)->boundaries().clear();
        getAgentStructAt(i)->setDistanceComputed(false);
        getAgentStructAt(i)->setMaxDist(0);
    }
}

void MultiAgentGrid::computeDistances(){
    if(_showText){
        cout<<"MultiAgentGrid::computeDistances() : begin"<<endl;
    }
    resetDistances();
    timespec foo_useless;
    foo_useless.tv_nsec=foo_useless.tv_sec=0;
    _computation_stat.reachability_time.assign(mAgentsInGrid.size(),foo_useless);
    _computation_stat.reachability_total_time=foo_useless;
    for(unsigned int i=0;i<mAgentsInGrid.size();++i){
        exec_time();
        initAllCellState(); //reset exploration status for all cells

        AgentInGrid * agent= getAgentStructAt(i);
        Robot * rob=agent->getRobot();
        shared_ptr<Configuration> q_agent_cur = rob->getCurrentPos();


        //TODO test with baseJnt for human too (have to implement/wait for  it)
        int firstIndexOfBaseDof=6;
        if(agent->isHuman()){
            firstIndexOfBaseDof=rob->getJoint("Pelvis")->getIndexOfFirstDof();
        }else{
            firstIndexOfBaseDof= dynamic_cast<p3d_jnt*>(rob->getRobotStruct()->baseJnt)->user_dof_equiv_nbr;
        }

        Vector2d pos;
        pos[0] = (*q_agent_cur)[firstIndexOfBaseDof + 0];
        pos[1] = (*q_agent_cur)[firstIndexOfBaseDof + 1];

        //TODO this was to set from an argument in EnvGrid::initGrid() (HRICS_EnvGrid.cpp:284), pertinent for us? why?
        //pos[0] = humanPos[0];
        //pos[1] = humanPos[1];
        MultiAgentCell* cell=dynamic_cast<MultiAgentCell*>(getCell(pos));
        computeDistancesForAgent(cell,i);

        exec_time(&_computation_stat.reachability_time[i]);
        _computation_stat.reachability_total_time.tv_sec+=_computation_stat.reachability_time[i].tv_sec;
        _computation_stat.reachability_total_time.tv_nsec+=_computation_stat.reachability_time[i].tv_nsec;
        if (_computation_stat.reachability_total_time.tv_nsec >= 1000000000L) {		/* Carry? */
                _computation_stat.reachability_total_time.tv_sec++ ;
                _computation_stat.reachability_total_time.tv_nsec = _computation_stat.reachability_total_time.tv_nsec - 1000000000L ;
            }
        if(_showText){
            cout<<"MultiAgentGrid::computeDistances() : done for agent #"<<i<<" : "<<rob->getName()<<endl;
        }
    }
    if(_showText){
        cout<<"MultiAgentGrid::computeDistances() : end"<<endl;
    }
}

struct compare_temporary_estimated_distances_of_cells{
bool operator() (MultiAgentCell * left, MultiAgentCell * right){
    return left->getTempEstim() > right->getTempEstim();
}
};
struct HeuristicDistance{
    enum distance_estim_t {EUCLIDEAN, MANHATAN, DIAGONAL};
    HeuristicDistance(distance_estim_t heuristic_type=EUCLIDEAN,double epsilon=1):
        heuristic(heuristic_type),epsilon(epsilon)
    {}
    double operator() (Eigen::Vector2d v1,Eigen::Vector2d v2){
        double h=0;
        switch (heuristic) {
        case MANHATAN:
            h= manhatan(v1-v2);
            break;
        case DIAGONAL:
            h= diagonal(v1-v2);
            break;
        default: //EUCLIDEAN
            h= (v1-v2).norm();
            break;
        }

        return h*epsilon;
    }
    distance_estim_t heuristic;
    double epsilon;
private:
    double manhatan(Eigen::Vector2d v){
        Eigen::Vector2d x(1,0),y(0,1);
        double vX=v.dot(x);
        double vY=v.dot(y);
        return vX+vY;
    }
    double diagonal(Eigen::Vector2d v){
        Eigen::Vector2d x(1,0),y(0,1);
        double vX=std::fabs(v.dot(x));
        double vY=std::fabs(v.dot(y));
        return (std::sqrt(2)-1)*min(vX,vY)+max(vX,vY); // ( sqrt(2)*min(x,y)+max(x,y)-min(x,y) )
    }
};

std::vector<MultiAgentCell*> MultiAgentGrid::computeOtherTrajectory(MultiAgentCell *origin, MultiAgentCell *dest, unsigned int agent_index,double epsilon,double dist_max,int max_expansions){
    //TODO reset temp
    //if(origin==getAgentStructAt(agent_index)->getOrigin()){
    //    return dest->getTraj(agent_index);
    //}
    unsigned int count(0),skip(10);
    int expansions_count(0);
    bool check_col_with_other_agents=false;//TODO check_col_with_other as argument / parameter
    if(epsilon<1){
        if(_showText){
            cout<<__FUNCTION__ << ": Warn: provided an epsilon < 1, reset to 1"<<endl;
        }
        epsilon=1;
    }
    HeuristicDistance H(HeuristicDistance::DIAGONAL,epsilon);
    vector<MultiAgentCell*> traj_reverse,traj;
    for(unsigned int i=0;i<getAgentStructAt(agent_index)->reachable().size();++i){
        MultiAgentCell * c=getAgentStructAt(agent_index)->reachable()[i];
        c->resetExplorationStatus();
    }
    if(dist_max<0){
        dist_max = numeric_limits<double>::infinity();
    }else{
        dist_max*=epsilon;
    }
    origin->setOpen();
    priority_queue<MultiAgentCell*,vector<MultiAgentCell*>,compare_temporary_estimated_distances_of_cells> openCells;
    openCells.push(origin);
    origin->setTempDist(0); // init
    origin->setTempEstim(H(dest->getCenter(),origin->getCenter()) + origin->getTempDist()); //estimation of total distance (heuristic + current dist)
    origin->setTempCameFrom(0);

    if(!origin->isReachable(agent_index) || !dest->isReachable(agent_index)){
        if(_draw_a_star){
            g3d_draw_allwin_active();
        }
        origin->resetExplorationStatus();
        return traj;
    }
    unsigned int agent_index_to_draw_grid_save=this->_agent_index_to_draw_grid;
    this->_agent_index_to_draw_grid=agent_index;

    while(openCells.size() >0 && expansions_count-max_expansions !=0){
        MultiAgentCell * current = openCells.top(); //la cell avec le cout estime le plus faible

        if(current==dest){
            break; //found
        }
        openCells.pop(); // remove current from open
        current->setClosed();
        expansions_count++;
        if(_draw_a_star && !(count++%skip)){
            g3d_draw_allwin_active();
        }
        vector<MultiAgentCell*> currentNeighbors = current->getNeighbors(agent_index);
        for(unsigned int i=0;i<currentNeighbors.size();++i){
            MultiAgentCell * neighbor=currentNeighbors[i];
            if(neighbor->getClosed()){
                continue;
            }
            //check colisions with other agents
            if(check_col_with_other_agents){
                bool is_in_col=false;
                for (unsigned int i=0;i<this->getAgents().size();i++){
                    if (i==agent_index){
                        continue;
                    }
                    MultiAgentCell *cell_other=dynamic_cast<MultiAgentCell*>(
                                this->getCell(Eigen::Vector2d(this->getAgentStructAt(i)->getRobot()->getCurrentPos()->at(6+0),this->getAgentStructAt(i)->getRobot()->getCurrentPos()->at(6+1)))
                                );
                    if(cell_other->computeDist(neighbor)<5){ //TODO distance threshold to set as parameter
                        if(p3d_col_test_robot_other(getAgentStructAt(i)->getCylinder()->getRobotStruct(),getAgentStructAt(agent_index)->getCylinder()->getRobotStruct(),FALSE)){
                            is_in_col=true;
                            break;
                        }
                    }
                }
                if(is_in_col){
                    neighbor->setClosed();
                    continue;
                }
            }
            double tentative_dist=current->getTempDist()+(current->getCenter()-neighbor->getCenter()).norm();
            if( !neighbor->getOpen() || tentative_dist < neighbor->getTempDist()){
                neighbor->setTempCameFrom(current);
                neighbor->setTempDist(tentative_dist);
                neighbor->setTempEstim( H(neighbor->getCenter(),dest->getCenter()) +  neighbor->getTempDist()); //estimation of total distance (heuristic + current dist)
                //open the cell only if the distance is inferior to max dist
                if(!neighbor->getOpen() && neighbor->getTempEstim() < dist_max && neighbor->getTempDist() < dist_max/epsilon){
                    neighbor->setOpen();
                    openCells.push(neighbor);
                }
            }
        }
    }
    this->_agent_index_to_draw_grid=agent_index_to_draw_grid_save;

    if(!openCells.empty()){
        MultiAgentCell * cell=dest;
        while(cell!=origin){
            traj_reverse.push_back(cell);
            cell=cell->getTempCameFrom();
        }
        traj_reverse.push_back(cell);
        traj.assign(traj_reverse.rbegin(),traj_reverse.rend());
        assert(traj.front()==origin);
    }
    initAllCellState();
    return traj;
}

bool MultiAgentGrid::areCellsInNeighbourhood(MultiAgentCell *c1, MultiAgentCell *c2, unsigned int a,double dist){
    // we use the fact that |d(A,x) - d(A,Y)| < d(X,Y)
    //  d() is shortest distance
    //so if this lower bound is inferior to the max distance (dist) we do check
    //with A*, otherwise, the lower bound is superior to the max dist so the
    //real distance is out of range, not in neighbourhood
    double lower_bound=fabs(c1->getDistance(a)-c2->getDistance(a));
    bool lower_bound_ok= (lower_bound - dist < 0.0001);
    double epsilon = 1;//TODO epsilon for A* in areCellsInNeighbourhood() should be settable
    if( lower_bound_ok ){
        std::vector<MultiAgentCell*> traj=this->computeOtherTrajectory(c1,c2,a,epsilon,dist);
        return !traj.empty();
    }else{
        return false;
    }

    //TODO remove old body of this function, or make other function
    Eigen::Vector2i coord1=c1->getCoord();
    Eigen::Vector2i coord2=c2->getCoord();
    Vector2i diff=coord2-coord1;
    Eigen::Vector2i candidate_coord=coord1;
    Eigen::Vector2d next_coord=candidate_coord.cast<double>();
    Eigen::Vector2d dir=diff.cast<double>();
    dir.normalize();
    MultiAgentCell *c=c1;
    while(c->isReachable(a)&& c!=c2){
        next_coord=next_coord+dir;
        candidate_coord.x()=std::floor(0.5+next_coord.x());
        candidate_coord.y()=std::floor(0.5+next_coord.y());
        c=dynamic_cast<MultiAgentCell*>(getCell(candidate_coord));
    }
    return c==c2;
}

void  MultiAgentGrid::computeDistancesForAgent(MultiAgentCell* cell, unsigned int agent_index)
{
    //check if this agent correspond to an existing meta agent having the same accessible space
    cout<< "computeDistancesForAgent check : "<< getAgents()[agent_index]->getName() << " : "
        << getAgentStructAt(agent_index)->getMetaAgent()->letter_id << " -> ";
    bool found=false;
    if(getAgentStructAt(agent_index)->getMetaAgent()->reachable().size()!=0){
        //accessibility computed for this meta agent
        //are they overlaping?
        MetaAgentInGrid::s_ptr meta_cur = getAgentStructAt(agent_index)->getMetaAgent();
        if(!cell->isReachable(agent_index)){
            //no => different spaces
            //find a meta agent overlaping with this
            for (unsigned int i=0;i<mAgentsInGrid.size();++i){
                if(i!=agent_index && cell->isDistanceComputed(i) && getAgentStructAt(agent_index)->getCylinder() == getAgentStructAt(i)->getCylinder()){
                    //found, change to corresponding meta agent
                    getAgentStructAt(i)->getMetaAgent()->addAgent(getAgentStructAt(agent_index),getAgentStructAt(i)->getMetaAgent());
                    found=true;
                    break;
                }
            }
            if (!found){
                //not found, clone to create a new meta agent
                meta_cur=meta_cur->shared_clone();
                meta_cur->reachable().clear();
                meta_cur->addAgent(getAgentStructAt(agent_index),meta_cur);
                cout << "new : "<< meta_cur->letter_id;
            }else{
                cout << "select existing : "<<getAgentStructAt(agent_index)->getMetaAgent()->letter_id;
                for (vector<MultiAgentCell*>::iterator it=getAgentStructAt(agent_index)->reachable().begin();it!=getAgentStructAt(agent_index)->reachable().end();++it){
                    (*it)->checkMetaAgent(agent_index);
                }
            }
        }else{
            cout <<"ok";
            found=true;
            bool check_st=true;
            for (vector<MultiAgentCell*>::iterator it=getAgentStructAt(agent_index)->reachable().begin();it!=getAgentStructAt(agent_index)->reachable().end();++it){
                check_st=(*it)->checkMetaAgent(agent_index) && check_st;
            }
            cout << " - check (" <<getAgentStructAt(agent_index)->reachable().size()<<  ") = "<<check_st;

        }
    }
    cout <<endl;
    cell->checkMetaAgent(agent_index) ;
    cell->setOpen();
    deque<MultiAgentCell*> openCells;
    openCells.push_back(cell);

    if(!found){
        cell->setAccessible(agent_index,true);
        cell->setReachable(agent_index,true);
        //cell->computeMoveCost(agent_index);
        getAgentStructAt(agent_index)->reachable().push_back(cell);
    }
    cell->setDistance(agent_index,0);
    cell->setDistanceIsComputed(agent_index);
    vector<MultiAgentCell*> tmp_traj(1,cell);
    cell->setTraj(agent_index,tmp_traj);

    while (openCells.size() > 0)
    {
        MultiAgentCell* currentCell = openCells.front(); //TODO check why front and not back (for efficiency)?
        vector<MultiAgentCell*> currentNeighbors = currentCell->getNeighbors(agent_index);
        if(!found && currentCell->isBoundary(agent_index)){
            if(std::find(getAgentStructAt(agent_index)->boundaries().begin(),getAgentStructAt(agent_index)->boundaries().end(),currentCell) != getAgentStructAt(agent_index)->boundaries().end()){
                //cerr<<"BOUNDARY ALREADY ADDED"<<endl;
            }
            getAgentStructAt(agent_index)->boundaries().push_back(currentCell);
        }

        double currentDist = 0;

        if (currentCell->isDistanceComputed(agent_index))
        {
            currentDist = currentCell->getDistance(agent_index);
        }
        else
        {
            currentCell->setDistanceIsComputed(agent_index);
            if(!found){
                currentCell->setReachable(agent_index,true);
                currentCell->setAccessible(agent_index,true);
                getAgentStructAt(agent_index)->reachable().push_back(currentCell);
            }
        }

        for (unsigned int i = 0; i < currentNeighbors.size(); i++)
        {
            MultiAgentCell* neighCell = currentNeighbors[i];
            if (!neighCell->getClosed())
            {
                double dist = currentDist + currentCell->computeDist(neighCell);
                if (neighCell->getOpen())
                {
                    if (dist > neighCell->getDistance(agent_index))
                    {
                        dist = neighCell->getDistance(agent_index);
                    }
                    else
                    {
                        neighCell->setTraj(agent_index,currentCell->getTraj(agent_index));
                        neighCell->getTraj(agent_index).push_back(neighCell);

                        neighCell->setVectorTraj(agent_index,currentCell->getVectorTraj(agent_index));
                        neighCell->getVectorTraj(agent_index).push_back(neighCell->getCenter());
                    }
                }
                else
                {
                    //neighCell is new, and is accessible \o/
                    //add this new path to reach it
                    neighCell->setTraj(agent_index,currentCell->getTraj(agent_index));
                    neighCell->getTraj(agent_index).push_back(neighCell);

                    neighCell->setVectorTraj(agent_index,currentCell->getVectorTraj(agent_index));
                    neighCell->getVectorTraj(agent_index).push_back(neighCell->getCenter());

                    neighCell->setOpen();
                    neighCell->checkMetaAgent(agent_index) ;
                    //add it to cells to explore
                    openCells.push_back(neighCell);

                    //add it to accessible cells
                    if(!found){
                        getAgentStructAt(agent_index)->reachable().push_back(neighCell);
                        neighCell->setAccessible(agent_index,true);
                        neighCell->setReachable(agent_index,true);
                    }
                }
                neighCell->setDistance(agent_index, dist);
                neighCell->setDistanceIsComputed(agent_index);
                if (dist > getAgentStructAt(agent_index)->getMaxDist()) {getAgentStructAt(agent_index)->setMaxDist(dist); }
                //neighCell->computeMoveCost(agent_index);
            }

        }
        currentCell->setClosed();
        openCells.pop_front();//delete first element
    }
    getAgentStructAt(agent_index)->setDistanceComputed(true);
    initAllCellState();

}

void  MultiAgentGrid::computeAltDistancesForAgent(MultiAgentCell* cell, unsigned int agent_index)
{
    if(!cell || cell->getAltDistance(agent_index,cell)==0){
        return; //already computed, good! / or no cell
    }
    cell->setOpen();
    deque<MultiAgentCell*> openCells;
    openCells.push_back(cell);

    cell->setAltDistance(agent_index,0,cell);

    while (openCells.size() > 0)
    {
        MultiAgentCell* currentCell = openCells.front(); //TODO check why front and not back (for efficiency)?
        vector<MultiAgentCell*> currentNeighbors = currentCell->getNeighbors(agent_index);

        double currentDist = 0;

        currentDist = currentCell->getAltDistance(agent_index,cell);

        for (unsigned int i = 0; i < currentNeighbors.size(); i++)
        {
            MultiAgentCell* neighCell = currentNeighbors[i];
            if (!neighCell->getClosed())
            {
                double dist = currentDist + currentCell->computeDist(neighCell);
                if (neighCell->getOpen())
                {
                    if (dist < neighCell->getAltDistance(agent_index,cell))
                    {
                        neighCell->setAltDistance(agent_index,dist,cell,currentCell);
                    }
                }
                else
                {
                    //neighCell is new, and is accessible \o/
                    neighCell->setAltDistance(agent_index,dist,cell,currentCell);

                    neighCell->setOpen();
                    //add it to cells to explore
                    openCells.push_back(neighCell);
                }
                //TODO this was commented on 3 Jul. 2014 14:00, is it OK? : (1 line, setAltDistance)
                //neighCell->setAltDistance(agent_index, dist,cell);
            }

        }
        currentCell->setClosed();
        openCells.pop_front();//delete first element
    }
    initAllCellState();
}

//vector<Vector2d> MultiAgentCell::getTempVectorTraj(){
//    vector<Vector2d> traj;
//    MultiAgentCell *cell=this;
//    while(cell!=0){
//        traj.push_back(cell->getCenter());
//        cell=cell->getTempCameFrom();
//    }
//}

pair<MultiAgentCell*,MultiAgentCell*> MultiAgentGrid::neighborSolution(unsigned int agent_1,unsigned int agent_2,pair<MultiAgentCell*,MultiAgentCell*> solution){
    pair<MultiAgentCell*,MultiAgentCell*> next;
    int r = rand()%2;
    bool _use_real_neighbours=false; //TODO make this an attribute
    next=solution;
    if(_use_real_neighbours){
    if(r==0){
        //change first point
        vector<MultiAgentCell*> neigh=solution.first->getNeighbors(agent_1);
        int r_i=rand()%neigh.size();
        next.first=neigh[r_i];
    }else{
        //change second point
        vector<MultiAgentCell*> neigh=solution.second->getNeighbors(agent_2);
        int r_i=rand()%neigh.size();
        next.second=neigh[r_i];
    }
    }else{
        MultiAgentCell * cell=(r==0 ? solution.first : solution.second),*new_cell=0;
        unsigned int agent=(r==0 ? agent_1 : agent_2);
        Vector2i r_pos;
        do{
            r_pos=cell->getCoord();
            r_pos[0]+=(rand()%7)-3;
            r_pos[1]+=(rand()%7)-3;
            if(r_pos[0]>0 && r_pos[0]<static_cast<int>(this->_nbCellsX) && r_pos[1]>=0 && r_pos[1]<static_cast<int>(this->_nbCellsY)){
                new_cell=dynamic_cast<MultiAgentCell*>(this->getCell(r_pos));
            }
        }while(!new_cell || !new_cell->isDistanceComputed(agent));
        if(r==0){
            next.first=new_cell;
        }else{
            next.second=new_cell;
        }
    }
    return next;
}


std::vector<FrontierPtr> MultiAgentGrid::searchFrontiers(unsigned int agent_1,unsigned int agent_2,double keep_distance){
    set<shared_ptr<FrontierCells> > frontiers;
    vector<shared_ptr<FrontierCells> > frontiers2, ret;
    unsigned int k_max=0;
    exec_time();
    for(unsigned int k=0;k<k_max;++k){
        //frontiers2=this->hillClimberFrontierSearch(agent_1,agent_2,keep_distance);
        frontiers.insert(frontiers2.begin(),frontiers2.end());
    }
    //if no frontier found, use simulated annealing instead
    k_max=0;
    if(frontiers.size()==0){
        for(unsigned int k=0;k<k_max;++k){
            frontiers2=this->simulatedAnnealingFrontierSearch(agent_1,agent_2,keep_distance);
            frontiers.insert(frontiers2.begin(),frontiers2.end());
        }
    }
    ret.insert(ret.end(),frontiers.begin(),frontiers.end());
    //ret=handInHandFrontierSearch(agent_1,agent_2,keep_distance);
    ret=exhaustiveFrontierSearch(agent_1,agent_2,keep_distance);
    exec_time(&_computation_stat.frontiers_time[getFrontierIndex(agent_2,agent_1)]);
    _computation_stat.frontiers_total_time.tv_nsec+=_computation_stat.frontiers_time[getFrontierIndex(agent_2,agent_1)].tv_nsec;
    _computation_stat.frontiers_total_time.tv_sec+=_computation_stat.frontiers_time[getFrontierIndex(agent_2,agent_1)].tv_sec;
    return ret;
}

vector<FrontierPtr> MultiAgentGrid::frontierSearch(unsigned int agent1, unsigned int agent2, double keep_distance){
    return exhaustiveFrontierSearch(agent1,agent2,keep_distance);
}

vector<MetaFrontierPtr> MultiAgentGrid::hillClimberFrontierSearch(unsigned int agent_1,unsigned int agent_2,double keep_distance){
    AgentInGrid *a1=getAgentStructAt(agent_1);
    AgentInGrid *a2=getAgentStructAt(agent_2);
    vector<shared_ptr<MetaFrontierCells> > frontier_points;
    // a solution is a set of two points, one for each agent position,.
    pair<MultiAgentCell*,MultiAgentCell*> solution,next;
    double next_eval,best_eval;
    unsigned int stop_counter(0),stop_counter_thresh(100);//stops after stop_counter_thresh
                                                          //consecutive tries that don't improve the cirteria
    if(_showText){
        cout<<" MultiAgentGrid::hillClimberFrontierSearch() : start for "<< mAgents[agent_1]->getName() << " vs " << mAgents[agent_2]->getName()<<endl;
    }

    if(!a1->isDistanceComputed() || ! a2->isDistanceComputed()){
        cerr << " MultiAgentGrid::hillClimberFrontierSearch() : ERROR distances not computed for agents, abort"<<endl;
        cerr << "                                               computeDistances() must be called before."<<endl;
        return frontier_points;
    }

    do{
        //get a random reachable point for agent1
        if(_use_reachable_for_frontier_search){
            int r=rand()%a1->reachable().size();
            solution.first=a1->reachable()[r];
        }else{
            int r=rand()%a1->boundaries().size();
            solution.first=a1->boundaries()[r];
        }
    }while(!solution.first->isDistanceComputed(agent_1));
    do{
        //get a random reachable point for agent2
        if(_use_reachable_for_frontier_search){
            int r=rand()%a2->reachable().size();
            solution.second=a2->reachable()[r];
        }else{
            int r=rand()%a2->boundaries().size();
            solution.second=a2->boundaries()[r];
        }
    }while(!solution.second->isDistanceComputed(agent_2));

    best_eval = solution.first->computeDist(solution.second);
    next=pair<MultiAgentCell*,MultiAgentCell*>(0,0); //NULL
    next_eval=numeric_limits<double>::infinity();
    do{
        //find next solution, changing one Cell in solution for a neighbour of it
        //select the best pair
        vector<MultiAgentCell*> tmp_neighbours=solution.first->getNeighbors(agent_1);
        vector<MultiAgentCell*> tmp_neighbours2=solution.second->getNeighbors(agent_2);
        for(unsigned int i=0;i<tmp_neighbours.size();++i){
            if(_use_reachable_for_frontier_search || tmp_neighbours[i]->isBoundary(agent_1)){
                double dist=tmp_neighbours[i]->computeDist(solution.second);
                if(dist<next_eval){
                    next_eval=dist;
                    next=pair<MultiAgentCell*,MultiAgentCell*>(tmp_neighbours[i],solution.second);
                }
            }
        }
        for(unsigned int i=0;i<tmp_neighbours2.size();++i){
            if(_use_reachable_for_frontier_search || tmp_neighbours2[i]->isBoundary(agent_2)){
                double dist=tmp_neighbours2[i]->computeDist(solution.first);
                if(dist<next_eval){
                    next_eval=dist;
                    next=pair<MultiAgentCell*,MultiAgentCell*>(solution.first,tmp_neighbours2[i]);
                }
            }
        }
        if (next_eval<keep_distance){
            //if close enough to be a frontier pair, store it
            frontier_points.push_back(next.first->addMetaFrontier(next.second,true,agent_1,agent_2,next_eval));
            //addFrontier(agent_1,agent_2,next.first);
            //addFrontier(agent_2,agent_1,next.second);
        }
        if(next_eval<best_eval){
            best_eval=next_eval;
            solution=next;
        }
        if(next_eval<best_eval){
            stop_counter=0;
        }else{
            ++stop_counter;
        }
    }while(stop_counter<stop_counter_thresh && best_eval>_frontier_search_stop_dist);

    if(_showText){
        cout<<" MultiAgentGrid::hillClimberFrontierSearch() : end with "<< frontier_points.size() << "frontiers pair found, stop with distance : "<< best_eval<<endl;
    }

    return frontier_points;
}

struct compareMetaFrontierWithCells{
    compareMetaFrontierWithCells(MultiAgentGrid::MetaAgentInGrid::s_ptr a1, MultiAgentGrid::MetaAgentInGrid::s_ptr a2, MultiAgentCell *c1,MultiAgentCell* c2){
        setCells(a1,c1,a2,c2);
    }
    bool operator()(MetaFrontierPtr it){
        return ((it)->cellOf(_a1)==_p.first  && (it)->cellOf(_a2)==_p.second) || ((it)->cellOf(_a2)==_p.first  && (it)->cellOf(_a1)==_p.second);
    }
    void setCells(MultiAgentGrid::MetaAgentInGrid::s_ptr a1,MultiAgentCell*
    c1,MultiAgentGrid::MetaAgentInGrid::s_ptr a2,MultiAgentCell* c2){
    setCells(a1,a2,pair<MultiAgentCell*,MultiAgentCell*>(c1,c2));} void
    setCells(MultiAgentGrid::MetaAgentInGrid::s_ptr a1,
    MultiAgentGrid::MetaAgentInGrid::s_ptr a2,
    pair<MultiAgentCell*,MultiAgentCell*> p){_a1=a1;_a2=a2;_p=p;} private:
    pair<MultiAgentCell*,MultiAgentCell*> _p;
    MultiAgentGrid::MetaAgentInGrid::s_ptr _a1,_a2;
};

std::vector<FrontierPtr > MultiAgentGrid::handInHandFrontierSearch(unsigned int agent1,unsigned int agent2, double keep_distance){
    unsigned int count=0;
    //getAgentStructAt(agent2)->frontiers(getAgentStructAt(agent1)).clear();
    //getAgentStructAt(agent1)->frontiers(getAgentStructAt(agent2)).clear();
    getAgentStructAt(agent1)->setFrontierFullyChecked(getAgentStructAt(agent2),true);
    getAgentStructAt(agent2)->setFrontierFullyChecked(getAgentStructAt(agent1),true);
    bool finished=false;
    double prev_dist=0;
    initAllCellState(); //reset open/closed markers
    AgentInGrid *a1=getAgentStructAt(agent1);
    AgentInGrid *a2=getAgentStructAt(agent2);
    vector<shared_ptr<FrontierCells> > frontier_points;
    // a solution is a set of two points, one for each agent position,.
    pair<MultiAgentCell*,MultiAgentCell*> solution,next,previous;
    unsigned int stop_counter(0),stop_counter_thresh(100);//stops after stop_counter_thresh
                                                          //consecutive tries that don't improve the cirteria
    if(_showText){
        cout<<" MultiAgentGrid::handInHandFrontierSearch() : start for "<< a1->getRobot()->getName() << " vs " << a2->getRobot()->getName()<<endl;
    }

    if(!a1->isDistanceComputed() || ! a2->isDistanceComputed()){
        cerr << " MultiAgentGrid::handInHandFrontierSearch() : ERROR distances not computed for agents, abort"<<endl;
        cerr << "                                               computeDistances() must be called before."<<endl;
        return frontier_points;
    }
    double frontier_search_stop_dist_bkp=_frontier_search_stop_dist;
    _frontier_search_stop_dist=keep_distance;//make the hill climber stop when one solution is found
    vector<MetaFrontierPtr> initial_frontiers = hillClimberFrontierSearch(agent1,agent2,keep_distance);
    _frontier_search_stop_dist=frontier_search_stop_dist_bkp;
    //get the cell pair from the hill climber
    if(initial_frontiers.size()>0){
        solution.first=initial_frontiers[0]->cellOf(a1->getMetaAgent());
        solution.second=initial_frontiers[0]->cellOf(a2->getMetaAgent());
    }else{
        return std::vector<std::tr1::shared_ptr<FrontierCells> > (0);
    }
    while (!finished && stop_counter<stop_counter_thresh){
        double next_dist=numeric_limits<double>::infinity();
        //find next solution, changing both Cell in solution for a neighbour of them
        //select the best pair
        vector<MultiAgentCell*> tmp_neighbours=solution.first->getNeighbors(agent1);
        tmp_neighbours.push_back(solution.first);
        vector<MultiAgentCell*> tmp_neighbours2=solution.second->getNeighbors(agent2);
        tmp_neighbours2.push_back(solution.second);
        double r01=static_cast<double>(rand())/RAND_MAX;
        if(1 || r01<exp((prev_dist-keep_distance*2)*4)){
            vector<MultiAgentCell*> neighInRange1,neighInRange2;
            neighInRange2=solution.first->findNeighborsInRange(agent2,keep_distance);
            neighInRange1=solution.second->findNeighborsInRange(agent1,keep_distance);
            tmp_neighbours.insert(tmp_neighbours.end(),neighInRange1.begin(),neighInRange1.end());
            tmp_neighbours2.insert(tmp_neighbours2.end(),neighInRange2.begin(),neighInRange2.end());
            //cout<<"/  \\  /  \\  /  \\  tunnel "<<r01<< " < " << exp(prev_dist-keep_distance) << " prev: "<<prev_dist <<endl;
        }
        //cout<<'.';
        for(unsigned int i=0;i<tmp_neighbours.size();++i){
            for(unsigned int j=0;j<tmp_neighbours2.size();++j){
                if(solution.first->isReachable(agent2) || solution.second->isReachable(agent1)){
                    assert(solution.first->isDistanceComputed(agent2) || solution.second->isDistanceComputed(agent1));
                    //accessible spaces overlap, agents can have lots of tf positions possible
                    //we do not compute them all, random search will be used online
                    getAgentStructAt(agent1)->setFrontierFullyChecked(getAgentStructAt(agent2),false);
                    getAgentStructAt(agent2)->setFrontierFullyChecked(getAgentStructAt(agent1),false);
                    finished=true;
                    break;
                }else if( !tmp_neighbours[i]->getClosed() &&
                          !tmp_neighbours2[j]->getClosed() &&
                          tmp_neighbours[i]->isBoundary(agent1)
                          && tmp_neighbours2[j]->isBoundary(agent2)){
                    //vector<MultiAgentCell*>::reverse_iterator r_it = find(a1->frontiers(getAgentStructAt(agent2)).rbegin(),a1->frontiers(getAgentStructAt(agent2)).rend(),tmp_neighbours[i]);
                    MultiAgentCell** r_it=&tmp_neighbours[i];
                    compareMetaFrontierWithCells comp_with_cells(a1->getMetaAgent(),a2->getMetaAgent(),tmp_neighbours[i],tmp_neighbours2[j]);
                    vector<MetaFrontierPtr>::reverse_iterator f_it;
                    //if(r_it!=a1->frontiers(getAgentStructAt(agent2)).rend()){
                        f_it=find_if((*r_it)->getMetaFrontier(agent1,agent2).rbegin(), (*r_it)->getMetaFrontier(agent1,agent2).rend(),comp_with_cells);
                        if( f_it != (*r_it)->getMetaFrontier(agent1,agent2).rend()){
                            //if this frontier already exists
                            assert((*f_it)->cellOf(a1->getMetaAgent()) == tmp_neighbours[i] && (*f_it)->cellOf(a2->getMetaAgent()) == tmp_neighbours2[j]);
                            continue;
                        }
                        for(unsigned int k=0;k<(*r_it)->getMetaFrontier(agent1,agent2).size();k++){
                            MetaFrontierPtr p = (*r_it)->getMetaFrontier(agent1,agent2)[k];
                            assert(!(p->cellOf(a1->getMetaAgent())==tmp_neighbours[i] && p->cellOf(a2->getMetaAgent())==tmp_neighbours2[j]));
                        }
                    //}
                    double dist=tmp_neighbours[i]->computeDist(tmp_neighbours2[j]);
                    if(dist<next_dist){
                        next_dist=dist;
                        next=pair<MultiAgentCell*,MultiAgentCell*>(tmp_neighbours[i],tmp_neighbours2[j]);
                    }
                }
            }
            if( finished ){
                break;
            }
        }
        if(next_dist<numeric_limits<double>::infinity()){
            //a next have been found
            //if(solution.first != next.first){
                solution.first->setClosed();
            //}
            //if(solution.second != next.second){
                solution.second->setClosed();
            //}
            solution=next;
            if(next_dist<keep_distance){
                count++;
                abort();
                //solution.first->addMetaFrontier(solution.second,agent1,agent2,next_dist);
                addFrontier(agent1,agent2,solution.first);
                addFrontier(agent2,agent1,solution.second);
                stop_counter=0;
            }else{
                stop_counter++;
            }
            prev_dist=next_dist;
        }else{
            finished=true;
        }
#ifdef DRAW_FRONTIER_HAND_IN_HAND
        _draw_frontier_search=true;
        _draw_frontier_search_cells=solution;
        _draw_frontier_search_is_pair_ok=(stop_counter==0);
        g3d_draw_allwin_active();
        if(_draw_frontier_search_is_pair_ok){
            usleep(10e3);
        }
        _draw_frontier_search=false;
#endif
    }
    if(_showText){
        if(!finished){
            cout << "                                               stop : found " <<getFrontierCells(agent1,agent2).size()<<" frontier. ("<<count<<')'<<endl;
        }else{
            cout << "                                               finished : found " <<getFrontierCells(agent1,agent2).size()<<" frontier. ("<<count<<')'<<endl;
        }
    }
    return std::vector<std::tr1::shared_ptr<FrontierCells> > (0);
}

void MultiAgentGrid::saveFrontiers(ofstream &output){
    std::map<MetaFrontierPtr,set<MetaAgentInGrid::s_ptr> > frontiers;
    //header/disclaimer
    output << "# file generated by the Multi-HandOver module of Move3d-planners\n"
           << "# Contains frontier cells listing for all agents in an environement\n"
           << "# that have been found when the file have been generated.\n"
           << "# It's your responsibility to check it is comform with the environement\n"
           << "# you are using."<<endl;

    output << "# grid size"<<endl;
    output << this->_nbCellsX<<','<<this->_nbCellsY<<endl;
    output << "# grid resolution"<<endl;
    output << this->_cellSize.x() << ',' << this->_cellSize.y() <<endl;

    output << "# frontiers"<<endl;
    for(unsigned int first=0;first<this->mAgents.size();++first){
        for(unsigned int second=0;second<this->mAgents.size();++second){
            vector<MultiAgentCell*> const & cells=getFrontierCells(first,second);
            for (unsigned int i=0;i<cells.size();++i){
                if(cells[i]->getMetaFrontier(first,second).size()){
                    MetaFrontierPtr mf=cells[i]->getMetaFrontier(first,second)[0];
                    frontiers.insert(make_pair(mf,set<MetaAgentInGrid::s_ptr>()));
                    //frontiers[mf].insert(this->getAgentStructAt(first)->getMetaAgent());
                    //frfrontiers[mf].insert(this->getAgentStructAt(second)->getMetaAgent());
                }
            }
        }
    }
    for(map<MetaFrontierPtr,set<MetaAgentInGrid::s_ptr> >::iterator it=frontiers.begin();
        it!=frontiers.end();it++)
    {
        MultiAgentCell *c=it->first->getCellA();
        output << c->getCoord().x() << ',' << c->getCoord().y() << ';';
        c=it->first->getCellB();
        output << c->getCoord().x() << ',' << c->getCoord().y() << ':';
    //     for (set<MetaAgentInGrid::s_ptr>::iterator ag=it->second.begin();
    //          ag!=it->second.end(); ag++)
    //     {
    //         vector<AgentInGrid*> & agents=(*ag)->getAgentsInGrid();
    //         for(unsigned int a=0;a<agents.size();a++){
    //             //output<<agents[a]
    //         }
    //     }
        output<<endl;
    }
    output.flush();
}

bool MultiAgentGrid::loadFrontiers(ifstream &input){
    bool ok=true;
    unsigned int n=20;
    int xm,ym,x,y;
    double rx,ry;
    int state=0;
    vector<pair<MultiAgentCell*,MultiAgentCell*> > cells;
    //int line_count=0;
    while(ok && !input.eof()){
        //cout<<line_count++<<" - ";
        //cout.flush();
        //char line[n];
        char c=0;
        istringstream s;
        string str;
        getline(input,str);
        s.str(str);
        bool end_of_line=false;
        unsigned int anti_loop=0;
        while(ok && !end_of_line && anti_loop<n && s.good()){
            anti_loop++;
            s >> std::ws;
            s >> c;
            if (isdigit(c)){
                int i;
                double v;
                s.putback(c);
                s >> v;
                i=v;
                switch (state) {
                case 0:
                    xm=i;
                    break;
                case 1:
                    ym=i;
                    break;
                case 2:
                    rx=v;
                    break;
                case 3:
                    ry=v;
                    if ( xm<0 || ym<0
                         ||
                         !( static_cast<unsigned int>(xm)==this->_nbCellsX && static_cast<unsigned int>(ym)==this->_nbCellsY
                            && rx==this->getCellSize().x() && ry==this->getCellSize().y() ) )
                    {
                        //the grid is different
                        ok=false;
                    }
                    break;
                case 4:
                    //x of cellA
                    x=i;
                    break;
                case 5:
                    //y of cellA
                    y=i;
                    cells.push_back(make_pair(dynamic_cast<MultiAgentCell*>(this->getCell(x,y)),(MultiAgentCell*)0));
                    break;
                case 6:
                    //x of cellB
                    x=i;
                    break;
                case 7:
                    //y of cellB
                    y=i;
                    cells.back().second=dynamic_cast<MultiAgentCell*>(this->getCell(x,y));
                    break;
                default:
                    break;
                }
                state++;
                if (state>=8){
                    end_of_line=true;
                    state=4;
                }
            }else if(c=='#' || c=='\n' || c=='\0'){
                //comment or end of line, stop
                end_of_line=true;//break
            }//in other cases, consume the char (can be , or ; ...)
        }
    }
    //cout<<endl;
    if(ok){
        for (unsigned int i=0;i<cells.size();i++){
            pair<MultiAgentCell*,MultiAgentCell*> p = cells[i];
            for (unsigned int a1=0;a1<this->getAgents().size();a1++){
                if(!p.first->isReachable(a1) || !p.first->isBoundary(a1)){
                    continue;
                }else{
                    for (unsigned int a2=0;a2<this->getAgents().size();a2++){
                        double keep_distance=this->getArmReach(a1)+this->getArmReach(a2);
                        double dist=p.first->computeDist(p.second);
                        if (p.second->isReachable(a2) && p.second->isBoundary(a2) && !areCellsInNeighbourhood(p.first,p.second,a2,dist*sqrt(2))){
                            if(dist <=keep_distance){
                                p.first->addMetaFrontier(p.second,true,a1,a2,dist);
                                addFrontier(a1,a2,p.first);
                                addFrontier(a2,a1,p.second);
                            }
                        }
                    }
                }
            }
        }
    }
    return ok;
}

std::vector<FrontierPtr> MultiAgentGrid::exhaustiveFrontierSearch(unsigned int agent_1,unsigned int agent_2,double keep_distance){
    //getAgentStructAt(agent_2)->frontiers(getAgentStructAt(agent_1)).clear();
    //getAgentStructAt(agent_1)->frontiers(getAgentStructAt(agent_2)).clear();
    //swap(agent_1,agent_2);
    //vector<shared_ptr<FrontierCells> > frontier_points;
    if(!getAgentStructAt(agent_1)->getMetaAgent()->isFrontierComputed(getAgentStructAt(agent_2)->getMetaAgent())){
        getAgentStructAt(agent_1)->getMetaAgent()->setFrontierComputed(getAgentStructAt(agent_2)->getMetaAgent());
        getAgentStructAt(agent_2)->getMetaAgent()->setFrontierComputed(getAgentStructAt(agent_1)->getMetaAgent());

        getAgentStructAt(agent_1)->setFrontierFullyChecked(getAgentStructAt(agent_2),true);
        getAgentStructAt(agent_2)->setFrontierFullyChecked(getAgentStructAt(agent_1),true);
        for (unsigned int i=0;i<getAgentStructAt(agent_1)->boundaries().size();++i){
            pair<MultiAgentCell*,MultiAgentCell*> p(getAgentStructAt(agent_1)->boundaries()[i],0),best=p;
#ifdef DRAW_FRONTIER_STEPS
            //draw the selected cell
            _draw_frontier_steps=true;
            _draw_frontier_steps_cells.assign(1,p.first);
            g3d_draw_allwin_active();
#endif
            double best_dist=keep_distance*2;
            vector<MultiAgentCell*> neighInRange2=p.first->findNeighborsInRange(agent_2,keep_distance);
#ifdef DRAW_FRONTIER_STEPS
            //draw the cells in the reach range
            _draw_frontier_steps=true;
            _draw_frontier_steps_cells=neighInRange2;
            g3d_draw_allwin_active();
#endif
            for (unsigned int j=0;j<neighInRange2.size();++j){
                p.second=neighInRange2[j];
                double dist=p.first->computeDist(p.second);
                if ( p.second->isBoundary(agent_2 ) && !areCellsInNeighbourhood(p.first,p.second,agent_2,dist*sqrt(2))){
                //if ( p.second->isBoundary(agent_2 ) && !areCellsInNeighbourhood(p.first,p.second,agent_2,dist*sqrt(2)) && !areCellsInNeighbourhood(p.second,p.first,agent_1,dist*sqrt(2))){
                    //TODO parameter to select exhaustive frontier check in all cases or not
                    if(false && (p.first->isReachable(agent_2) || p.second->isReachable(agent_1))){
                        assert(p.first->isDistanceComputed(agent_2) || p.second->isDistanceComputed(agent_1));
                        //accessible spaces overlap, agents can have lots of tf positions possible
                        //we do not compute them all, random search will be used online
                        getAgentStructAt(agent_1)->setFrontierFullyChecked(getAgentStructAt(agent_2),false);
                        getAgentStructAt(agent_2)->setFrontierFullyChecked(getAgentStructAt(agent_1),false);
                        break;
                    }else {
                        if(dist<=best_dist && dist <=keep_distance){
                            best=p;
                            best_dist=dist;
                        }
                    }
                }
            }
            if(best_dist<=keep_distance){
                best.first->addMetaFrontier(best.second,true,agent_1,agent_2,best_dist);
                addFrontier(agent_1,agent_2,best.first);
                addFrontier(agent_2,agent_1,best.second);
            }
            if( !getAgentStructAt(agent_1)->frontierFullyChecked(getAgentStructAt(agent_2)) && !getAgentStructAt(agent_2)->frontierFullyChecked(getAgentStructAt(agent_1))){
                break;
            }
            //add the cell to the frontiers list of agent_1 if necessary
            //if(p.first->isFrontier(agent_1,agent_2)){
            //    addFrontier(agent_1,agent_2,getAgentStructAt(agent_1)->boundaries[i]);
            //}
#ifdef DRAW_FRONTIER_STEPS
            _draw_frontier_steps=false;
#endif
#ifdef DRAW_FRONTIER_HAND_IN_HAND
            _draw_frontier_search=true;
            _draw_frontier_search_cells=best;
            _draw_frontier_search_is_pair_ok=(best_dist<=keep_distance);
            g3d_draw_allwin_active();
            if(_draw_frontier_search_is_pair_ok){
                usleep(10e3);
            }
            _draw_frontier_search=false;
#endif
        }
        //add the cells to the frontiers list of agent_2 if necessary
        //for (unsigned int j=0;j<getAgentStructAt(agent_2)->boundaries.size();++j){
        //    if(getAgentStructAt(agent_2)->boundaries[j]->isFrontier(agent_2,agent_1)){
        //        addFrontier(agent_2,agent_1,getAgentStructAt(agent_2)->boundaries[j]);
        //    }
        //}
    }
    return std::vector<std::tr1::shared_ptr<FrontierCells> > (0);
}

vector<shared_ptr<FrontierCells> > MultiAgentGrid::simulatedAnnealingFrontierSearch(unsigned int agent_1,unsigned int agent_2,double keep_distance){
    AgentInGrid *a1=getAgentStructAt(agent_1);
    AgentInGrid *a2=getAgentStructAt(agent_2);
    vector<shared_ptr<FrontierCells> > frontier_points;
    // a solution is a set of two points, one for each agent position,.
    pair<MultiAgentCell*,MultiAgentCell*> solution;
    double energy;
    double energy_max=this->_frontier_search_stop_dist;
    double energy_ok=keep_distance;
    double temperature=this->_frontier_search_start_temp;
    unsigned int k;//iteration counter
    unsigned int k_max=_frontier_search_max_it;

    if(!a1->isDistanceComputed() || ! a2->isDistanceComputed()){
        cerr << " MultiAgentGrid::simulatedAnnealingFrontierSearch() : ERROR distances not computed for agents, abort"<<endl;
        cerr << "                                                      computeDistances() must be called before."<<endl;
        return frontier_points;
    }
    if(_showText){
        cout << "MultiAgentGrid::simulatedAnnealingFrontierSearch() : start for "<<a1->getRobot()->getName()<< " vs " << a2->getRobot()->getName()<<endl;
    }
    if(_save_log_frontier){
        _log_frontier_seeking[getFrontierIndex(agent_1,agent_2)].clear();
    }
    //INIT
    //initialize to random positions.
    do{
        //get a random reachable point for agent1
        int r=rand()%a1->reachable().size();
        solution.first=a1->reachable()[r];
    }while(!solution.first->isDistanceComputed(agent_1));
    do{
        //get a random reachable point for agent2
        int r=rand()%a2->reachable().size();
        solution.second=a2->reachable()[r];
    }while(!solution.second->isDistanceComputed(agent_2));

    energy=solution.first->computeDist(solution.second);

    pair<MultiAgentCell*,MultiAgentCell*> solution_k;
    double energy_k;
    unsigned int log_this=0;
    unsigned int log_this_inc=k_max/500;
    k=0;
    while(k<k_max && energy>energy_max){
        temperature = temperature*_frontier_search_temp_decrease;
        solution_k=neighborSolution(agent_1,agent_2,solution);
        energy_k=solution_k.first->computeDist(solution_k.second);
        if(_save_log_frontier && log_this==k){
            _log_frontier_seeking[getFrontierIndex(agent_1,agent_2)].push_back(pair<double,pair<MultiAgentCell*,MultiAgentCell*> >(energy_k,solution_k));
            log_this+=log_this_inc;
        }
        if((energy_k<energy) || ((rand()/RAND_MAX) < exp((energy-energy_k)/temperature) ) ){
            //accept new solution if energy is lower
            //or with probability exp(-dE/T)
            solution=solution_k;
            energy=energy_k;
            if(energy<=energy_ok){
                // is deprecated need to detect if over obstacle:
                // frontier_points.push_back(solution.first->addFrontierComplete(solution.second,agent_1,agent_2,energy));
                addFrontier(agent_1,agent_2,solution.first);
                addFrontier(agent_2,agent_1,solution.second);
            }
        }
        ++k;
    }
    if(_showText){
        cout<< "MultiAgentGrid::simulatedAnnealingFrontierSearch() : computing frontiers done for "<< a1->getRobot()->getName()<< " vs " << a2->getRobot()->getName()<<endl;
    }
    return frontier_points;
}

void MultiAgentGrid::initVectorFrontiers(){
    //the vector represents a symetrical matrix with consant diagonal (value stored at index 0)
    //a matrix of 4x4 would be represented in the vector as:
    // { diagonal value, 0x1, 0x2, 0x3, 1x2, 1x3, 2x3 }
    unsigned int n=mAgentsInGrid.size()-1;
    //this->_frontiers.assign(getFrontierIndex(n-1,n)+1,vector<shared_ptr<FrontierCells> >(0));
    if(_save_log_frontier){
        this->_log_frontier_seeking.assign(getFrontierIndex(n-1,n)+1,vector<pair<double,pair<MultiAgentCell*,MultiAgentCell*> > >(0));
    }
}

unsigned int MultiAgentGrid::getFrontierIndex(unsigned int i,unsigned int j){
    if(i>j){swap(i,j);}
    else if(i==j){return 0;}

    unsigned int n=mAgentsInGrid.size();
    unsigned int d=i*n - static_cast<double>(i*(i+1))/2.0;//+1 should be added, as at 0 is the diagonal value
    return d+j-i;		    //-1 should be added
}

pair<unsigned int,unsigned int> MultiAgentGrid::getFrontierAgentIndices(unsigned int p){
    unsigned int n=mAgentsInGrid.size();
    double i,j,i1;//les inconnues
    double delta=-1;
    pair<unsigned int, unsigned int> ret(0,0);
    if(p==0){return ret;}

    unsigned int t=1;
    while(delta<0){
        delta=(1.5-n)*(1.5-n)-2*(p-t++);//post incrementation of t
    }
    i1=(n-1.5-sqrt(delta));

    for (int ind=i1;ind>=0;--ind){
        i=ind;
        j=i*i/2+i*(1.5-n)+p;
        if(i<j && j<n && getFrontierIndex(i,j)==p){
            ret.first=i;ret.second=j;
            break;
        }
    }
    return ret;
}

vector<MultiAgentCell*> const & MultiAgentGrid::getFrontierCells(unsigned int i, unsigned int j){
    return this->getAgentStructAt(i)->frontiers(getAgentStructAt(j));
}

void MultiAgentGrid::searchAllFrontiers(){
    //this->initVectorFrontiers();
    _computation_stat.reset_frontiers(getFrontierIndex(mAgentsInGrid.size()-2,mAgentsInGrid.size()-1)+1);
    for(unsigned int i=0;i<mAgentsInGrid.size();++i){
        for(unsigned int j=0;j<i;++j){
            this->searchFrontiers(i,j,this->getHandOverReach(i,j));
            //this->_frontiers.at(getFrontierIndex(j,i))=this->searchFrontiers(i,j,this->getReach(i,j));
        }
    }
}

double MultiAgentGrid::getArmReach(AgentInGrid *a){
    double reach=-numeric_limits<double>::infinity();
    if(a->getRobot()->getHriAgent()){
        reach= a->getRobot()->getHriAgent()->manip->reach_tolerance;
    }else{
        cerr<<"MultiAgentGrid::getArmReach(): ERROR no HRI_AGENT for this robot : "<< a->getRobot()->getName()
              <<"assign -inf to it"<<endl;
    }
    return reach;
}

double MultiAgentGrid::getArmReach(unsigned int i){
    return this->getArmReach(this->getAgentStructAt(i));
}

double MultiAgentGrid::getHandOverReach(unsigned int i,unsigned int j){
    return (getArmReach(i)+getArmReach(j));
}

void MultiAgentGrid::addFrontier(unsigned int agent,unsigned int other, MultiAgentCell * cell){
    //assert(cell->getMetaFrontier(agent,getAgentStructAt(other)->getMetaAgent()).size());
    assert(cell->getMetaFrontier(agent,other) == cell->getMetaFrontier(agent,getAgentStructAt(other)->getMetaAgent() ));
    getAgentStructAt(agent)->frontiers(getAgentStructAt(other)).push_back(cell);
}

vector<MultiAgentCell*> & MultiAgentGrid::getFrontierCheckedOk(unsigned int a1, unsigned int a2){
    return getAgentStructAt(a1)->getMetaAgent()->frontiers_ok(getAgentStructAt(a2)->getMetaAgent());
}

boost::mt19937 rand_gen;
boost::uniform_01<double> uniform01;
FrontierPtr MultiAgentGrid::getRandomFrontier(unsigned int a1, unsigned int a2, double timeout_ms){
    FrontierPtr frontier;
    if(getAgentStructAt(a1)->frontierFullyChecked(getAgentStructAt(a2))){
        //get a random frontier in the list
        unsigned int index=std::floor(uniform01(rand_gen)*getAgentStructAt(a1)->frontiers(getAgentStructAt(a2)).size());
        MultiAgentCell * cell=getAgentStructAt(a1)->frontiers(getAgentStructAt(a2))[index];
        vector<MetaFrontierPtr> meta_frontiers =cell->getMetaFrontier(a1,a2);
        index=std::floor(uniform01(rand_gen)*meta_frontiers.size());
        frontier = cell->createRealFrontierFor(a1,a2,meta_frontiers[index]);
        //if(frontiers.size()){
        //    frontier=frontiers[0];
        //}else{

        //}
    }else{
        //select a cell in the accessible list of an agent, and select another in the neighboring of it, accessible by the second
        unsigned int index=std::floor(uniform01(rand_gen)*getAgentStructAt(a1)->reachable().size());
        MultiAgentCell * cell1=getAgentStructAt(a1)->reachable()[index];
        vector<MultiAgentCell*> neighbors=cell1->findNeighborsInRange(a2,getArmReach(a1)+getArmReach(a2));
        if(neighbors.size()>0){
            index=std::floor(uniform01(rand_gen)*neighbors.size());
            MultiAgentCell* cell2=neighbors[index];
            cell1->addMetaFrontier(cell2,this->areCellsInNeighbourhood(cell1,cell2,a1,cell1->computeDist(cell2)*sqrt(2)),a1,a2,cell1->computeDist(cell2));
            addFrontier(a1,a2,cell1);
            addFrontier(a2,a1,cell2);
            vector<FrontierPtr> frontiers = cell1->createAllRealFrontiersFor(a1,a2);
            assert(frontiers.size());
            frontier=frontiers[0];
        }
    }
    return frontier;
}

//---------------------------------------------------------------------------
// Cell
//---------------------------------------------------------------------------

char MultiAgentCell::MetaAgentInCell::_letter_cnt = 'A';

struct compMetaInGrid{
    bool operator()( MultiAgentCell::MetaAgentInCell::s_ptr meta_in_cell ){
        return meta_in_cell->getMetaAgentInGrid() == meta_in_grid ;
    }
    MultiAgentGrid::MetaAgentInGrid::s_ptr meta_in_grid;
};

MultiAgentCell::MetaAgentInCell::s_ptr MultiAgentCell::MetaAgentInCell::shared_clone(MultiAgentGrid::MetaAgentInGrid::s_ptr meta_in_grid){
    s_ptr ptr(new MetaAgentInCell(meta_in_grid,this->getCylinderInCell(),this->isReachabilityComputed(),this->isReachable(),this->isBoundary()));
    return ptr;
}

void MultiAgentCell::MetaAgentInCell::addRealAgent(AgentInCell *a,s_ptr s_this){
    if(a->getMetaAgent() && a->getMetaAgent()!=s_this){
        a->getMetaAgent()->removeAgent(a);
    }
    if(std::find(_realAgents.begin(),_realAgents.end(),a) == _realAgents.end()){
        this->_realAgents.push_back(a);
    }
    a->setMetaAgent(s_this);
}

void MultiAgentCell::MetaAgentInCell::removeAgent(AgentInCell *a){
    vector<AgentInCell*>::iterator it = std::find(_realAgents.begin(),_realAgents.end(),a);
    if(it!=_realAgents.end()){
        _realAgents.erase(it);
    }
}

bool MultiAgentCell::checkMetaAgent(unsigned int index){
    bool is_ok=true;
    compMetaInGrid pred_meta_in_grid;
    MultiAgentGrid::MetaAgentInGrid::s_ptr real_meta=this->getMetaFromGrid(index);
    if(real_meta != this->getAgentAt(index)->getMetaAgent()->getMetaAgentInGrid()){
        //the meta agents does not correspond
        is_ok=false;
        pred_meta_in_grid.meta_in_grid=real_meta;
        std::vector<MetaAgentInCell::s_ptr>::iterator it_find=find_if(this->_metaAgents.begin(),this->_metaAgents.end(),pred_meta_in_grid);
        if (it_find == _metaAgents.end()){
            //not found, create new
            //_metaAgents.push_back(getAgentAt(index)->getMetaAgent()->shared_clone(real_meta));
            _metaAgents.push_back(MetaAgentInCell::s_ptr(new MetaAgentInCell(getMetaFromGrid(index),getAgentAt(index)->getCylinderInCell())));
            _metaAgents.back()->addRealAgent(this->getAgentAt(index),_metaAgents.back());
        }else{
            //found, change to this
            (*it_find)->addRealAgent(this->getAgentAt(index),*it_find);
        }
    }
    return is_ok;
}

MultiAgentCell::MultiAgentCell() :
        _Open(false),
        _Closed(false),
        _costIsComputed(false)/*,
        m_humanDistIsComputed(false),
        m_robotDistIsComputed(false)*/
{
    resetData();
}

MultiAgentCell::MultiAgentCell(int i, Vector2i coord, Vector2d corner, MultiAgentGrid* grid) :
        API::TwoDCell(i,corner,grid),
        _Coord(coord),
        _Open(false),
        _Closed(false),
        _costIsComputed(false)/*,
        m_humanDistIsComputed(false),
        m_robotDistIsComputed(false)*/
{
    resetData();
}

void MultiAgentCell::resetData(){
    _tmp_color=-1;
    if(!_grid){return;}
    MultiAgentGrid *grid= dynamic_cast<MultiAgentGrid*>(this->_grid);
    AgentInCell aic;
    for(unsigned int i=0;i<_agentsInCell.size();++i){
        delete _agentsInCell[i];
    }
    _agentsInCell.clear();
    _metaAgents.clear();
    _cylinders.clear();
    std::set<Robot*> cylinders;
    for(unsigned int i=0;i<grid->getAgents().size();++i){
        cylinders.insert(grid->getAgentStructAt(i)->getCylinder() );
    }
    for(set<Robot*>::iterator it=cylinders.begin();it!=cylinders.end();++it){
        _cylinders.push_back(CylinderInCell::s_ptr(new CylinderInCell(*it)));
    }
    aic.agentInGrid=0;
    aic.distance=0;
    aic.distanceIsComputed=false;
    aic.distance_alternatives.clear();
    aic.origin_alternatives.clear();
    aic.come_from_alternatives.clear();
    aic.initFrontiersVector(grid->getAgents().size());
    _agentsInCell.reserve(grid->getAgents().size());
    for (unsigned int i=0;i<grid->getAgents().size();++i){
        _agentsInCell.push_back(new AgentInCell(aic));
        //set the robot properly
        getAgentAt(i)->agentInGrid=grid->getAgentStructAt(i);
        getAgentAt(i)->index=i;
        //set the meta agent according to the cylinder
        for(unsigned int i_c=0;i_c<_cylinders.size();++i_c){
            if(getAgentAt(i)->agentInGrid->getCylinder() == _cylinders[i_c]->getCylinder()){
                compMetaInGrid pred_comp_meta_in_grid;
                pred_comp_meta_in_grid.meta_in_grid=getAgentAt(i)->agentInGrid->getMetaAgent();
                vector<MetaAgentInCell::s_ptr>::iterator it_find=find_if(_metaAgents.begin(),_metaAgents.end(),pred_comp_meta_in_grid);
                if(it_find != _metaAgents.end()){
                    (*it_find)->addRealAgent(getAgentAt(i),*it_find);
                }else{
                    _metaAgents.push_back(
                                MetaAgentInCell::s_ptr(
                                    new MetaAgentInCell(getAgentAt(i)->agentInGrid->getMetaAgent(),_cylinders[i_c])));
                    (_metaAgents.back())->addRealAgent(getAgentAt(i),_metaAgents.back());
                }
                break;
            }
        }
    }
}

double MultiAgentCell::getAltDistance(unsigned int index, MultiAgentCell *alt_origin){
    double dist(numeric_limits<double>::infinity());
    AgentInCell* a=(getAgentAt(index));
    unsigned int i(0);
    while (i<a->origin_alternatives.size() && a->origin_alternatives[i]!=alt_origin) {
        ++i;
    }
    if(i<a->origin_alternatives.size()){
        dist=a->distance_alternatives[i];
    }
    return dist;
}

MultiAgentCell* MultiAgentCell::getAltCameFrom(unsigned int index, MultiAgentCell *alt_origin){
    MultiAgentCell* c=0;
    AgentInCell* a=(getAgentAt(index));
    unsigned int i(0);
    while (i<a->origin_alternatives.size() && a->origin_alternatives[i]!=alt_origin) {
        ++i;
    }
    if(i<a->origin_alternatives.size()){
        c=a->come_from_alternatives[i];
    }
    return c;
}

vector<MultiAgentCell*> MultiAgentCell::getAltTraj(unsigned int index, MultiAgentCell *alt_origin,bool reverse){
    vector<MultiAgentCell*> traj;
    AgentInCell* a=(getAgentAt(index));
    unsigned int i(0);
    while (i<a->origin_alternatives.size() && a->origin_alternatives[i]!=alt_origin) {
        ++i;
    }
    if(i<a->origin_alternatives.size()){
        MultiAgentCell *c_t=this;
        while(c_t && c_t!=alt_origin){
            traj.push_back(c_t);
            c_t=c_t->getAltCameFrom(index,alt_origin);
            if(c_t==0){
                cerr<<"alternative 2d path is broken"<<endl;
            }
        }
        traj.push_back(c_t);
        //here, the path is from this to alt_origin, we normally want the other direction:
        //but we can ask for this one too
        if(!reverse){
            std::reverse(traj.begin(),traj.end());
        }
    }
    return traj;
}
double MultiAgentCell::getDistanceAny(unsigned int agent, MultiAgentCell *other){
    double d=numeric_limits<double>::infinity();
    if(other==getAgentAt(agent)->agentInGrid->getOrigin()){
        //other is the origin
        d= getDistance(agent);
    }else if(this==getAgentAt(agent)->agentInGrid->getOrigin()){
        //this is the origin
        d= other->getDistance(agent);
    }else if(this->getAltDistance(agent,other) < numeric_limits<double>::infinity()){
        //other is an alternative origin
        d= this->getAltDistance(agent,other);
    }else if(other->getAltDistance(agent,this) < numeric_limits<double>::infinity() ){
        //this is an alternative origin
        d= other->getAltDistance(agent,this);
    }
    if (d==numeric_limits<double>::infinity()){
        cout<<"getDistanceAny: not found for agent "<<getAgentAt(agent)->agentInGrid->getRobot()->getName()<<endl;
        cout << "reachable > cell this: "<<this->isReachable(agent)<<" ; other: "<<other->isReachable(agent)<<endl;
    }
    
    return d;
}

vector<MultiAgentCell*> MultiAgentCell::getTrajAny(unsigned int agent, MultiAgentCell *origin){
    if(origin==getAgentAt(agent)->agentInGrid->getOrigin()){
        //other is the origin
        return getTraj(agent); // get traj from origin to this
    }else if(this==getAgentAt(agent)->agentInGrid->getOrigin()){
        //this is the origin
        vector<MultiAgentCell*> tmp = origin->getTraj(agent);//get path from this to origin
        std::reverse(tmp.begin(),tmp.end()); // reverse the path => from origin to this
        return tmp;
    }else if(getAltDistance(agent,origin)<numeric_limits<double>::infinity()){
        return getAltTraj(agent,origin);//get path from origin to this
    }else if(origin->getAltDistance(agent,this) < numeric_limits<double>::infinity() ){
        return origin->getAltTraj(agent,this,true);//3rd arg=true => get reversed path
    }
    return vector<MultiAgentCell*>(0);

}
void MultiAgentCell::setAltDistance(unsigned int index, double value, MultiAgentCell *alt_origin,MultiAgentCell *previous_cell){
    AgentInCell* a=(getAgentAt(index));
    unsigned int i(0);
    while (i<a->origin_alternatives.size() && a->origin_alternatives[i]!=alt_origin) {
        ++i;
    }
    if(i<a->origin_alternatives.size()){
        a->distance_alternatives[i]=value;
        if(previous_cell){
            a->come_from_alternatives[i]=previous_cell;
        }
    }else{
        //create and set the alternative
        a->origin_alternatives.push_back(alt_origin);
        a->distance_alternatives.push_back(value);
        a->come_from_alternatives.push_back(previous_cell);
        if(a->origin_alternatives.size()>dynamic_cast<MultiAgentGrid*>(this->_grid)->getMaxAlternativeDistancesForCells()){
            //ensure the number of alternatives does not explodes
            a->origin_alternatives.pop_front();
            a->distance_alternatives.pop_front();
            a->come_from_alternatives.pop_front();
        }
    }
}

bool MultiAgentCell::checkIsInCollision(unsigned int agent_this, MultiAgentCell *other, unsigned int agent_other){
    double d=this->computeDist(other);
    double col_dist=getAgentAt(agent_this)->getMetaAgent()->getMetaAgentInGrid()->getCylinderRadius();
    col_dist+=getAgentAt(agent_other)->getMetaAgent()->getMetaAgentInGrid()->getCylinderRadius();
    return d<=col_dist;
}

std::vector<FrontierPtr> MultiAgentCell::createAllRealFrontiersFor(unsigned int a1, unsigned int a2){
    vector<MetaFrontierPtr> & meta = getMetaFrontier(a1,a2);
    vector<FrontierPtr> frontiers;
    for (vector<MetaFrontierPtr>::iterator it = meta.begin();it!=meta.end();++it){
        FrontierPtr ptr = createRealFrontierFor(a1,a2,*it);
        frontiers.push_back(ptr);//add the frontier to the vector to return
    }
    return frontiers;
    //TODO sort
}

FrontierPtr MultiAgentCell::createRealFrontierFor(unsigned int a1,unsigned int a2, MetaFrontierPtr mfrontier){
    assert(this==mfrontier->cellOf(getMetaFromGrid(a1)) || getMetaFromGrid(a1)==getMetaFromGrid(a2) );//a1 cell is this OR we can't discrimine a1 vs a2
    FrontierPtr ptr = getRealFrontier(a1,a2,mfrontier);
    if(!ptr){
        MultiAgentCell *cellA=mfrontier->getCellA(),*cellB=mfrontier->getCellB();
        unsigned int aA(a1),aB(a2);
        if(cellB==this){
            std::swap(aA,aB);
        }else if(cellA==this){
            //nothing
        }else{
            assert(false);
        }
        //if(cellA!=mfrontier->cellOf(getMetaFromGrid(aA))){
        //    std::swap(aA,aB);
        //    //if swapped, cellB == this, as then ptr->cellOf(a1) == ptr->cellB()
        //}
        assert(mfrontier->cellOf(getMetaFromGrid(aB))->isReachable(aB));
        assert(mfrontier->cellOf(getMetaFromGrid(aA))->isReachable(aA));
        assert(mfrontier->cellOf(getMetaFromGrid(aA))==cellA);
        //the frontier is not set
        ptr = FrontierPtr(new FrontierCells);

        ptr->agentA=cellA->getAgentAt(aA);
        ptr->agentA_ind=aA;
        ptr->agentB=cellB->getAgentAt(aB);
        ptr->agentB_ind=aB;
        ptr->meta_frontier=mfrontier;
        assert(ptr->check_status==FrontierCells::FCS_NONE);
        //TODO add this correctly
        //if(mfrontier->getCheckedOk()){
        //    ptr->check_status=FrontierCells::FCS_OBJECT;
        //    ptr->object_ok=true;
        //    ptr->object_position= mfrontier->getObjectPosition();
        //}
        mfrontier->setRealFrontier(getAgentAt(aA)->agentInGrid,getAgentAt(aB)->agentInGrid,ptr);
        cellA->getAgentAt(aA)->frontiers(cellB->getAgentAt(aB)).push_back(ptr);
        cellB->getAgentAt(aB)->frontiers(cellA->getAgentAt(aA)).push_back(ptr);
        assert(ptr->cellA()->isReachable(ptr->agentA_ind) && ptr->cellB()->isReachable(ptr->agentB_ind));
        assert(ptr->cellOf(a1)==this);
        assert(ptr->cellA() != ptr->cellB());
    }else{
        assert(ptr->cellOf(a1)==this);
        assert(ptr->cellA()->isReachable(ptr->agentA_ind) && ptr->cellB()->isReachable(ptr->agentB_ind));
        assert(ptr->cellA() != ptr->cellB());
    }
    return ptr;
}

FrontierPtr MultiAgentCell::getRealFrontier(unsigned int a1, unsigned int a2, MetaFrontierPtr mfrontier){
    unsigned int aA(a1),aB(a2);
    if(mfrontier->getCellB()==this){
        std::swap(aA,aB);
        //swap so that p->cellOf(a1) == this
    }
    FrontierPtr p= mfrontier->getRealFrontier(dynamic_cast<MultiAgentGrid*>(this->_grid)->getAgentStructAt(aA),dynamic_cast<MultiAgentGrid*>(this->_grid)->getAgentStructAt(aB));
    assert(!p || p->cellOf(a1)==this);
    return p;
}

void MultiAgentCell::computeAccessibility(unsigned int index_agent)
{

//    Robot* cyl = dynamic_cast<MultiAgentGrid*>(_grid)->getCylinder(index_agent);

    //detection of collision
    CylinderInCell::s_ptr cyl=getAgentAt(index_agent)->getCylinderInCell();
    Robot * cylinder=cyl->getCylinder();

    if(cylinder==0){
        cerr<<"MultiAgentCell::computeReach("<<index_agent<<") : error, bad agent descriptor"<<endl;
        return;
    }
    //shared_ptr<Configuration> q_agent_cur = agent->getCurrentPos();
    //shared_ptr<Configuration> q_agent = agent->getCurrentPos();
    //(*q_agent)[6] = 0;
    //(*q_agent)[7] = 0;
    //agent->setAndUpdate(*q_agent);

    shared_ptr<Configuration> q = cylinder->getCurrentPos();
    (*q)[6] = this->getCenter()[0];
    (*q)[7] = this->getCenter()[1];
    cylinder->setAndUpdate(*q);
    this->setAccessible(index_agent,!cylinder->isInCollisionWithOthersAndEnv());
    (*q)[6] = 0;
    (*q)[7] = 0;
    cylinder->setAndUpdate(*q);
    //agent->setAndUpdate(*q_agent_cur);
}


//double MultiAgentCell::computeMoveCost(unsigned int index_agent){
//
//
//    MultiAgentGrid* grid = dynamic_cast<MultiAgentGrid*>(this->_grid);
//    double speed = grid->getAgentStructAt(index_agent)->speed;
//
//    double duration= this->getDistance(index_agent)/speed;
//    getAgentAt(index_agent)->moveCost=duration;
//    return getAgentAt(index_agent)->moveCost;
//}


std::vector<MultiAgentCell*> MultiAgentCell::getNeighbors(unsigned int index)
{
    std::vector<MultiAgentCell*> neighbors;
    Vector2i coord = getCoord();
    double nbCellX = dynamic_cast<MultiAgentGrid*>(_grid)->getNbCellX();
    double nbCellY = dynamic_cast<MultiAgentGrid*>(_grid)->getNbCellY();
    for (int i = -1; i <= 1; i++)
    {
        for (int j = -1; j <= 1; j++)
        {
            if (i != 0 || j != 0)
            {

                if (coord[0]+i < nbCellX && coord[0]+i >= 0 && coord[1]+j < nbCellY && coord[1]+j >= 0)
                {
                    MultiAgentCell* cell = dynamic_cast<MultiAgentCell*>(dynamic_cast<MultiAgentGrid*>(_grid)->getCell(coord[0]+i,coord[1]+j));
                    if ( !cell->isAccessibilityComputed(index)){
                        cell->computeAccessibility(index);
                    }
                    if (cell->isAccessible(index))
                    {
                        neighbors.push_back(cell);
                    }else{
                        this->getAgentAt(index)->setBoundary(true); //set this cell as boundary for this agent as it has
                                                                    //a neighbour cell not accessible. this cell is near an
                                                                    //obstacle
                    }
                }
            }
        }
    }

    return neighbors;
}

std::vector<MultiAgentCell*> MultiAgentCell::findNeighborsInRange(unsigned int agent, double range){
    std::vector<MultiAgentCell*> neighbors;
    Vector2i coord = getCoord();
    Vector2d size=this->getCellSize();
    double range_square = pow(range,2);
    //we define an ellipse in the discrete plan that represent the circle in the continuous space
    //the ellipse is because of the possibility to have different cell size over the x and y axis (why?)
    //first the rectangle in which the ellipse is inscribed:
    int x_half_offset = std::ceil(range/size.x());
    int y_half_offset = std::ceil(range/size.y());
    double x_ratio=size.x(),y_ratio=size.y();//for the ellipse equation
    double nbCellX = dynamic_cast<MultiAgentGrid*>(_grid)->getNbCellX();
    double nbCellY = dynamic_cast<MultiAgentGrid*>(_grid)->getNbCellY();
    for (int x = -x_half_offset; x <= x_half_offset; x++)
    {
        double xx2=pow(x*x_ratio,2);
        for (int y = -y_half_offset; y <= y_half_offset; y++)
        {
            if (x != 0 || y != 0 )
            {
                if( xx2 + pow(y*y_ratio,2) <=  range_square ){
                    //the point is inside the ellipse (x/a)^2 + (y/b)^2 <= r^2
                    if (coord[0]+x < nbCellX && coord[0]+x >= 0 && coord[1]+y < nbCellY && coord[1]+y >= 0)
                    {
                        MultiAgentCell* cell = dynamic_cast<MultiAgentCell*>(dynamic_cast<MultiAgentGrid*>(_grid)->getCell(coord[0]+x,coord[1]+y));
                        if (cell && cell->isReachable(agent))
                        {
                            neighbors.push_back(cell);
                        }
                    }
                }
            }
        }
    }

    return neighbors;

}


double MultiAgentCell::computeDist(MultiAgentCell* neighCell)
{
    return sqrt(pow(getCenter()[0] - neighCell->getCenter()[0] , 2) + pow(getCenter()[1] - neighCell->getCenter()[1] , 2));
}

void MultiAgentCell::resetAccessibility()
{
    for (unsigned int i=0;i<_agentsInCell.size();++i){
        resetAccessibility(i);
    }
}

void MultiAgentCell::resetAccessibility(unsigned int index){
    AgentInCell * ag=_agentsInCell[index];
    ag->unsetAccessibilityComputed();
}

void MultiAgentCell::resetAllDistances(){
    for (unsigned int i=0;i<_agentsInCell.size();++i){
        resetDistances(i);
    }
}

void MultiAgentCell::resetDistances(unsigned int index){
    AgentInCell * ag=_agentsInCell[index];
    resetTraj(index);
    ag->distance=0;
    ag->distanceIsComputed=false;
    ag->setBoundary(false);
}

void MultiAgentCell::resetTraj(){
    for (unsigned int i=0;i<_agentsInCell.size();++i){
        resetTraj(i);
    }
}

void MultiAgentCell::resetTraj(unsigned int index)
{
    AgentInCell * ag=_agentsInCell[index];
    ag->distanceIsComputed = false;
    ag->traj.clear();
    ag->vect_traj.clear();
}


bool compareFrontiersDist(shared_ptr<FrontierCells> a,shared_ptr<FrontierCells> b){
    return (a->dist()<b->dist());
}

shared_ptr<MetaFrontierCells> MultiAgentCell::addMetaFrontier(MultiAgentCell *other, bool over_obstacle, unsigned int agent_this, unsigned int agent_other, double dist,bool check_exists){
    return addMetaFrontier(other, over_obstacle, this->getAgentAt(agent_this)->getMetaAgent(),other->getAgentAt(agent_other)->getMetaAgent(),dist,check_exists);
}

shared_ptr<MetaFrontierCells> MultiAgentCell::addMetaFrontier(MultiAgentCell *other, bool over_obstacle, MetaAgentInCell::s_ptr agent_this, MetaAgentInCell::s_ptr agent_other, double dist,bool check_exists){
    MetaFrontierPtr ptr;
    //assert agent_other is a meta agent declared in the cell other
    assert(find(other->_metaAgents.begin(),other->_metaAgents.end(),agent_other)!=other->_metaAgents.end());
    assert(agent_this->isReachable() && agent_other->isReachable());
    vector<MetaFrontierPtr> & frontiers_agent_this=agent_this->getFrontiers(agent_other);
    vector<MetaFrontierPtr>::iterator it_find=frontiers_agent_this.end();
    if(check_exists){
        MetaFrontierCells::CheckCellsAre comp(this,agent_this->getMetaAgentInGrid(),other,agent_other->getMetaAgentInGrid());
        it_find=find_if(frontiers_agent_this.begin(),frontiers_agent_this.end(),comp);
    }
    //if not found or not searched
    if(it_find==frontiers_agent_this.end()){
        ptr=MetaFrontierPtr(new MetaFrontierCells(agent_this->getMetaAgentInGrid(),this,agent_other->getMetaAgentInGrid(),other,dist,over_obstacle));
        frontiers_agent_this.push_back(ptr);
        agent_other->getFrontiers(agent_this).push_back(ptr);
    }else{
        ptr=*it_find;
    }
    return ptr;
}

shared_ptr<FrontierCells> MultiAgentCell::addFrontierComplete(MultiAgentCell *other, bool over_obstacle, unsigned int agent_this, unsigned int agent_other, double dist,bool check_exists){
#ifndef NDEBUG
    if(!check_exists){
        for(unsigned int i=0;i<this->getMetaFrontier(agent_this,agent_other).size();++i){
            //asserts the frontier does not already exist
            assert(this->getMetaFrontier(agent_this,agent_other)[i]->cellOf(getAgentAt(agent_other)->getMetaAgent()->getMetaAgentInGrid())!=other ||
                    this->getMetaFrontier(agent_this,agent_other)[i]->cellOf(getAgentAt(agent_this)->getMetaAgent()->getMetaAgentInGrid()) != this);
        }
    }
#endif
    MetaFrontierPtr meta=addMetaFrontier(other,over_obstacle,agent_this,agent_other,dist,check_exists);
    return createRealFrontierFor(agent_this,agent_other,meta);
}

FrontierCells::~FrontierCells(){
    if(trajA){
        delete trajA;
    }
    if(trajB){
        delete trajB;
    }
}

FrontierPtr FrontierCells::neighbourFrontier(double keep_distance, double min_distance){
    vector<MultiAgentCell*> neigA = this->cellA()->findNeighborsInRange(this->agentA_ind,keep_distance);
    vector<MultiAgentCell*> neigB = this->cellB()->findNeighborsInRange(this->agentB_ind,keep_distance);

    //TODO less random, more smartness in the neighbour frontier selection
    unsigned int iA,iB;
    iA=rand() % (neigA.size() + 1);
    iB=rand() % (neigB.size() + 1);
    MultiAgentCell *neigA_cell=(iA<neigA.size() ? neigA[iA] : this->cellA());
    MultiAgentCell *neigB_cell=(iB<neigB.size() ? neigB[iB] : this->cellB());
    //instaciate eventually existing meta frontiers, returns all existing frontiers
    vector<FrontierPtr> frontiers = neigA_cell->createAllRealFrontiersFor(agentA_ind,agentB_ind);
    unsigned int i=0;
    //find if the selected frontiers already exists
    while(i<frontiers.size() && frontiers[i]->cellOf(agentB_ind) != neigB_cell){ ++i; }
    if(i<frontiers.size()){
        //it is found, return it
        return frontiers[i];
    }else{
        //it does not exists, create it
        double dist=neigA_cell->computeDist(neigB_cell);
        if(dist<keep_distance && dist>min_distance){
            cerr<< "error deprecated need to compute if over obstacle"<<endl;
            abort();
            //return neigA_cell->addFrontierComplete(neigB_cell,agentA_ind,agentB_ind,dist);
        }
    }
    //returns an empty frontier, meaning nothing have been found
    FrontierPtr empty;
    return empty;
}

vector<FrontierPtr> FrontierCells::findNeighbourFrontiers(double keep_distance,double min_distance){
    vector<MultiAgentCell*> neigA = this->cellA()->getNeighbors(this->agentA_ind);
    vector<MultiAgentCell*> neigB = this->cellB()->getNeighbors(this->agentB_ind);
    vector<FrontierPtr> frontiers_found;

    for(vector<MultiAgentCell*>::iterator itA=neigA.begin();itA!=neigA.end();++itA){
        for(vector<MultiAgentCell*>::iterator itB=neigB.begin(); itB != neigB.end() ; ++itB){
            if(*itA!=*itB){
                //two different cells
                MetaFrontierCells::CheckCellsAre comp(*itA,agentA->getMetaAgent()->getMetaAgentInGrid(),*itB,agentB->getMetaAgent()->getMetaAgentInGrid());
                //find if a meta frontier already exists for this pair of cells
                //it has same cells
                vector<MetaFrontierPtr>::iterator meta=find_if((*itA)->getMetaFrontier(agentA_ind,agentB_ind).begin(),
                                              (*itA)->getMetaFrontier(agentA_ind,agentB_ind).end(),
                                              comp);
                //check dist is in range
                double dist=(*itA)->computeDist(*itB);
                if(dist<keep_distance && dist>min_distance){
                    if(meta!=(*itA)->getMetaFrontier(agentA_ind,agentB_ind).end()){
                        //found an existing meta, get/create the real associated
                        frontiers_found.push_back((*itA)->createRealFrontierFor(agentA_ind,agentB_ind,*meta));
                    }else{
                        //does not exists, create
                        abort();
                        //frontiers_found.push_back( (*itA)->addFrontierComplete(*itB,agentA_ind,agentB_ind,dist) );
                    }
                }
            }
        }
    }
    return frontiers_found;

}


vector<FrontierPtr> FrontierCells::findNeighbourFrontiersJumping(double keep_distance,double min_distance,unsigned int number){
    //some neighbours of the original frontier cells:
    vector<MultiAgentCell*> neigA = this->cellA()->findNeighborsInRange(this->agentA_ind,keep_distance/2);
    vector<MultiAgentCell*> neigB = this->cellB()->findNeighborsInRange(this->agentB_ind,keep_distance/2);
    vector<FrontierPtr> frontiers_found; // to be returned

    if(neigA.size() !=0 || neigB.size()!=0){
        while(frontiers_found.size()<number){
            vector<MultiAgentCell*>::iterator it1,it2;
            vector<MultiAgentCell*> neigOther;
            unsigned int a1_ind,a2_ind;
            MultiAgentCell::AgentInCell *a1,*a2;
            //alternate between cellA and cellB neighbours
            if(rand()<RAND_MAX/2){
                it1=neigA.begin()+(rand()%neigA.size());
                a1_ind=agentA_ind;
                a2_ind=agentB_ind;
                a1=agentA;
                a2=agentB;
            }else{
                it1=neigB.begin()+(rand()%neigB.size());
                a1_ind=agentB_ind;
                a2_ind=agentA_ind;
                a1=agentB;
                a2=agentA;
            }
            //get a random neighbour of the selected cell to be the other cell of the frontier
            neigOther = (*it1)->findNeighborsInRange(a2_ind,keep_distance);
            if(neigOther.size()){
                it2=neigOther.begin()+(rand()%neigOther.size());
                if(*it1!=*it2){
                    //two different cells
                    MetaFrontierCells::CheckCellsAre comp(*it1,a1->getMetaAgent()->getMetaAgentInGrid(),
                                                          *it2,a2->getMetaAgent()->getMetaAgentInGrid());
                    //find if a meta frontier already exists for this pair of cells
                    //it has same cells
                    vector<MetaFrontierPtr>::iterator meta=find_if((*it1)->getMetaFrontier(a1_ind,a2_ind).begin(),
                                                                   (*it1)->getMetaFrontier(a1_ind,a2_ind).end(),
                                                                   comp);
                    //check dist is in range
                    double dist=(*it1)->computeDist(*it2);
                    if(dist<keep_distance && dist>min_distance){
                        if(meta!=(*it1)->getMetaFrontier(a1_ind,a2_ind).end()){
                            //found an existing meta, get/create the real associated
                            frontiers_found.push_back((*it1)->createRealFrontierFor(a1_ind,a2_ind,*meta));
                        }else{
                            //does not exists, create
                            abort();
                            //frontiers_found.push_back( (*it1)->addFrontierComplete(*it2,a1_ind,a2_ind,dist) );
                        }
                    }
                }
            }
        }
    }
    return frontiers_found;

}

void FrontierCells::deSymetrize(){
    if(isSymetric()){
        this->_symetric = false;
        this->r_check_status = this->check_status;
        this->r_grabA=grabA;
        this->r_grabB=grabB;
        if(this->object_position)
            this->r_object_position = confPtr_t(new Configuration(*this->object_position));
        this->r_status_ok.assign(this->status_ok.begin(),this->status_ok.end());
        if(this->trajA)
            this->r_trajA = new API::Trajectory(*this->trajA);
        if(this->trajB)
            this->r_trajB = new API::Trajectory(*this->trajB);
        if(this->confA)
            this->r_confA = confPtr_t(new Configuration(*this->confA));
        if(this->confB)
            this->r_confB = confPtr_t(new Configuration(*this->confB));
    }
}
