#include "ent_surface.hpp"
#include "project.hpp"
#include <fstream>

namespace Ent {


INIT_MOVE3D_STATIC_LOGGER(Surface, "planners.entities.surface");

Surface::Surface(std::string global_json_path):
    _global_json_path(global_json_path)
{
}

Surface::~Surface()
{
}

std::string Surface::getPath()
{
    std::string file=_global_json_path;
    file.append("/Surface.json");
    return file;
}


bool Surface::fetchData()
{
    bool ok = importJsonFile(getPath());
    return ok;
}


bool Surface::importJsonFile(const std::string &path)
{
    std::ifstream ifs;
    Json::Value root;
    bool ok=true;

    ifs.open(path.c_str());

    if(!ifs){
        ok=false;
        M3D_ERROR("could not open "<<path);
    }
    if(ok){
        try{
            ifs>>root;
        }catch (std::runtime_error e){
            M3D_ERROR("error while parsing file "<<path<<": "<<e.what());
            ok=false;
        }catch(std::exception e){
            M3D_ERROR("error while parsing file "<<path<<": "<<e.what());
            ok=false;
        }
    }
    if(ok){
        if(!root){
            M3D_ERROR("could not parse json file "<<path);
            ok=false;
        }
    }
    if(ok){
        if(!createSurfaceInObjects(root)){
            ok=false;
            M3D_ERROR("could not properly parse objects in file "<<path);
        }
    }
    return ok;
}


bool Surface::createSurfaceInObjects(Json::Value const & value)
{
    bool ok = true;
    M3D_INFO("List of surfaces entities:");
    for(Json::Value::iterator it = value.begin(); it !=value.end(); ++it)
    {
        //        Json::Value key = it.key();
        Json::Value cont = (*it);

        if(cont.isMember("complete_name"))
        {
            for(Json::Value::iterator it2 = cont["complete_name"].begin(); it2 !=cont["complete_name"].end(); ++it2)
            {
                if (global_Project->getActiveScene()->getRobotByName((*it2).asString()))
                {
                    Robot* r = global_Project->getActiveScene()->getRobotByName((*it2).asString());
                    if (r)
                    {
                        M3D_INFO(r->getName());
                        ok = ok && setToObjRob(r, cont["surfaces"]);
                    }
                }
            }
        }
        if (cont.isMember("partial_name"))
        {
            for(Json::Value::iterator it2 = cont["partial_name"].begin(); it2 !=cont["partial_name"].end(); ++it2)
            {
                std::vector<Robot*> vr = global_Project->getActiveScene()->getAllRobotWithNameContaining((*it2).asString());
                for (unsigned int i = 0; i < vr.size();i++)
                {
                    M3D_INFO(vr.at(i)->getName());
                    ok = ok && setToObjRob(vr.at(i), cont["surfaces"]);
                }
            }
        }

        if (!cont.isMember("complete_name") && !cont.isMember("partial_name") && !cont.isMember("_comment"))
        {
            M3D_ERROR("complete_name and partial_name does not exit for an entry");
            ok = false;
        }
    }
    return ok;
}

bool Surface::setToObjRob(Robot* rob, Json::Value const & value)
{
    bool ok = true;
    if(value.isArray())
    {
        for(Json::Value::const_iterator it = value.begin(); it !=value.end(); ++it)
        {
            Json::Value oneFormJson = (*it);
            //        cout << oneFormJson.toStyledString() << endl;

            if (oneFormJson.isMember("rectangle"))
            {
                std::vector<p3d_point> vp;
                for(Json::Value::iterator it2 = oneFormJson["rectangle"].begin(); it2 !=oneFormJson["rectangle"].end(); ++it2)
                {
                    if ((*it2).size() != 3)
                    {
                        M3D_ERROR("not a point to define the rectangle");
                        ok = false;
                    }
                    else
                    {
                        p3d_point tmp;
                        tmp.x = (*it2)[0].asDouble();
                        tmp.y = (*it2)[1].asDouble();
                        tmp.z = (*it2)[2].asDouble();
                        vp.push_back(tmp);
                    }
                }
                if (vp.size() != 4)
                {
                    M3D_ERROR("not enough points to define the rectangle");
                    ok = false;
                }
                p3d_vector3 norm;
                if (oneFormJson.isMember("normal") && oneFormJson["normal"].size() == 3)
                {

                    norm[0] = oneFormJson["normal"][0].asDouble();
                    norm[1] = oneFormJson["normal"][1].asDouble();
                    norm[2] = oneFormJson["normal"][2].asDouble();
                }
                else
                {
                    M3D_ERROR("no norm for the rectangle");
                    ok = false;
                }

                if (ok)
                {
                    RectangleForm* rect = new RectangleForm(vp.at(0),vp.at(1),vp.at(2),vp.at(3),norm);
                    rob->getObjectRob()->addSupports(rect);
                    rob->getObjectRob()->setSupport(true);
                }
            }
            else
            {
                M3D_ERROR("undefined form type");
                ok =false;
            }
        }
    }
    else
    {
        if (value.isString() && value.asString() == "default")
        {
            std::tr1::shared_ptr<Configuration> q_cur = rob->getCurrentPos();
            std::tr1::shared_ptr<Configuration> q = (std::tr1::shared_ptr<Configuration>)new Configuration(*q_cur);
            (*q)[6] = 0;
            (*q)[7] = 0;
            (*q)[8] = 0;
            (*q)[9] = 0;
            (*q)[10] = 0;
            (*q)[11] = 0;
            rob->setAndUpdate(*q);

            p3d_vector3 m;
            m[0] = 0;
            m[1] = 0;
            m[2] = 1;

            GeometricComputations GC = GeometricComputations();

            std::vector<p3d_point> vs = GC.getUpperPointsInObject(rob,0.03);
            if (vs.size() !=4)
                ok = false;
            else
            {
                RectangleForm* rect = new RectangleForm(vs.at(0),vs.at(1),vs.at(2),vs.at(3),m);
                rob->getObjectRob()->addSupports(rect);
                rob->getObjectRob()->setSupport(true);
            }

            rob->setAndUpdate(*q_cur);

        }
        else
            ok = false;
    }
    return ok;
}



}
