#include "Sem_Actions.hpp"
#include <fstream>
#include <libgen.h>
#include <dirent.h>
#include "Sem_AgentType.hpp"

namespace Sem {

using namespace std;

INIT_MOVE3D_STATIC_LOGGER(ActionType, "planners.semantics.actiontype");

// ####################
//  ActionType //

std::vector<std::string> ActionType::__stringValues = std::vector<std::string>(0);
ActionType::ActionType():
    _value(-1)
{
}

ActionType::ActionType(const std::string &s)
{
    std::vector<std::string>::const_iterator it=std::find(ActionType::__stringValues.begin(),__stringValues.end(),s);
    if (it!=ActionType::__stringValues.end()){
        this->_value=it-__stringValues.begin();
    }else{
        __stringValues.push_back(s);
        this->_value=__stringValues.size()-1;
    }
}

ActionType::~ActionType()
{
}

string ActionType::toString() const
{
    if(isDefined()){
        return __stringValues[this->_value];
    }else{
        return "undef!";
    }
}

std::vector<ActionType> ActionType::currentlyDefinedActionTypes()
{
    std::vector<ActionType> v;
    for(unsigned int i=0;i<__stringValues.size();i++){
        v.push_back(ActionType(i));
    }
    return v;
}

ActionType::ActionType(int val)
{
    if(val>=0){
        if (static_cast<unsigned int>(val) <__stringValues.size()){
            _value=val;
        }else{
            _value=-1;
        }
    }else{
        _value=-1;
    }
}

bool ActionType::isDefined() const
{
    if(_value >=0 ){
        return true;
    }else{
        return false;
    }
}

// ####################
//  ActionDefinition //

INIT_MOVE3D_STATIC_LOGGER(ActionDefinition, "planners.semantics.actiondefinition");

ActionDefinition::ActionDefinition():
    _requiredParts(0),
    _hriParts(0)
{
    _actionType=ActionType();
}

ActionDefinition::ActionDefinition(ActionType type):
    _actionType(type),
    _requiredParts(0),
    _hriParts(0)
{
}

ActionDefinition::~ActionDefinition()
{

}

const std::vector<PlanningType> &ActionDefinition::requiredParts() const
{
    return _requiredParts;
}

void ActionDefinition::setRequiredParts(const std::vector<PlanningType> &requiredParts)
{
    _requiredParts = requiredParts;
}
ActionType ActionDefinition::actionType() const
{
    return _actionType;
}

void ActionDefinition::setActionType(const ActionType &actionType)
{
    _actionType = actionType;
}
std::vector<PlanningType> ActionDefinition::hriParts() const
{
    return _hriParts;
}

void ActionDefinition::setHriParts(const std::vector<PlanningType> &hriParts)
{
    _hriParts.assign(hriParts.begin(),hriParts.end());
}

std::set<AgentType> &ActionDefinition::include()
{
    return _include;
}

std::set<AgentType> &ActionDefinition::exclude()
{
    return _exclude;
}
const std::set<AgentType> & ActionDefinition::include() const
{
    return _include;
}

void ActionDefinition::setInclude(const std::set<AgentType> &include)
{
    _include = include;
}
const std::set<AgentType> &ActionDefinition::exclude() const
{
    return _exclude;
}

void ActionDefinition::setExclude(const std::set<AgentType> &exclude)
{
    _exclude = exclude;
}

void ActionDefinition::addExclude(AgentType agent_type)
{
    _exclude.insert(agent_type);
}

bool ActionDefinition::isWhiteListed(AgentType agent_type) const{
    bool white=true;//white list all by default
    if(_include.size()){
        //excluded by default
        white=false;
        if(_include.find(agent_type) != _include.end()){
            white=true;
        }
    }
    if (white && _exclude.size()){
        //included by default
        white = true;
        if(_exclude.find(agent_type) != _exclude.end()){
            white=false;
        }
    }
    return white;
}




// ####################
//  Actions //

INIT_MOVE3D_STATIC_LOGGER(Actions, "planners.semantics.actions");

Actions::Actions(string global_json_path, bool proactive):
    _global_json_path(global_json_path)
{
    setProactive(proactive);
}

Actions::~Actions()
{

}

const ActionDefinition *Actions::getActionDefinition(ActionType t)
{
    //std::map<ActionType,ActionDefinition *>::const_iterator it;
    if(_actions.count(t) ==0){
        if (!proactive()){
            fetchAction(t);
        }
    }
    if(_actions.count(t) !=0){
        return _actions[t];
    }else{
        return 0;
        //return _actions.insert(std::make_pair(t,new ActionDefinition(t))).first->second;
    }
}

string Actions::getPath()
{
    std::string file=_global_json_path;
    file.append("/Actions/");
    return file;
}

string Actions::getPath(ActionType t)
{
    return getPath()+t.toString()+std::string(".json");
}

bool Actions::fetchData()
{
    DIR* dir;
    bool ok(true);
    struct dirent *ent;
    std::string path = getPath();
    if ((dir = opendir(path.c_str())) != NULL){
        while ((ent = readdir (dir)) != NULL) {
            std::string name(ent->d_name);
            if ( name.find(".json") == name.size()-5){
                bool it_ok=true;
                it_ok = importJsonFile(path+name);
                ok = ok && it_ok;
            }
        }
    }
    return ok;
}

bool Actions::fetchAction(ActionType t){
    return importJsonFile(getPath(t),t);
}

bool Actions::importJsonFile(const std::string &path,ActionType t)
{
    std::ifstream ifs;
    Json::Value root;
    bool ok=true;

    ifs.open(path.c_str());

    if(!ifs){
        ok=false;
        M3D_ERROR("could not open "<<path);
    }
    if(ok){
        try{
            ifs>>root;
        }catch (std::runtime_error e){
            M3D_ERROR("error while parsing file "<<path<<": "<<e.what());
            ok=false;
        }catch(std::exception e){
            M3D_ERROR("error while parsing file "<<path<<": "<<e.what());
            ok=false;
        }
    }
    if(ok){
        if(!root){
            M3D_ERROR("could not parse json file "<<path);
            ok=false;
        }
    }
    if(ok){
        if(!addActionDef(t,root)){
            ok=false;
            M3D_ERROR("could not properly parse action definition for action "
                      <<t.toString()<<" in file "<<path);
        }
    }
    return ok;
}

bool Actions::importJsonFile(const std::string &path){
    char * c_path = new char [path.length()+1];
    std::strcpy (c_path, path.c_str());
    std::string base = basename(c_path);
    std::string::size_type suffix = base.find(".json");
    std::string type = base.substr(0,suffix);
    ActionType t(type);
    delete[] c_path;
    return importJsonFile(path,t);
}

bool Actions::addActionDef(ActionType t, const Json::Value &value)
{
    bool ok=true;
    ActionDefinition *act = jsonToActionDefinition(t,value);
    if(!act){
        ok=false;
    }
    if(ok){
        this->addActionDef(act);
    }
    return ok;
}

void Actions::addActionDef(ActionDefinition *action)
{
    this->_actions[action->actionType()] = action;
}

bool Actions::save()
{
    bool ok(true);
    for (std::map<ActionType,ActionDefinition*>::const_iterator it=_actions.begin();it!=_actions.end();++it){
        Json::Value root;
        bool it_ok= appendActionToJson(root,*it->second);
        if (it_ok){
            std::string path = getPath(it->first);
            //first of all, load to not overwrite
            std::ifstream ifs(path.c_str());
            std::ofstream ofs;
            if(ifs){
                if(!importJsonFile(path)){
                    M3D_ERROR("not writting to "<<path<<" because the existing file couldn't be parsed, don't want to lose data. Delete it if you want to.");
                    it_ok=false;
                }
            }
            if(it_ok){
                ofs.open(path.c_str());
                if (!ofs){
                    it_ok=false;
                }
            }
            if(it_ok){
                ofs<<root;
                ofs.close();
            }
        }
        ok= it_ok && ok;
    }
    return ok;
}

bool Actions::appendActionToJson(Json::Value &root, const ActionDefinition &act)
{
    bool ok(true);
    for ( std::vector<PlanningType>::const_iterator it  =act.requiredParts().begin();it!=act.requiredParts().end();++it){
        root["planning_parts"]["required"].append(Json::Value(it->toString()));
    }
    for ( std::vector<PlanningType>::const_iterator it  =act.hriParts().begin();it!=act.hriParts().end();++it){
        root["planning_parts"]["hri"].append(Json::Value(it->toString()));
    }
    if(act.include().size() && act.exclude().size()){
        M3D_ERROR("cannot define both include and exclude lists for action definition");
        ok=false;
    }
    if(ok){
        for ( std::set<AgentType>::const_iterator it  =act.include().begin();it!=act.include().end();++it){
            std::string s;
            if(it->isDefined()){
                root["include"].append(Json::Value(it->toString()));
            }else{
                M3D_ERROR("no string defined for agent type "<<it->toHriAgentType());
                ok=false;
            }
        }
        for ( std::set<AgentType>::const_iterator it  =act.exclude().begin();it!=act.exclude().end();++it){
            std::string s;
            if(it->isDefined()){
                root["exclude"].append(Json::Value(it->toString()));
            }else{
                M3D_ERROR("no string defined for agent type "<<it->toHriAgentType());
                ok=false;
            }
        }
    }

    return ok;
}

ActionDefinition* Actions::jsonToActionDefinition(ActionType t, const Json::Value &value)
{
    bool ok =true;
    ActionDefinition *act = 0;
    std::vector<PlanningType> req;
    std::vector<PlanningType> hri;
    act = new ActionDefinition(t);


    if (!value.isMember("planning_parts")){
        M3D_ERROR("No planning parts defined for action type "<<t.toString()<<"in json file");
        ok = false;
    }
    if(ok){
        Json::Value const & planparts_val = value["planning_parts"];
        for (Json::Value::const_iterator it = planparts_val.begin();
             it != planparts_val.end();
             ++it)
        {
            std::vector<PlanningType> *the_vect;
            if (it.key().asString() == "required"){
                the_vect = & req;
            }else if (it.key().asString() == "hri"){
                the_vect= &hri;
            }
            if(the_vect){
                for (Json::Value::const_iterator it2=(*it).begin();
                     it2!= (*it).end();
                     ++it2)
                {
                    the_vect->push_back(PlanningType((*it2).asString()));
                }
            }
        }

        if(value.isMember("cost_settings")){
            value["cost_settings"].asString();
            M3D_ERROR ("cost_settings not implemented yet");
            //TODO cost settings for action <-> hrics
        }

        if(value.isMember("include") && value.isMember("exclude")){
            if(value["include"].size() && value["exclude"].size()){
                M3D_ERROR("cannot define both include and exclude lists for action definition for "<<t.toString());
                ok=false;
            }
        }
        if(ok && value.isMember("include")){
            const Json::Value &include=value["include"];
            for(Json::Value::const_iterator it = include.begin();it!=include.end();++it){
                AgentType type((*it).asString());
                if(type.isDefined()){
                    act->include().insert(type);
                }else{
                    M3D_WARN("found unknown agent type "<<(*it).asString()<<" in \"include\" field in json action file");
                }
            }
        }
        if(ok && value.isMember("exclude")){
            const Json::Value &exclude=value["exclude"];
            for(Json::Value::const_iterator it = exclude.begin();it!=exclude.end();++it){
                AgentType type((*it).asString());
                if(type.isDefined()){
                    act->exclude().insert(type);
                }else{
                    M3D_WARN("found unknown agent type "<<(*it).asString()<<" in \"exclude\" field in json action file");
                }
            }
        }
    }

    if(ok){
        act->setRequiredParts(req);
        act->setHriParts(hri);
    }else{
        delete act;
        act=0;
    }
    return act;
}

bool Actions::proactive() const
{
    return _proactive;
}

void Actions::setProactive(bool proactive)
{
    _proactive = proactive;
    if(_proactive){
        this->fetchData();
    }
}



} // namespace Sem

