#include "Sem_ActionsManager.hpp"

namespace Sem {

INIT_MOVE3D_STATIC_LOGGER(ActionsManager,"planners.semantics.actionsmanager");

ActionsManager::ActionsManager(Actions *actions, PlanParts *planparts)
{
    setActions(actions);
    setPlanParts(planparts);
}

ActionsManager::~ActionsManager()
{

}

bool ActionsManager::isActionPossible(ActionType act_type, HRI_AGENT_TYPE agent_type,std::string &report)
{
    bool possible=true;
    std::ostringstream oss_report;
    report="ok";

    const ActionDefinition *act = actions()->getActionDefinition(act_type);
    if(!act){
        oss_report<<"action "<<act_type.toString()<<" not defined";
        possible=false;
    }
    if(possible){
        if (!act->isWhiteListed(agent_type)){
            possible=false;
            oss_report << "action \""<<act_type.toString()<<"\" is marked as not feasible by agents of type "<< AgentType::agentTypeToHumanString(agent_type);
        }
    }
    if(possible){
        //check required planning parts
        //the action is possible if all plan parts are non-empty
        for (std::vector<PlanningType>::const_iterator it=act->requiredParts().begin(); it!=act->requiredParts().end();++it){
            std::vector<PlanningPart*> jsets=planParts()->getPlanningParts(agent_type,*it);
            if (possible){
                if (jsets.size()==0){
                    possible = false;
                    oss_report << "agent type "<<  AgentType::agentTypeToHumanString(agent_type)
                               << " has no plan part for planning type "<<it->toString()
                               << " required for the action "<<act_type.toString();
                }
            }
        }
    }
    if(!possible){
        report = oss_report.str();
        M3D_INFO(report);
    }
    assert ( possible == (report == "ok"));
    return possible;

}

std::vector<PlanningPart> ActionsManager::getJointsForAction(ActionType act_type, HRI_AGENT_TYPE agent_type, bool is_hri)
{
    std::set<PlanningType> planTypeSet;

    const ActionDefinition* def = actions()->getActionDefinition(act_type);
    assert(def->actionType() == act_type);
    planTypeSet.insert(def->requiredParts().begin(),def->requiredParts().end());
    if(is_hri){
        M3D_DEBUG("hriParts size: " << def->hriParts().size() );
        for (unsigned int i=0;i<def->hriParts().size() ; i++){
            planTypeSet.insert(def->hriParts()[i]);
        }
        //planTypeSet.insert(def->hriParts().begin(),def->hriParts().end());
    }

    std::vector<PlanningPart> parts_buff, parts_prev;
    parts_prev.push_back(PlanningPart());
    parts_prev.back().setAgentType(agent_type);
    unsigned int nb_comb(1);
    //create the PlanningParts from (all) the combinations of PlanningParts needed for the action
    /* parts_prev = [ empty PlanningPart ];
     * for each t in planning types
     *   for each prev in parts_prev
     *      for each new in planningParts(agent,t)
     *         parts_buff.append(prev+new)
     *   parts_prev = parts_buff
     * return parts_prev
     */
    for (std::set<PlanningType>::const_iterator it=planTypeSet.begin(); it != planTypeSet.end(); ++it){
        std::vector<PlanningPart*> parts_new = planParts()->getPlanningParts(agent_type,*it);
        if(parts_new.size()>0){
            nb_comb*=parts_new.size();
            for(std::vector<PlanningPart>::const_iterator prev = parts_prev.begin();prev != parts_prev.end();prev++){
                for(std::vector<PlanningPart*>::const_iterator newp=parts_new.begin();newp!=parts_new.end();newp++){
                    parts_buff.push_back(PlanningPart(*prev));
                    parts_buff.back().concat(**newp);
                }
            }
            parts_prev.clear();
            std::swap(parts_prev,parts_buff);
        }
        //now parts_buff is empty, parts_prev contains the data
    }

    assert(parts_prev.size() == nb_comb);


    return parts_prev;
}
Actions *ActionsManager::actions() const
{
    return _actions;
}

void ActionsManager::setActions(Actions *actions)
{
    _actions = actions;
}
PlanParts *ActionsManager::planParts() const
{
    return _planParts;
}

void ActionsManager::setPlanParts(PlanParts *planParts)
{
    _planParts = planParts;
}



} // namespace Sem

