#ifndef SEM_ACTIONSMANAGER_H
#define SEM_ACTIONSMANAGER_H

#include "Logging/Logger.h"
#include "database/semantics/Sem_Actions.hpp"
#include "database/semantics/Sem_PlanParts.hpp"
#include "database/semantics/Sem_AgentType.hpp"


namespace Sem {

/**
 * @brief The ActionsManager class
 *
 * @example
 *     std::string sem_repport;
 *     if(!Sem::SemanticsManager::mainSemanticsManager()->actionsManager()->isActionPossible(Sem::ActionType("pick"),mainAgent->getHriAgent()->type,sem_repport)){
 *         cout << sem_repport<<endl;
 *         return false;
 *     }
 *
 *     Json::Value json;
 *     std::vector<Sem::PlanningPart> planParts =
 *             Sem::SemanticsManager::mainSemanticsManager()->actionsManager()->getJointsForAction(Sem::ActionType("pick"),
 *                                                                                                 mainAgent->getHriAgent()->type,
 *                                                                                                 false);
 *     for (std::vector<Sem::PlanningPart>::const_iterator it = planParts.begin(); it!= planParts.end();it++){
 *         ...
 *     }
 *
 */
class ActionsManager
{
    MOVE3D_STATIC_LOGGER;
public:
    ActionsManager(Actions *actions, PlanParts *planparts);
    ~ActionsManager();

    /**
     * @brief ActionsManager::isActionPossible
     * @param act_type[in]
     * @param agent_type[in]
     * @param report[out]
     * @return
     *
     * possible reports (and reasons why it returns false):
     * - "action <action type> not defined" : no definition found for this action (in json files)
     * - "action <action type> is marked as not feasible by agents of type <agent type>" : the agent type is in the exlcude list,
     *   or not in the include list, and one of them exists;
     * - "agent type <agent type> has no plan part for planning type <planning type> required for the action <action type>" : there is a planning part required by
     *   the action that is not defined for the agent type.
     * - "ok" : if and only if return true
     */
    bool isActionPossible(ActionType act_type, HRI_AGENT_TYPE agent_type, std::string &report);

    /**
     * @brief getJointsForAction uses Actions and PlanParts to retrieve set of joints to use.
     * @param act_type
     * @param agent_type
     * @param is_hri
     * @return a vector of PlanningPart, each of them is an alternative (e.g. left arm/right arm).
     */
    std::vector<PlanningPart> getJointsForAction(ActionType act_type, HRI_AGENT_TYPE agent_type, bool is_hri);



    Actions *actions() const;
    void setActions(Actions *actions);

    PlanParts *planParts() const;
    void setPlanParts(PlanParts *planParts);

private:
    Actions *_actions;
    PlanParts *_planParts;
};

} // namespace Sem

#endif // SEM_ACTIONSMANAGER_H
