/*
 * TreePlanner.cpp
 *
 *  Created on: Aug 31, 2009
 *      Author: jmainpri
 */

#include "TreePlanner.hpp"

#include "planner/cost_space.hpp"

#include "API/Grids/PointCloud.hpp"

#include "API/Roadmap/node.hpp"
#include "API/Roadmap/compco.hpp"
#include "API/Roadmap/graph.hpp"

#include "../p3d/env.hpp"
#include "planEnvironment.hpp"
#include "ReplanningSimulator.hpp"

#include "Planner-pkg.h"
#include "Collision-pkg.h"

using namespace std;
using namespace tr1;
using namespace API;


/**
 * Actions to be done after the expansion attempted from the given node, in case of success.
 */
void TreeExpansion::expansionSucceeded(Node * node)
{
	// re-initialize the number of consecutive expansion failures of the node
	node->getNodeStruct()->n_fail_extend = 0;
}


/**
 * Actions to be done after the expansion attempted from the given node, in case of failure.
 */
void TreeExpansion::expansionFailed(Node * node)
{
	// increment the number of consecutive expansion failures of the node
	node->getNodeStruct()->n_fail_extend++;

	// check whether the node should be discarded
	if ((ENV.getBool(Env::discardNodes)) && (node->getNodeStruct()->n_fail_extend > m_MaxExpandNodeFailure))
		node->getNodeStruct()->IsDiscarded = true;
}


/*!
 * Constructor
 */
TreePlanner::TreePlanner(Robot* R, Graph* G) :
Planner(R,G),
m_nbConscutiveFailures(0),
m_nbExpansion(0),
m_nbFailedExpansion(0)
{}

/*!
 * Destructor
 */
TreePlanner::~TreePlanner()
{}

unsigned TreePlanner::init()
{
	Planner::init();
	m_nbConscutiveFailures = 0;
	m_nbExpansion = 0;
	m_nbInitNodes = _Graph->getNumberOfNodes();
	return m_nbInitNodes;
}

/**
 * Time of day chrono
 */
double TreePlanner::getTime()
{
  double time=0.0;
  ChronoTimeOfDayTimes(&time);
  return time;
}

/*!
 * Checks out that the plannification
 * problem fits such requirement
 */
bool TreePlanner::preConditions()
{
	if (!_Robot->setAndUpdate(*_Start->getConfiguration())) {
		cout << "TreePlanner::preConditions => Start config does not respect kin. constraints" << endl;
		return false;
	}

	if (_Start->getConfiguration()->isOutOfBounds()) {
		cout << "TreePlanner::preConditions => Start is out of bounds" << endl;
		return false;
	}

	if (_Start->getConfiguration()->isInCollision()) {
		cout << "TreePlanner::preConditions => Start in collision" << endl;
		return false;
	}

	if (ENV.getBool(Env::biDir))
	{
		if (*_Start->getConfiguration() == *_Goal->getConfiguration()) {
			cout << "TreePlanner::preConditions => Start and goal are the same" << endl;
			return false;
		}

		if (!_Graph->searchConf(*_Start->getConfiguration())) {
			cout << "TreePlanner::preConditions => Config. start not in graph" << endl;
			return false;
		}

		if (!_Graph->searchConf(*_Goal->getConfiguration())) {
			cout << "TreePlanner::preConditions => Config. goal not in graph" << endl;
			return false;
		}
	}

	if (ENV.getBool(Env::expandToGoal))
	{
		if (_Goal->getConfiguration()->isOutOfBounds()) {
			cout << "TreePlanner::preConditions => Goal is out of bounds" << endl;
			return false;
		}

		if (!_Robot->setAndUpdate(*_Goal->getConfiguration())) {
			cout << "TreePlanner::preConditions => Goal config does not respect kin. constraints" << endl;
			return false;
		}

		if (_Goal->getConfiguration()->isInCollision()) {
			cout << "TreePlanner::preConditions => Goal in collision" << endl;
			return false;
		}
	}

	if (ENV.getBool(Env::drawPoints)) {
		PointsToDraw = new PointCloud;
	}

	return true;
}


/**
 * Check the stopping conditions of the Tree Planner.
 * @return TRUE if one of the conditions is satisfied, and FALSE otherwise
 */
bool TreePlanner::checkStopConditions()
{
    if (PlanEnv->getBool(PlanParam::stopPlanner)) {
        cout << "Stop: planning canceled by user." << endl;
        return true;
    }

    if (PlanEnv->getBool(PlanParam::planWithTimeLimit) &&
            m_time >= PlanEnv->getDouble(PlanParam::timeLimitPlanning)) {
        cout << "Stop: the time limit has been reached." << endl;
        return true;
    }

    if (m_nbExpansion >= (unsigned) PlanEnv->getInt(PlanParam::plannerMaxIterations)) {
        cout << "Stop: the maximum number of expansions has been reached." << endl;
        return true;
    }

    if (m_nbConscutiveFailures >= (unsigned) ENV.getInt(Env::NbTry)) {
        cout << "Stop: the maximum number of consecutive failures has been reached." << endl;
        return true;
    }

    if (_Graph->getNumberOfNodes() >= (unsigned) ENV.getInt(Env::maxNodeCompco)) {
        cout << "Stop: the maximum number of nodes has been reached." << endl;
        return true;
    }

    return false;
}


/*!
 * Tries to connect one node from
 * one component to the other
 */
bool TreePlanner::connectNodeToComp(Node* node, Node* compNode)
{
    return _Graph->linkNodeAndMerge(node, compNode);
}

/*!
 * Main function to connect to the other component
 */
bool TreePlanner::connectionToTheOtherCompco(Node* toNode)
{
	if( ENV.getBool(Env::tryClosest))
	{
		bool WeigtedRot = ENV.getBool(Env::isWeightedRotation);
		ENV.setBool(Env::isWeightedRotation,false);
		
		Node* closestNode = _Graph->nearestWeightNeighbour(toNode, _Graph->getLastNode()->getConfiguration(),
                false, ENV.getInt(Env::DistConfigChoice));
		
		ENV.setBool(Env::isWeightedRotation,WeigtedRot);
		
		return connectNodeToComp(_Graph->getLastNode(), closestNode );
	}
	else if(ENV.getBool(Env::randomConnectionToGoal))
	{
	    Node* randomNode = toNode->getConnectedComponent()->randomNode();
		return connectNodeToComp(_Graph->getLastNode(), randomNode);
	}
	else
		return connectNodeToComp( _Graph->getLastNode(), toNode );
}

const int nb_Graph_Saved = 100;
int ith_Graph_Saved = 0;

//! Main Function of the Tree Planner,
//! Bi-Directionality is handled here
unsigned int TreePlanner::run()
{
	shared_ptr<Configuration> tmp = _Robot->getCurrentPos();

    if (!preConditions()) {
        m_fail = true;
        cout << "Stoped in Tree planner, pre condition" << endl;
        return 0;
    }

	unsigned int NbCurCreatedNodes = 0;
    unsigned int NbTotCreatedNodes = 0;

    Node* fromNode = _Start;
    Node* toNode = _Goal;

    double ts(0.0);
    bool first_iteration(true);
    bool connected(false);

    if (!globalRePlanningEnv) {
        m_time = 0.0;
        ChronoOn();
    } else {
        // Stores the init time in ts on first call
        m_time = globalRePlanningEnv->timeSinceLastCall(first_iteration, ts);
    }

	while (!checkStopConditions())
	{
		ENV.setInt(Env::progress,
                (int) ((double) _Graph->getNumberOfNodes() / (double) ENV.getInt(Env::maxNodeCompco)));

        // Do not expand in the case of a balanced bidirectional expansion, when components are unbalanced
        if (!ENV.getBool(Env::expandToGoal)
                || !((ENV.getBool(Env::biDir) && ENV.getBool(Env::expandBalanced))
                        && (fromNode->getConnectedComponent()->getNumberOfNodes()
                                > toNode->getConnectedComponent()->getNumberOfNodes() + 2)))
        {
            // expand one way one time (Main function of Tree like planners)
            NbCurCreatedNodes = expandOneStep(fromNode);
            ++m_nbExpansion;

            if (NbCurCreatedNodes > 0)
            {
                NbTotCreatedNodes += NbCurCreatedNodes;
                m_nbConscutiveFailures = 0;

                // If it expands towards a goal, tries to link with local method
                if (!connected && ENV.getBool(Env::expandToGoal) && connectionToTheOtherCompco(toNode))
                    connected = true;

                if (!ENV.getBool(Env::drawDisabled))
                {
                    if (connected && PlanEnv->getBool(PlanParam::rrtExtractShortestPath) &&
                            ENV.getBool(Env::drawTraj))
                    {
                        Trajectory* traj = _Graph->extractAStarShortestPath(_q_start, _q_goal);
                        if (traj)
                        {
                            _draw_func();
                            delete traj;
                        }
                    }
                    else if (ENV.getBool(Env::drawExploration))
                        _draw_func();
                }
            } else
                m_nbConscutiveFailures++;
        }

		if (ENV.getBool(Env::biDir))
            swap(fromNode, toNode);

		if (globalRePlanningEnv)
            m_time += globalRePlanningEnv->timeSinceLastCall(first_iteration, ts);
        else
            ChronoTimes(&m_time, &ts);
	}

    if (!globalRePlanningEnv)
        ChronoOff();

    if ((!ENV.getBool(Env::drawDisabled))
            && (ENV.getBool(Env::drawExploration) || ENV.getBool(Env::drawTraj)))
        _draw_func();

    ENV.setInt(Env::nbQRand, m_nbExpansion);
    _Robot->setAndUpdate(*tmp);
    return (NbTotCreatedNodes);
}

/**
 * Decide whether the goal bias should be applied, based on a random trial.
 * @return TRUE if the goal bias is applied, and FALSE otherwise
 */
bool TreePlanner::goalBias() const
{
	return (p3d_random(0., 1.) < ENV.getDouble(Env::Bias));
}
