/*
 * BaseExpansion.cpp
 *
 *  Created on: Jun 12, 2009
 *      Author: jmainpri
 */


#include "BaseExpansion.hpp"

#include "API/planningAPI.hpp"

#include "Roadmap/compco.hpp"

#include "P3d-pkg.h"
#include "Planner-pkg.h"

using namespace std;
using namespace tr1;

BaseExpansion::BaseExpansion() :
m_ExpansionNodeMethod(NEAREST_EXP_NODE_METH),
m_MaxExpandNodeFailure(10),
m_kNearestPercent(10),
m_ExpansionDirectionMethod(GLOBAL_CS_EXP),
m_IsDirSampleWithRlg(false),
m_fromComp(NULL),
m_toComp(NULL)
{ 
	cout << "no graph in expansion method" << endl; 
    _nbConnectibleNodes=0;
}

BaseExpansion::BaseExpansion(Graph* ptrGraph) :

m_ExpansionNodeMethod(NEAREST_EXP_NODE_METH),
m_MaxExpandNodeFailure(10),
m_kNearestPercent(10),
m_ExpansionDirectionMethod(GLOBAL_CS_EXP),
m_IsDirSampleWithRlg(false),
m_Graph(ptrGraph),
m_fromComp(NULL),
m_toComp(NULL)
{
_nbConnectibleNodes=0;
}

BaseExpansion::~BaseExpansion(){}

/**
 * Returns a distance define by the environement
 */
double BaseExpansion::step()
{
	return(p3d_get_env_dmax() * ENV.getDouble(Env::extensionStep));
}

double BaseExpansion::getDMax()
{
	return p3d_get_env_dmax();
}

//! Interpolation along the given path.
confPtr_t BaseExpansion::interpolate(confPtr_t conf, LocalPath * path, double factor)
{
	return (factor == 1.) ? conf : path->configAtParam(factor * path->getParamMax());
}

double BaseExpansion::positionAlongPath(LocalPath& path, double param)
{
	return					path.getParamMax() <= 0. ? 1. : 
	MIN(1., param / path.getParamMax() );
}

/**
 * Return the path parameter that is 
 * the closest to step()
 */
double BaseExpansion::pathDelta(LocalPath& path)
{
	double pathDeta =	path.getParamMax() == 0. ? 1 : 
	MIN(1., step() /  path.getParamMax() );	
	
	return pathDeta;
}

/**
 * Returns the localpath of length
 * equal or inferior to step
 */
LocalPath BaseExpansion::getExtensiontPath( confPtr_t qi, confPtr_t qf )
{
	LocalPath directionLocalpath(qi,qf);
	
	double pathDelta = directionLocalpath.getParamMax() == 0. ? 1 
	: MIN(1., step() / directionLocalpath.getParamMax() );
	
	LocalPath extensionLocalpath( directionLocalpath, pathDelta , false );
	return extensionLocalpath;
}

/**
 * Function that balances the ratio of exploration towards refinement
 */
bool BaseExpansion::refinementControl(LocalPath& path, double rho, Node& compNode)
{
	double radius = 0;

	if(ENV.getBool(Env::useRefiRadius))
	    radius = ENV.getDouble(Env::refiRadius);
	else
	    radius = step();

	if (ENV.getBool(Env::printRadius))
	{
        cout << "radius = " << radius << endl;
        cout << "path.length() = " << path.getParamMax() << endl;
        double ratio = (double) compNode.getConnectedComponent()->getCompcoStruct()->nbRefinNodes
                / (double) compNode.getConnectedComponent()->getNumberOfNodes();
        cout << "ratio of RNODES = " << ratio << endl;
        cout << endl;
    }

	// in the case of a refinement node
	if (path.getParamMax() <= radius)
    {
        if ((double) compNode.getConnectedComponent()->getCompcoStruct()->nbRefinNodes
                > rho * compNode.getConnectedComponent()->getNumberOfNodes())
            return false;
        else
            compNode.getConnectedComponent()->getCompcoStruct()->nbRefinNodes++;
    }

	return true;
}

/**
 * Function called when an expansion fails.
 */
void BaseExpansion::expansionFailed(Node& node) 
{
	if(m_ExpansionNodeMethod == RANDOM_IN_SHELL_METH)  
		p3d_SetNGood(0);

	node.getNodeStruct()->n_fail_extend++;

	if((ENV.getBool(Env::discardNodes)) &&
		 (node.getNodeStruct() != m_Graph->getGraphStruct()->search_start) &&
		 (node.getNodeStruct() != m_Graph->getGraphStruct()->search_goal) &&
		 (node.getNodeStruct()->n_fail_extend > m_MaxExpandNodeFailure)) 
	{
		node.getNodeStruct()->IsDiscarded = true;
		update_parent_nfails(node.getNodeStruct());

		m_Graph->getGraphStruct()->n_consec_pb_level ++;
		m_Graph->getGraphStruct()->n_consec_fail_pb_level ++;
	}
}


/**
 * Gives successive co
 */
bool BaseExpansion::nextStep(LocalPath& path,
														 Node* directionNode,
														 double& pathDelta,
														 shared_ptr<LocalPath>& newPath,
														 Env::expansionMethod method)
{
	if( method == Env::Connect )
	{
		//cout << "create path satisfying connect method" << endl;
		newPath = shared_ptr<LocalPath>(new LocalPath(path, pathDelta));
		
		if(pathDelta == 0.)
		{ 
			return false; 
		}
		else 
		{
			return true;
		}
	}
	else
	{
		pathDelta = path.getParamMax() == 0. ? 1. : MIN(1., step() / path.getParamMax());
		shared_ptr<Configuration> ptrEnd;
		
		if(pathDelta == 1. && directionNode)
		{
			ptrEnd = directionNode->getConfiguration();
		}
		else
		{
			ptrEnd = path.configAtParam(pathDelta * path.getLocalpathStruct()->range_param);
		}
		ptrEnd->setConstraints();
		newPath = shared_ptr<LocalPath>(new LocalPath(path.getBegin(),ptrEnd));
	}

    return(newPath->isValid());
}

/**
 * Gives successive co
 */
bool BaseExpansion::nextStep(LocalPath& path,
														 shared_ptr<Configuration>& directionConfig,
														 double& pathDelta,
														 shared_ptr<LocalPath>& newPath,
														 Env::expansionMethod method)
{
	
	if(method == Env::Connect)
	{
		cout << "Error : Method Doesn't work with connect" << endl;
	}
	else
	{
		pathDelta = path.getParamMax() <= 0. ? 1. : MIN(1., this->step() / path.getParamMax());
		
		newPath = shared_ptr<LocalPath>(
																		new LocalPath(path.getBegin(),
																									pathDelta == 1. && directionConfig ?
																									directionConfig :
																									path.configAtParam(pathDelta * path.getLocalpathStruct()->range_param)));
	}
	
	return(newPath->isValid());
}

/**
 * Function that adds a node to the graph
 */
Node* BaseExpansion::addNode(Node* currentNode, LocalPath& path, double pathDelta, Node* directionNode,
        unsigned& nbCreatedNodes)
{
    currentNode->getNodeStruct()->n_fail_extend = 0;

	if (pathDelta == 1. && directionNode)
	{
	    m_Graph->linkNodeAndMerge(currentNode, directionNode);
        return (directionNode);
	}
	else
	{
		nbCreatedNodes++;
		return (m_Graph->insertNode(currentNode, path));
	}
}


Node* BaseExpansion::getExpansionNode(Node* compNode, confPtr_t direction, int distance)
{
    if (p3d_GetCostMethodChoice() == MONTE_CARLO_SEARCH)
    {
        return m_Graph->getNode(compNode->getConnectedComponent()->getCompcoStruct()->dist_nodes->N);
    }

    int K;
    vector<Node*> KNearestNeighbors;

    switch (distance)
    {

    case NEAREST_EXP_NODE_METH:
        // choose the nearest node of the connected component
        return (m_Graph->nearestWeightNeighbour(compNode, direction, p3d_GetIsWeightedChoice(), distance));

    case K_NEAREST_EXP_NODE_METH:
        // randomly select one of the K nearest nodes of the connected component
        K = ceil(compNode->getConnectedComponent()->getNumberOfNodes() / 100.);
        if (K == 1)
            return m_Graph->nearestWeightNeighbour(compNode, direction, p3d_GetIsWeightedChoice(), distance);
        else {
            KNearestNeighbors = compNode->getConnectedComponent()->KNearestWeightNeighbour(direction, K,
                    numeric_limits<double>::max(), p3d_GetIsWeightedChoice(), distance);
            unsigned index = (int) floor(p3d_random(0.0, (double) KNearestNeighbors.size() - EPS6));
            return KNearestNeighbors[index];
        }

    case BEST_SCORE_EXP_METH:
        /* Select the node which has the best score: weight*dist */
        return (m_Graph->nearestWeightNeighbour(compNode, direction, p3d_GetIsWeightedChoice(), distance));

    case K_BEST_SCORE_EXP_METH: //The same as BEST_SCORE_EXP_METH : something as to be done here
        /*int NearestPercent = m_kNearestPercent;
        int KNearest = MAX(1,(int)((NearestPercent*(compNode->getConnectedComponent()->getNumberOfNodes()))/100.));*/
        return (m_Graph->nearestWeightNeighbour(compNode, direction, p3d_GetIsWeightedChoice(), distance));

    case RANDOM_IN_SHELL_METH:
        /* Select randomly among all the nodes inside a given portion of shell */
        return (m_Graph->getNode(
                hrm_selected_pb_node(m_Graph->getGraphStruct(), direction->getConfigStruct(),
                        compNode->getNodeStruct()->comp)));

    case RANDOM_NODE_METH:
        return (m_Graph->getNode(p3d_RandomNodeFromComp(compNode->getConnectedComponent()->getCompcoStruct())));

    default:
        /* By default return the nearest node of the component */
        return (m_Graph->nearestWeightNeighbour(compNode, direction, p3d_GetIsWeightedChoice(), distance));
    }
}


shared_ptr<Configuration> BaseExpansion::getExpansionDirection(Node* expandComp, Node* goalComp,
        bool samplePassive, Node*& directionNode)
{
    if (p3d_GetCostMethodChoice() == MONTE_CARLO_SEARCH)
    {
        confPtr_t q = m_Graph->getRobot()->shootDir(samplePassive);

        p3d_addConfig(m_Graph->getRobot()->getRobotStruct(),
                      q->getConfigStruct(),
                      expandComp->getConnectedComponent()->getCompcoStruct()->dist_nodes->N->q,
                      q->getConfigStruct());

        return q;
    }

    shared_ptr<Configuration> q;
    int savedRlg;

    if (m_IsDirSampleWithRlg)
    {
        // Save the previous Rlg setting to shoot without Rlg
        savedRlg = p3d_get_RLG();
        p3d_set_RLG(false);
    }

    // Selection in the entire CSpace and
    // biased to the Comp of the goal configuration
    if (ENV.getBool(Env::isGoalBiased) && p3d_random(0., 1.) <= ENV.getDouble(Env::Bias))
    {
        // select randomly a node in the goal component as direction of expansion
        directionNode = m_Graph->randomNodeFromComp(goalComp);
        q = directionNode->getConfiguration();
    }
    else
    {
        switch (m_ExpansionDirectionMethod)
        {
            case SUBREGION_CS_EXP:
                // Selection in a subregion of the CSpace
                // (typically close to the current tree)
                // and  biased to the goal configuration
                q = shared_ptr<Configuration> (
                        new Configuration(m_Graph->getRobot()));

#ifdef P3D_PLANNER
                p3d_shoot_inside_box(m_Graph->getRobot()->getRobotStruct(),
                                     q->getConfigStruct(),
																		 expandComp->getConnectedComponent()->getCompcoStruct()->box_env_small,
                                     (int) samplePassive);
#else
						printf("P3D_PLANNER not compiled in %s in %s",__func__,__FILE__);
#endif

            break;
#ifdef LIGHT_PLANNER
			case NAVIGATION_BEFORE_MANIPULATION:
			{
				m_Graph->getRobot()->deactivateCcConstraint();

				q = m_Graph->getRobot()->shoot();
				Node* closest = m_Graph->nearestWeightNeighbour(expandComp,q,false,ONLY_ROBOT_BASE);

				LocalPath path(closest->getConfiguration(),q);
//				cout << "Param max = " << path.getParamMax() << " , step = " << step() << endl;
				q = path.configAtParam(2*step());

//				cout << "1 :" << endl;
//				q->print();

				m_Graph->getRobot()->setAndUpdate(*q);
				q = m_Graph->getRobot()->shootAllExceptBase();
				q->setConstraintsWithSideEffect();

//				cout << "2 :" << endl;
//				q->print();

//				m_Graph->getRobot()->setAndUpdate(*q);
//				g3d_draw_allwin_active();

				m_Graph->getRobot()->activateCcConstraint();
//				q->setConstraintsWithSideEffect();
//				m_Graph->getRobot()->setAndUpdate(*q);
//				g3d_draw_allwin_active();
			}
			break;
#endif
			case GLOBAL_CS_EXP:
          default:
            // Selection in the entire CSpace
            q = m_Graph->getRobot()->shoot(samplePassive);
        }
    }
    if (!m_IsDirSampleWithRlg)
    {
        //Restore the previous Rlg setting
        p3d_set_RLG(savedRlg);
    }
    return (q);
}
