/*
 * ManhattanLike-RRT.cpp
 *
 *  Created on: Aug 31, 2009
 *      Author: jmainpri
 */

#include "P3d-pkg.h"
#include "Collision-pkg.h"

#include "API/Roadmap/compco.hpp"
#include "API/Roadmap/graph.hpp"

#include "planner/planEnvironment.hpp"
#include "planner/Diffusion/Variants/ManhattanLike-RRT.hpp"

using namespace std;


/**
 * Perform a single expansion step of ML-RRT, growing the given connected component.
 * @return the number of created nodes
 */
unsigned ManhattanLikeRRT::expandOneStep(ConnectedComponent * fromComp)
{
	// sample the expansion direction (i.e. the configuration toward which the expansion will be performed)
	// taking into account the goal bias, if relevant
	Node * directionNode(NULL);
	confPtr_t directionConf;
	bool authorizedBias = !PlanEnv->getBool(PlanParam::anytime) || !this->trajFound();
	if (ENV.getBool(Env::isGoalBiased) && authorizedBias && (p3d_random(0., 1.) < ENV.getDouble(Env::Bias)))
	{
		// bias toward the goal configuration
		directionNode = _Goal;
		directionConf = _q_goal;
	}
	else
		// or random sampling in the entire C-Space
		directionConf = _Robot->shoot();

	// select, from the component to be grown, the node from which the expansion will be attempted
	Node * expandNode = ENV.getBool(Env::randomConnectionToGoal) ? fromComp->randomNode()
			: fromComp->nearestNeighbor(directionConf);

	// copy passive DoFs of q_near in q_rand
	expandNode->getConfiguration()->copyPassive(*directionConf);

	// directed expansion of the active DoFs
	unsigned nbCreatedNodes(0);
	switch (ENV.getExpansionMethod())
	{
	case Env::Extend:
		nbCreatedNodes = expansion->extend(expandNode, directionConf, directionNode);
		break;
	case Env::Connect:
		nbCreatedNodes = expansion->connect(expandNode, directionConf, directionNode);
		break;
	default:
		cerr << "*** ERROR: ManhattanLikeRRT::expandOneStep --> method not implemented !!! ***" << endl;
		nbCreatedNodes = 0;
	}

	// directed expansion of the passive DoFs
	nbCreatedNodes += this->passiveExpandProcess(nbCreatedNodes == 0 ? expandNode : _Graph->getLastNode(),
			directionNode);

	return nbCreatedNodes;
}


unsigned ManhattanLikeRRT::passiveExpandProcess(Node * expansionNode, Node * directionNode)
{
	unsigned totalCreatedNodes(0);
	bool expansionSucceeded(true);

	Configuration invalConf(_Robot);
	vector<p3d_jnt*> oldJoints;

	// Keep iterating while a collision is encountered during the connect expansion,
	// unless the last passive expansion failed.
	while (expansionSucceeded && p3d_get_current_q_inv(_Robot->getRobotStruct(), invalConf.getConfigStruct()))
	{
		vector<p3d_jnt*> joints;
		vector<p3d_jnt*> newJoints;
		expansionSucceeded = false;

		// Get the colliding passive dofs and select those that have not been expanded yet.
		if (getCollidingPassiveJntList(_Robot, invalConf, joints) &&
				selectNewJntInList(joints, oldJoints, newJoints))
		{
			// Create a copy of the expansion configuration.
			// This configuration will become the new target configuration,
			// after its passive dofs have been changed.
			confPtr_t newRandConf = expansionNode->getConfiguration()->copy();

			for (int i = 0; i < ENV.getInt(Env::MaxPassiveExpand) && !expansionSucceeded; i++)
			{
				shoot_jnt_list_and_copy_into_conf(*newRandConf, newJoints);
				unsigned nbCreatedNodes = expansion->connect(expansionNode, newRandConf, directionNode);
				expansionSucceeded = nbCreatedNodes > 0;

				// If the expansion succeeded, update the expansion node for the next iteration
				if (expansionSucceeded)
				{
					expansionNode = _Graph->getLastNode();
					totalCreatedNodes += nbCreatedNodes;
				}
			}
		}
	}

	return totalCreatedNodes;
}


bool ManhattanLikeRRT::getCollidingPassiveJntList(Robot * R, Configuration & qinv, vector<p3d_jnt *> & joints)
{
	p3d_poly * polys[2];

	R->setAndUpdateWithoutConstraints(qinv);
	if (p3d_col_test() <= 0) {
		cout << "No collision detected" << endl;
		return (false);
	}

	// NOTE: KCD only retuns the first collision pair !!!
	// NOTE: ONLY THE PASSIVE JOINT INVOLVED IN THE COLLISION IS RETURNED
	//       BUT ALL THE PARENT JOINTS SHOULD BE ALSO CONSIDERED ???
	p3d_col_get_report(0, &polys[0], &polys[1]);
	for (uint i(0); i < 2; i++)
		if (polys[i]->p3d_objPt->jnt != NULL)
			if (!p3d_jnt_get_is_active_for_planner(polys[i]->p3d_objPt->jnt))
				joints.push_back(polys[i]->p3d_objPt->jnt);

	return (joints.size() > 0);
}


bool ManhattanLikeRRT::selectNewJntInList(vector<p3d_jnt *> & joints, vector<p3d_jnt *> & oldJoints,
		vector<p3d_jnt *> & newJoints)
{
	for (uint i(0); i < joints.size(); i++)
	{
		bool found(false);
		for (uint j(0); j < oldJoints.size(); j++)
		{
			if (oldJoints[j] == joints[i])
			{
				found = true;
				break;
			}
		}
		if (!found)
		{
			newJoints.push_back(joints[i]);
			oldJoints.push_back(joints[i]);
		}
	}

	return (newJoints.size() > 0);
}


void ManhattanLikeRRT::shoot_jnt_list_and_copy_into_conf(Configuration & qrand, vector<p3d_jnt *> & joints)
{
	//double perturb = 0.1; // NOTE: THIS SHOULD BE A PARAMETER
	double perturb = 0.3; // NOTE: THIS SHOULD BE A PARAMETER

	// NOTE : the random shoot should be centered at q_inv !!!
	//        (doesn't matter in the case of "circular" joints)
	for (uint i(0); i < joints.size(); i++)
	{
		p3d_jnt* joint(joints[i]);
		for (int j(0); j < joint->dof_equiv_nbr; j++)
		{
			double vmin, vmax;
			double val, rval;
			p3d_jnt_get_dof_rand_bounds(joint, j, &vmin, &vmax);
			int k(joint->index_dof + j);
			if (!p3d_jnt_is_dof_circular(joint, j))
				val = p3d_random(vmin, vmax);
			else
			{
				double midrange = (vmax - vmin) / 2.0;
				// perturbation factor
				midrange *= perturb;
				rval = p3d_random(-midrange, midrange);
				val = qrand.getConfigStruct()[k] + rval;
				val = MAX(MIN(val, vmax), vmin);
			}
			qrand.getConfigStruct()[k] = val;
		}
	}
}

