/*
 * Multi-RRT.cpp
 *
 *  Created on: Nov 29, 2012
 *      Author: ddevaurs
 */

#include "API/Roadmap/compco.hpp"
#include "API/Roadmap/graph.hpp"

#include "planner/planEnvironment.hpp"
#include "planner/Diffusion/Variants/Multi-RRT.hpp"

using namespace std;
using namespace API;


/**
 * Main function of the Multi-RRT.
 * @return the number of nodes added to the graph
 */
unsigned MultiRRT::run()
{
	this->chronoStart();

	// check the pre-conditions
	if (!this->preConditions()) {
		cout << "Failure: pre-conditions are not satisfied." << endl;
		return 0;
	}

	unsigned compCounter(0); // index of the current component considered for expansion
	unsigned nbCreatedNodes(0); // number of nodes created by one expansion step
	unsigned totalCreatedNodes(0); // total number of nodes created so far
	bool unbalancedComp;

	// variable used to keep track of the cost of the best trajectory found so far, in the anytime variant
	double bestTrajCost = numeric_limits<double>::max();
	bool firstTraj(true);

    // Multi-RRT loop: start by checking the stopping conditions
	while (!this->checkStopConditions())
	{
		// choose the component that will be expanded
		vector<ConnectedComponent *> components = _Graph->getConnectedComponents();
		compCounter = (compCounter + 1) % components.size();

		// determine whether the components are balanced, when relevant
		if (ENV.getBool(Env::expandBalanced))
		{
			// compute the minimum number of nodes contained in a component
			unsigned minNbNodes = components[0]->getNumberOfNodes();
			for (unsigned i(1); i < components.size(); ++i)
				if (components[i]->getNumberOfNodes() < minNbNodes)
					minNbNodes = components[i]->getNumberOfNodes();

			unbalancedComp = (components[compCounter]->getNumberOfNodes() > (minNbNodes + 10));
		}

		// no expansion is performed when the connected components are unbalanced (when relevant)
		if (!ENV.getBool(Env::expandBalanced) || !unbalancedComp)
		{
			++m_nbExpansion;

			// perform a single expansion step of RRT
			nbCreatedNodes = this->expandOneStep(components[compCounter]);

			// success, if some nodes have been created
			if (nbCreatedNodes > 0)
			{
				totalCreatedNodes += nbCreatedNodes;
				m_nbConscutiveFailures = 0;

				Node * lastNode = _Graph->getLastNode();
				confPtr_t lastConf = lastNode->getConfiguration();

				// attempt a connection toward the closest component (when relevant)
				if (!PlanEnv->getBool(PlanParam::anytime) || !this->trajFound())
				{
					// look for the closest component
					Node * closestNode(NULL);
					double minDistance(0.);
					for (unsigned i(0); i < components.size(); ++i)
					{
						// ignore the expanded component
						if (i == compCounter)
							continue;

						// look for the node whose configuration is the closest to that of the last node
						Node * tempNode = components[i]->nearestNeighbor(lastConf);

						// check whether this node is closer than the one found so far (if there is one)
						double distance = lastConf->squareDistance(tempNode->getConfiguration().get());
						if (!closestNode || (distance < minDistance))
						{
							closestNode = tempNode;
							minDistance = distance;
						}
					}

					// attempt the expansion
					localPathPtr_t path(new LocalPath(lastConf, closestNode->getConfiguration()));
					if (expansion->isValidConnection(path))
						_Graph->addEdges(lastNode, closestNode, path);
				}

				// anytime variant
				bool drawNewTraj = false;
				if (PlanEnv->getBool(PlanParam::anytime) && this->trajFound())
				{
					bool extractNewTraj(true);

					// try to add cycles (when relevant)
					if (ENV.getBool(Env::addCycles))
                    {
                        extractNewTraj = (firstTraj || expansion->addCycles());
                    }
                    else if (ENV.getBool(Env::addCycles1))
                    {
                        extractNewTraj = (firstTraj || expansion->addCycles1());
                    }

                    // extract the new trajectory from the graph (when the display is enabled)
					if (extractNewTraj && !ENV.getBool(Env::drawDisabled))
					{
						Trajectory * trajectory = this->extractTrajectory();
						if (trajectory)
						{
							double trajCost = trajectory->cost();
							if (trajCost < bestTrajCost)
							{
								// keep track of the best trajectory found so far
								bestTrajCost = trajCost;
								trajectory->replaceP3dTraj();
								drawNewTraj = ENV.getBool(Env::drawTraj);
								cout << "best trajectory cost = " << bestTrajCost << endl;
							}
							delete trajectory;
							firstTraj = false;
						}
					}
				}

				// redraw the scene (when relevant)
				if (!ENV.getBool(Env::drawDisabled) && (ENV.getBool(Env::drawExploration) || drawNewTraj))
					_draw_func();
			}
			// failure, otherwise
			else {
				m_nbFailedExpansion++;
				m_nbConscutiveFailures++;
			}
		}
		this->chronoUpdate();
	}

	this->chronoStop();
	return totalCreatedNodes;
}


/**
 * Check whether the Multi-RRT has found a trajectory.
 * @return TRUE if the criteria required for a trajectory to be found are met, and FALSE otherwise
 */
bool MultiRRT::trajFound()
{
	// success if the graph contains a single connected component
	if (_Graph->getNumberOfCompco() == 1)
		return true;

	// success if all the waypoints belong to the same connected component
	unsigned i(0);
	bool connected(true);
	ConnectedComponent * component = waypointNodes[0]->getConnectedComponent();
	while (connected && i < waypointNodes.size())
		connected = (component == waypointNodes[++i]->getConnectedComponent());
	return connected;
}


/**
 * Load the user-defined waypoints (start, goal, and potential intermediate waypoints).
 */
void MultiRRT::loadWaypoints()
{
	// initialize the lists of waypoints
	waypoints.push_back(_q_start);
	waypointNodes.push_back(_Start);

	// load the additional user-defined configurations as waypoints (when relevant)
	for (unsigned i(0); i < _Robot->getNbConf(); ++i)
	{
		confPtr_t conf = _Robot->getConf(i);
		Node * node = new Node(_Graph, conf);
		_Graph->addNode(node);
		++m_nbInitNodes;

		waypoints.push_back(conf);
		waypointNodes.push_back(node);
	}

	// finalize the lists of waypoints
	waypoints.push_back(_q_goal);
	waypointNodes.push_back(_Goal);
}

/**
 * Extract a trajectory going from one waypoints to another one.
 */
Trajectory * MultiRRT::extractTrajectory(confPtr_t q_start, confPtr_t q_goal)
{
	return _Graph->extractAStarShortestPath(q_start, q_goal);
}

/**
 * Extract a trajectory going through all the waypoints.
 */
Trajectory * MultiRRT::extractTrajectory()
{
	// simple case: only two waypoints (start and goal)
	if (waypoints.size() == 2)
		return _Graph->extractAStarShortestPath(_q_start, _q_goal);

    // general case: multiple waypoints
	else if (waypoints.size() > 2)
	{
    	// create a container for the pairwise distance matrix of the waypoints
		vector<vector<double> > distances;
		for (unsigned i(0); i < waypoints.size(); ++i)
		{
			vector<double> distanceRow;
			for (unsigned j(0); j < waypoints.size(); ++j)
				distanceRow.push_back(0.);
			distances.push_back(distanceRow);
		}

    	// compute the pairwise distance matrix of the waypoints
		for (unsigned i(0); i < waypoints.size(); ++i)
		{
			for (unsigned j(i + 1); j < waypoints.size(); ++j)
			{
				Trajectory * trajectory = _Graph->extractAStarShortestPath(waypoints[i], waypoints[j]);
				if (trajectory)
				{
					distances[i][j] = trajectory->cost();
					delete trajectory;
				}
				if (ENV.getBool(Env::isCostSpace) && (ENV.getInt(Env::costDeltaMethod) == MECHANICAL_WORK))
				{
					trajectory = _Graph->extractAStarShortestPath(waypoints[j], waypoints[i]);
					if (trajectory)
					{
						distances[j][i] = trajectory->cost();
						delete trajectory;
					}
				}
				else
					distances[j][i] = distances[i][j];
			}
		}

    	// solve the TPS problem to find the shortest path
		vector<unsigned> shortestPath;
		this->TPSbyExhaustiveSearch(distances, shortestPath);

		// extract a trajectory going through all the waypoints, following the shortest path
		vector<unsigned>::iterator start = shortestPath.begin();
		vector<unsigned>::iterator goal = start;
		++goal;
		Trajectory * trajectory = _Graph->extractAStarShortestPath(waypoints[*start], waypoints[*goal]);
		for (unsigned i(2); i < waypoints.size(); ++i)
		{
			start = goal++;
			Trajectory * traj = _Graph->extractAStarShortestPath(waypoints[*start], waypoints[*goal]);
			if (!trajectory->concat(*traj))
				return NULL;
		}

		return trajectory;
    }

    // error: less than two waypoints are present
	else
		return NULL;
}


/**
 * Solve the Traveling Salesman Problem by exhaustive search.
 * @param distances: pairwise distance matrix given as input
 * @param solution: shortest path returned as output
 * @return length of the shortest path
 */
double MultiRRT::TPSbyExhaustiveSearch(vector<vector<double> > & distances, vector<unsigned> & solution)
{
	// initialization
	unsigned N = distances.size();
	unsigned permutation[N];
	for (unsigned i(0); i < N; ++i) {
		permutation[i] = i;
		solution.push_back(i);
	}
	double minLength = 0.;
	for (unsigned i(1); i < N; ++i)
		minLength += distances[i - 1][i];

	// go through each permutation and check whether it corresponds to a shorter path
	while (next_permutation(permutation, permutation + N))
	{
		double length = 0.;
		for (unsigned i(1); i < N; ++i)
			length += distances[permutation[i - 1]][permutation[i]];

		if (length < minLength) {
			minLength = length;
			for (unsigned i(0); i < N; ++i)
				solution[i] = permutation[i];
		}
	}
	return minLength;
}


/**
 * Solve the Traveling Salesman Problem using the Nearest Neighbor heuristic.
 * @param distances: pairwise distance matrix given as input
 * @param solution: shortest path returned as output
 * @return length of the shortest path
 */
double MultiRRT::TPSwithNearestNeighbor(vector<vector<double> > & distances, vector<unsigned> & solution)
{
	unsigned N = distances.size();
	double pathLengths[N];
	list<unsigned> paths[N];
	list<unsigned> candidates;
	list<unsigned>::iterator state;

	// for each waypoint used as a starting point,
	// compute the shortest path going through all the waypoints, starting from this waypoint
	for (unsigned i(0); i < N; ++i)
	{
		// initialize the path with the ith waypoint, and the list of candidates with the other waypoints
		pathLengths[i] = 0.;
		paths[i].push_back(i);
		for (unsigned j(0); j < N; ++j)
			if (j != i)
				candidates.push_back(j);

		// while there are at least two unvisited waypoints, look for the next one to visit
		while (candidates.size() > 1)
		{
			// look for the nearest neighbor of the last visited waypoint, among the candidates
			double minDistance = numeric_limits<double>::max();
			list<unsigned>::iterator nearestState;
			for (state = candidates.begin(); state != candidates.end(); ++state)
			{
				double distance = distances[paths[i].back()][*state];
				if (distance < minDistance) {
					minDistance = distance;
					nearestState = state;
				}
			}

			// add this nearest waypoint to the path, and remove it from the list of candidates
			pathLengths[i] += minDistance;
			paths[i].push_back(*nearestState);
			candidates.erase(nearestState);
		}

		// the last unvisited waypoint is added to the path
		pathLengths[i] += distances[paths[i].back()][candidates.front()];
		paths[i].push_back(candidates.front());
		candidates.clear();
	}

	// return the shortest path and its length
	unsigned shortestPathID = 0;
	for (unsigned i(1); i < N; ++i)
		if (pathLengths[i] < pathLengths[shortestPathID])
			shortestPathID = i;
	for (state = paths[shortestPathID].begin(); state != paths[shortestPathID].end(); ++state)
		solution.push_back(*state);

	return pathLengths[shortestPathID];
}


/**
 * Solve the Traveling Salesman Problem using the Multi-Fragment heuristic.
 * @param distances: pairwise distance matrix given as input
 * @param solution: shortest path returned as output
 * @return length of the shortest path
 */
double MultiRRT::TPSwithMultiFragments(vector<vector<double> > & distances, vector<unsigned> & solution)
{
	unsigned N = distances.size();
	list<unsigned>::iterator state1, state2, bestState1, bestState2;
	list<unsigned> edges[N];
	list<unsigned> candidates;
	for (unsigned i(0); i < N; ++i)
		candidates.push_back(i);

	// while there are at least three candidates, look for the shortest edge that do not create a cycle
	while (candidates.size() > 2)
	{
		double minDistance = numeric_limits<double>::max();

		// go through each pair of candidates
		for (state1 = candidates.begin(); state1 != candidates.end(); ++state1)
		{
			state2 = state1;
			++state2;
			for (; state2 != candidates.end(); ++state2)
			{
				if (distances[*state1][*state2] < minDistance)
				{
					// check for cycle
					bool done = edges[*state1].empty() || edges[*state2].empty();
					bool cycle = false;
					unsigned current = *state1;
					unsigned previous = current;
					while (!done && !cycle)
					{
						if (edges[current].size() == 1) {
							if (edges[current].front() == *state2)
								cycle = true;
							else if (edges[current].front() != previous) {
								previous = current;
								current = edges[current].front();
							}
							else
								done = true;
						}

						if (edges[current].size() == 2) {
							if (edges[current].front() == *state2 || edges[current].back() == *state2)
								cycle = true;
							else if (edges[current].front() != previous) {
								previous = current;
								current = edges[current].front();
							}
							else {
								previous = current;
								current = edges[current].back();
							}
						}
					}

					// if there is no cycle, keep track of the new minimum
					if (!cycle) {
						minDistance = distances[*state1][*state2];
						bestState1 = state1;
						bestState2 = state2;
					}
				}
			}
		}

		// store the shortest edge (as adjacency information)
		edges[*bestState1].push_back(*bestState2);
		edges[*bestState2].push_back(*bestState1);

		// remove from the list of candidates the points that now have two outgoing edges
		if (edges[*bestState1].size() == 2)
			candidates.erase(bestState1);
		if (edges[*bestState2].size() == 2)
			candidates.erase(bestState2);
	}

	// return the shortest path (whose extremities are the two remaining candidates) and its length
	double pathLength = 0.;
	unsigned it1 = candidates.front();
	unsigned it2 = edges[it1].front();
	solution.push_back(it1);
	solution.push_back(it2);
	pathLength += distances[it1][it2];
	while (it2 != candidates.back())
	{
		if (edges[it2].front() != it1) {
			it1 = it2;
			it2 = edges[it1].front();
		}
		else {
			it1 = it2;
			it2 = edges[it1].back();
		}
		solution.push_back(it2);
		pathLength += distances[it1][it2];
	}

	return pathLength;
}

