/*
 * RRT-Star.cpp
 *
 *  Created on: Nov 04, 2013
 *      Author: ddevaurs
 */

#include <boost/math/special_functions/factorials.hpp>

#include "API/Roadmap/compco.hpp"
#include "API/Roadmap/graph.hpp"

#include "planner/planEnvironment.hpp"
#include "planner/Diffusion/Variants/RRT-Star.hpp"

using namespace std;


/*
 * Constructor.
 */
StarExpansion::StarExpansion(Graph * graph) : RRTExpansion(graph)
{
	Robot * robot = m_Graph->getRobot();

	// get the list of joints that are passive in some kinematic constraints
	list<Joint *> passiveJoints;
	robot->getListOfPassiveJoints(passiveJoints);

	// compute the dimension and the volume of the configuration space
	dimension = 0;
	double volumeOfSpace(1.);
	for (unsigned i(0); i < robot->getNumberOfJoints(); ++i)
	{
		Joint * joint = robot->getJoint(i);

		// ignore passive joints
		if (find(passiveJoints.begin(), passiveJoints.end(), joint) != passiveJoints.end())
			continue;

		dimension += joint->computeDimension();
		volumeOfSpace *= joint->computeSpaceVolume();
	}

	// compute gamma*
	if (dimension == 2)
		gamma_star = 2. * sqrt(1.5 * volumeOfSpace / M_PI);
	else if (dimension == 3)
		gamma_star = 2. * pow(volumeOfSpace / M_PI, 1. / 3.);
	else if (dimension == 4)
		gamma_star = pow(40. * volumeOfSpace / (M_PI * M_PI), 1. / 4.);
	else if (dimension == 5)
		gamma_star = pow(72. * volumeOfSpace / (M_PI * M_PI), 1. / 5.);
	else if (dimension == 6)
		gamma_star = 2. * pow(7. * volumeOfSpace / (M_PI * M_PI * M_PI), 1. / 6.);
	else if (dimension == 7)
		gamma_star = pow(960. * volumeOfSpace / (M_PI * M_PI * M_PI), 1. / 7.);
	else if (dimension == 8)
		gamma_star = 2. * pow(27. * volumeOfSpace / (M_PI * M_PI * M_PI * M_PI), 1. / 8.);
	else if (dimension == 9)
		gamma_star = pow(16800. * volumeOfSpace / (M_PI * M_PI * M_PI * M_PI), 1. / 9.);
	else
	{
		double volumeOfUnitSphere;
		if ((dimension % 2) == 0)
			volumeOfUnitSphere = pow(M_PI, dimension / 2) / boost::math::factorial<double>(dimension / 2);
		else
			volumeOfUnitSphere = pow(M_PI, (dimension - 1) / 2) * pow(2., (dimension + 1) / 2)
					/ boost::math::double_factorial<double>(dimension);

		gamma_star = 2. * pow((1. + 1. / dimension) * volumeOfSpace / volumeOfUnitSphere, 1. / dimension);
	}

	// compute gamma
	gamma = gamma_star * PlanEnv->getDouble(PlanParam::starRadius);
}


/**
 * Check whether the given path is a valid connection.
 * @return TRUE if the connection is valid, and FALSE otherwise
 */
bool StarExpansion::isValidConnection(localPathPtr_t path)
{
	// N.B.: the connection is attempted only if the path length is less than a specified threshold
	// (which is some multiple of the extension step)
	if (path->getParamMax() > ENV.getDouble(Env::minimalFinalExpansionGap) * this->step())
		return false;
	else
		// check whether the path is collision-free
		return path->isValid();
}


/**
 * Add a node to the graph, after an expansion that has produced the given path, starting from fromNode.
 * @param directionNode: node toward which the expansion was attempted.
 * @return the added node
 */
Node * StarExpansion::addNode(localPathPtr_t path, Node * fromNode, Node * directionNode)
{
	// Conditional activation of RRT*
//	if (m_Graph->getNumberOfCompco() > 1)
//		return RRTExpansion::addNode(path, fromNode, directionNode);

	// create a new node and add it to the graph (when needed)
	confPtr_t newConf = path->getEnd();
	Node * newNode = directionNode ? directionNode : new Node(m_Graph, newConf);
	if (!directionNode)
		m_Graph->addNode(newNode);

	// compute the list of nearest neighbors of the new node (within a ball in C-space)
	double nbNodes = (double) m_Graph->getNumberOfNodes();
	double radius = gamma * pow(log(nbNodes) / nbNodes, 1. / (double) dimension);
	vector<Node *> nearNodes = fromNode->getConnectedComponent()->neighborsInBall(newConf, radius);

	// validNodes will store the subset of neighbors that can be connected to newNode by a valid path
	vector<Node *> validNodes;
	validNodes.push_back(fromNode);
	vector<localPathPtr_t> validPaths;
	validPaths.push_back(path);

	// among its neighbors, look for a parent for the new node:
	// it is the node minimizing the cost of the path (in the graph) from the Start to the new node
	Node * minNode = fromNode;
	localPathPtr_t minPath = path;
	double minCostFromStart = fromNode->sumCost(true) + path->cost();

	for (unsigned i(0); i < nearNodes.size(); ++i)
	{
		if (nearNodes[i] == fromNode)
			continue;

		localPathPtr_t path(new LocalPath(nearNodes[i]->getConfiguration(), newConf));
		if (path->isValid())
		{
			validNodes.push_back(nearNodes[i]);
			validPaths.push_back(path);

			double costFromStart = nearNodes[i]->sumCost(true) + path->cost();
			if (costFromStart < minCostFromStart)
			{
				minCostFromStart = costFromStart;
				minNode = nearNodes[i];
				minPath = path;
			}
		}
	}

	// link the new node to minNode (i.e. link it along the minimum-cost path)
	m_Graph->addEdges(minNode, newNode, minPath);
	newNode->sumCost() = minCostFromStart;

	// rewire the graph:
	// if having newNode as its parent allows one of its neighbors to be connected to the Start
	// along a lower-cost path, change the current connections
	for (unsigned i(0); i < validNodes.size(); ++i)
	{
		if (validNodes[i] == minNode || !validNodes[i]->parent())
			continue;

		if (validNodes[i]->sumCost() > newNode->sumCost() + validPaths[i]->getReverseCost())
		{
			m_Graph->removeEdges(validNodes[i]->parent(), validNodes[i]);
			m_Graph->addEdges(newNode, validNodes[i], validPaths[i]->buildReversePath());
		}
	}

	return newNode;
}

