#include "Planner-pkg.h"
#include "API/Roadmap/graph.hpp"
#include "planner/planEnvironment.hpp"
#include "planner/PRM/DirectedPRM.hpp"
#include "planner/TrajectoryOptim/Classic/smoothing.hpp"
#include "Graphic-pkg.h"

using namespace std;
using namespace tr1;
using namespace API;

void DirectedPRM::pause(){
    bool was_running=ENV.getBool(Env::isRunning);
    g3d_draw_allwin_active();
    ENV.setBool(Env::isRunning,false);
    int c;
    printf(" ... [PAUSED] ...");
    fflush( stdout );
    do c = getchar(); while ((c != '\n') && (c != EOF));
    ENV.setBool(Env::isRunning,was_running);
}

bool DirectedPRM::initialize(){
    if(!_Robot){
        printf("error: DirectedPRM::initialize: planner has no robot\n");
        return false;
    }
    if(!_Graph){
        printf("error: DirectedPRM::initialize: planner has no graph\n");
        return false;
    }
    _Graph->resetConnections();
    if(!this->setStartNode(_Robot->getInitialPosition()))
        return false;
    if(!this->setGoalNode(_Robot->getGoTo()))
        return false;
    this->updateConnections();
    _Init=true;
    return true;
}

bool DirectedPRM::setStartNode(shared_ptr<Configuration> Cs){

    if(!Cs->setConstraintsWithSideEffect()){
        printf("error: DirectedPRM::setStartNode: start configuration does not respect the constraints\n");
        return false;
    }
    if(Cs->isInCollision()){
        printf("error: DirectedPRM::setStartNode: start configuration is in collision\n");
        return false;
    }
    Node* N=_Graph->searchConf(*Cs);
    bool added=false;
    if(!N){
        N=new Node(_Graph,Cs);
        if(!N){
            printf("error: DirectedPRM::setStartNode: cannot allocate start node\n");
            return false;
        }
        _Graph->addNode(N);
        m_nbAddedNode++;
        if(N->getConfiguration()->isConnectible())
            m_nbConnectibleNodes++;
        added=true;
    }

    _Start=N;
    _Start->backwardConnectToSource();

    _Graph->getGraphStruct()->search_start=_Start->getNodeStruct();
    _q_start = Cs;

    if(added)
        this->linkNode(_Start);

    return true;
}

bool DirectedPRM::setGoalNode(shared_ptr<Configuration> Cg){

    if(!Cg->setConstraintsWithSideEffect()){
        printf("error: DirectedPRM::setGoalNode: goal configuration does not respect the constraints\n");
        return false;
    }
    if(Cg->isInCollision()){
        printf("error: DirectedPRM::setGoalNode: goal configuration is in collision\n");
        return false;
    }
    Node* N=_Graph->searchConf(*Cg);
    bool added=false;
    if(!N){
        N=new Node(_Graph,Cg);
        if(!N){
            printf("error: DirectedPRM::setGoalNode: cannot allocate goal node\n");
            return false;
        }
        _Graph->addNode(N);
        m_nbAddedNode++;
        if(N->getConfiguration()->isConnectible())
            m_nbConnectibleNodes++;
        added=true;
    }

    _Goal=N;
    _Goal->forwardConnectToGoal();

    _Graph->getGraphStruct()->search_goal=_Goal->getNodeStruct();
    _q_goal=Cg;

    if(added)
        this->linkNode(_Goal);

    return true;
}

bool DirectedPRM::checkStopConditions(){

    if(ENV.getBool(Env::expandToGoal) && this->trajFound()){
        cout << "Stop: trajectory found."<<endl;
        return true;
    }

    if (PlanEnv->getBool(PlanParam::stopPlanner)) {
        cout << "Stop: planning canceled by user." << endl;
        return true;
    }

    if (PlanEnv->getBool(PlanParam::planWithTimeLimit) &&
            m_time >= PlanEnv->getDouble(PlanParam::timeLimitPlanning)) {
        cout << "Stop: the time limit has been reached." << endl;
        return true;
    }

    if (m_nbConscutiveFailures >= (unsigned) ENV.getInt(Env::NbTry)) {
        cout << "Stop: the maximum number of consecutive failures has been reached." << endl;
        return true;
    }

    if (_Graph->getNumberOfNodes() >= (unsigned) ENV.getInt(Env::maxNodeCompco)) {
        cout << "Stop: the maximum number of nodes has been reached." << endl;
        return true;
    }

    return false;
}

bool DirectedPRM::preConditions(){

    if (_Start->getConfiguration()->equal(*_Goal->getConfiguration())) {
        cout<<"error: DirectedPRM::preConditions: graph creation failed: start and goal are the same"<<endl;
        return false;
    }

    LocalPath *LP=new LocalPath(_Start->getConfiguration(),_Goal->getConfiguration());
    if(LP->isValid()){
        _Graph->addOrientedEdge(_Start,_Goal,localPathPtr_t(LP));
        cout<<"DirectedPRM::preConditions: direct connection"<<endl;
        return true;

    }
    else{
        cout<<"DirectedPRM::preConditions: no direct connection"<<endl;
        delete LP;
    }

	return true;
}

void DirectedPRM::expandOneStep(){

    ++m_nbShoots;
    confPtr_t q=_Robot->shoot();

    if(q->setConstraintsWithSideEffect() && (!q->isInCollision())){
        ++m_nbExpansions;
		m_nbConscutiveFailures = 0;
        ++m_nbAddedNode;
        if(q->isConnectible())
            m_nbConnectibleNodes++;
        Node * N=new Node(_Graph,q);
        _Graph->addNode(N);
        this->linkNode(N);
	}
    else{
        ++m_nbFailures;
		++m_nbConscutiveFailures;
    }
}

unsigned int DirectedPRM::run(){

	double ts(0.);
    m_time = 0.;

	if (!this->preConditions()) {
		cout << "Failure: pre-conditions are not satisfied." << endl;
		return 0;
	}

    while (!this->checkStopConditions()){

		this->expandOneStep();

		// if the expansion is successful
        if (m_nbConscutiveFailures==0){
			// draw the scene
            if (!ENV.getBool(Env::drawDisabled) && ENV.getBool(Env::drawExploration))
				_draw_func();
		}
	}

	return m_nbAddedNode;
}

bool DirectedPRM::trajFound(){
    if(!_Init)
        return false;
    if(!_Graph)
        return false;
    if(!_Start)
        return false;
    if(!_Goal)
        return false;
    unsigned int start_ssc=_Start->getSccId();
    unsigned int goal_ssc=_Goal->getSccId();
    return  (start_ssc==goal_ssc || _Graph->areSccConnected(start_ssc,goal_ssc));
}

bool DirectedPRM::linkNode(Node *N1){
    if(ENV.getBool(Env::expandToGoal) && this->trajFound())
        return false;
    double dmax=p3d_get_DMAX();
    bool use_dist=(ENV.getBool((Env::useDist)) && dmax>0.);
    for(unsigned int scc_id=0;scc_id<_Graph->getStronglyConnectedComponents().size();scc_id++){
        if(N1->getSccId()!=scc_id){
            vector<Node*> scc=_Graph->getStronglyConnectedComponents()[scc_id];
            // Try forward connection to scc if relevant
            if(!_Graph->areSccConnected(N1->getSccId(),scc_id)){
                vector<double> distances=_Graph->sortSCC(scc_id,N1,true);
                for(unsigned int node_id=0;node_id<scc.size();node_id++){
                    if(use_dist && node_id<distances.size() && distances[node_id]>dmax){
                        break;
                    }
                    Node *N2=scc[node_id];
                    localPathPtr_t LP12=localPathPtr_t(new LocalPath(N1->getConfiguration(),N2->getConfiguration()));
                    if(LP12->isValid()){
                        _Graph->addOrientedEdge(N1,N2,LP12);
                        break;
                    }
                }
            }
            // Try backward connection to scc if relevant
            if((!ENV.getBool(Env::expandToGoal) || !this->trajFound()) && N1->getSccId()!=scc_id && !_Graph->areSccConnected(scc_id,N1->getSccId())){
                vector<double> distances=_Graph->sortSCC(scc_id,N1,false);
                for(unsigned int node_id=0;node_id<scc.size();node_id++){
                    if(use_dist && node_id<distances.size() && distances[node_id]>dmax)
                        break;
                    Node *N2=scc[node_id];
                    localPathPtr_t LP21=localPathPtr_t(new LocalPath(N2->getConfiguration(),N1->getConfiguration()));
                    if(LP21->isValid()){
                        _Graph->addOrientedEdge(N2,N1,LP21);
                        break;
                    }
                }
            }
        }
        if(ENV.getBool(Env::expandToGoal) && this->trajFound())
            break;
    }
    return true;
}

void DirectedPRM::updateConnections(){
    _Graph->unflagAll();
    _Start->updateForward();
    _Graph->unflagAll();
    _Goal->updateBackward();
    cout<<"Connections updated"<<endl;
}

void DirectedPRM::extractSmoothAndReplaceTraj(){
    Trajectory *extractTraj=_Graph->extractAStarShortestPath(_Start,_Goal);
    if(!extractTraj){
        printf("error: DirectedPRM::extractSmoothAndReplaceTraj: no trajectory was extracted\n");
        return;
    }
    extractTraj->replaceP3dTraj();
    delete extractTraj;
    if(PlanEnv->getBool(PlanParam::withSmoothing)){
        g3d_draw_allwin_active();
        Robot *robot=_Graph->getRobot();
        API::Smoothing optimTrj(robot,robot->getRobotStruct()->tcur);
        double old_cost=optimTrj.cost();
        cout<<"| Smoothing:"<<endl;
        cout<<"|    Previous cost: "<<old_cost<<endl;
        optimTrj.removeUselessNodes();
        optimTrj.runShortCut(PlanEnv->getInt(PlanParam::smoothMaxIterations));
        optimTrj.removeUselessNodes();
        optimTrj.replaceP3dTraj();
        double cost=optimTrj.cost();
        cout<<"|    New cost: "<<cost<<endl;
        cout<<"|    Smoothing gain: "<<100*(old_cost-cost)/old_cost<<" %"<<endl;
    }
}

