#include "Geometry.h"
#include "Logging/Logger.h"

namespace m3dGeometry
{


Eigen::Vector3d p3dToEigen(const p3d_point v){
    Eigen::Vector3d r;
    r.x() =v.x;
    r.y() =v.y;
    r.z() =v.z;
    return r;
}

Eigen::Affine3d p3dToEigen(const p3d_matrix4 pos){
    Eigen::Affine3d f;
    for (unsigned int x=0;x<3;x++){
        for (unsigned int y=0;y<4;y++){
            f(x,y)=pos[x][y];
        }
    }
    for (unsigned int y=0;y<3;y++){
        f(3,y)=0;
        //std::cout<<pos[3][y]<<" ";
    }
    f(3,3)=1;
    //std::cout<<pos[3][3]<<std::endl;
    return f;
}

p3d_matrix4* eigenToP3d(const Eigen::Affine3d &f){
    p3d_matrix4 *pos=new p3d_matrix4[1];
    for (unsigned int x=0;x<4;x++){
        for (unsigned int y=0;y<4;y++){
            (*pos)[x][y] = f(x,y);
        }
    }
    return pos;
}

p3d_point getTranslation(const p3d_matrix4 t){
    p3d_point p;
    p.x=t[0][3];
    p.y=t[1][3];
    p.z=t[2][3];
    return p;
}

Eigen::Vector2d getConfBase2DPos(const Configuration &q){
    Eigen::Vector2d v;
    v[0] = q[6];
    v[1] = q[7];
    return v;
}

void setBaseOrientation(Robot * rob,const Eigen::Matrix3d & rot_mat){
    Eigen::Vector3d angles = rot_mat.eulerAngles(0,1,2);
    setBaseOrientation(rob,angles);
}
void setBaseOrientation(Robot *rob, const Eigen::Vector3d & angles){
    Configuration q = *rob->getCurrentPos();
    for(unsigned int i=0;i<3;i++){
        q.at(6+3+i)=angles[i];
    }
    rob->setAndUpdate(q);
}

void setBasePosition(Robot *rob, const Eigen::Vector3d & pos){
    Configuration q=*rob->getCurrentPos();
    for(unsigned int i=0;i<3;i++){
        q.at(6+i)=pos[i];
    }
    rob->setAndUpdate(q);
}

void setBasePosition2D(Robot *rob, const Eigen::Vector2d & pos){
    Configuration q=*rob->getCurrentPos();
    for(unsigned int i=0;i<2;i++){
        q.at(6+i)=pos[i];
    }
    rob->setAndUpdate(q);
}

void setBase6D(Robot *rob,const Eigen::Vector3d &pos, const Eigen::Vector3d &angles){
    setBasePosition(rob,pos);
    setBaseOrientation(rob,angles);
}


void setBase4D(Robot *rob, const Eigen::Vector3d &pos, double theta){
    Eigen::Vector3d angles;
    angles.x()=angles.y()=0;
    angles.z()=theta;
    setBasePosition(rob,pos);
    setBaseOrientation(rob,angles);
}

void setBaseTransform(Robot* rob, const Eigen::Affine3d & tf){
    setBaseOrientation(rob,tf.rotation());
    setBasePosition(rob,tf.translation());
}


void setBaseFromConf(Robot *rob, const Configuration & conf){
    Configuration q = *rob->getCurrentPos();
    for(unsigned int i=0;i<6;i++){
        q[6+i] = conf[6+i];
    }
    rob->setAndUpdate(q);
}

Eigen::Affine3d changeFrame(const Eigen::Affine3d &posF1, const Eigen::Affine3d &F1, const Eigen::Affine3d &F2)
{
    Eigen::Affine3d new_pos;
    Eigen::Affine3d abs_pos = F1*posF1;
    new_pos =  F2.inverse() * abs_pos;
    return new_pos;
}

Eigen::Vector3d changeFramePoint(const Eigen::Vector3d p1, const Eigen::Affine3d &F1, const Eigen::Affine3d &F2){
    Eigen::Vector3d new_p, abs_p;
    abs_p = F1*p1;
    new_p= F2.inverse() * abs_p;
    return new_p;
}

Eigen::Affine3d absoluteFramePos()
{
    Eigen::Affine3d f;
    f=Eigen::Translation3d(0,0,0);
    return f;
}

void test()
{
    INIT_MOVE3D_LOG("planners.m3dgeometry.test");
    Eigen::Affine3d pos1,f1,f2,pos2,abs1,abs2;
    f1=Eigen::Translation3d(1,0,0);
    f2=Eigen::Translation3d(2,0,0); f2 *= Eigen::AngleAxisd(0.5*M_PI,Eigen::Vector3d::UnitZ());
    //test changeFrame
    pos1=Eigen::Translation3d(1,2,3); pos1 *= Eigen::AngleAxisd(0.5*M_PI,Eigen::Vector3d::UnitZ());
    pos2=changeFrame(pos1,f1,f2);

    //M3D_DEBUG("\n"<<pos1.translation()<<"\n==>\n"<<pos2.translation());
    M3D_DEBUG("\n"<<Eigen::Matrix4d(pos1.matrix())<<"\n==>\n"<<Eigen::Matrix4d(pos2.matrix()));
    M3D_DEBUG("\n"<< pos1 * Eigen::Vector3d::UnitX()<<"\n==>\n"<< pos2 *Eigen::Vector3d::UnitX());
    M3D_DEBUG("\n"<<(pos1*f1).matrix());

    Eigen::Vector3d v=changeFramePoint(Eigen::Vector3d::UnitY(),f1,f2);
    M3D_DEBUG("changeFramePoint\n"<<v);
}


}

