#include "MultiAgentHandOvers.h"

#include "P3d-pkg.h"
#include "Collision-pkg.h"
#include "VimanTags.h"
#include "Geometry.h"
#include "HRICS_Workspace.hpp"

#include "GTP/taskManagerInterface.hpp"



namespace HRICS {
namespace MHO {

const std::string HandOverTestDistance::_name="distance";
const std::string HandOverTestGrabs::_name="grabs";

FrontierTester::FrontierTester(){}
void FrontierTester::addTester(AbstractHandOverTest *tester, bool is_default){
    this->_testers[tester->getName()]=tester;
    if(is_default){
        this->_defaults[tester->getCheckStatusLevel()]=tester->getName();
    }
}

FrontierTester::~FrontierTester(){
    for(map<std::string,AbstractHandOverTest*>::iterator it=_testers.begin();it!=_testers.end();++it){
        delete it->second;
    }
}

void FrontierTester::addDefault(const string &name){
    this->_defaults[getTester(name)->getCheckStatusLevel()]=name;
}

AbstractHandOverTest* FrontierTester::getTester(const string &name){
    if(this->_testers.find(name) == this->_testers.end()){
        return 0;
    }
    return this->_testers[name];
}
AbstractHandOverTest* FrontierTester::getDefault(FrontierCells::FrontierCheckStatus target_status){
    return getTester(this->_defaults[target_status]);
}

std::vector< std::pair<std::string,bool> > FrontierTester::listTesterNames(bool default_only){
    vector< std::pair<std::string,bool> > list;
    for(map<string,AbstractHandOverTest*>::iterator it=_testers.begin();it!=_testers.end();it++){
        map<FrontierCells::FrontierCheckStatus,string>::iterator it_def=_defaults.find(it->second->getCheckStatusLevel());
        bool is_def=false;
        if(it_def->second==it->first){is_def=true;}
        if(!default_only || is_def){
            list.push_back(make_pair(it->first,is_def));
        }
    }
    return list;
}
std::vector< std::pair<AbstractHandOverTest*,bool> > FrontierTester::listTesters(bool default_only){
    vector< std::pair<AbstractHandOverTest*,bool> > list;
    for(map<string,AbstractHandOverTest*>::iterator it=_testers.begin();it!=_testers.end();it++){
        map<FrontierCells::FrontierCheckStatus,string>::iterator it_def=_defaults.find(it->second->getCheckStatusLevel());
        bool is_def=false;
        if(it_def != _defaults.end() && it_def->second==it->first){is_def=true;}
        if(!default_only || is_def){
            list.push_back(make_pair(it->second,is_def));
        }
    }
    return list;
}

bool FrontierTester::test(FrontierPtr frontier, std::string const & method){
    //TODO ensure that the frontier have been checked with previous required checks
    //use pre/post predicates based on FrontierCells::FCS_*
    this->_chrono.reset();
    bool ok=this->getTester(method)->check(frontier);
    assert(ok==frontier->status_ok[this->getTester(method)->getCheckStatusLevel()]);
    assert(frontier->check_status>=this->getTester(method)->getCheckStatusLevel());
    //if(!ok){
    //    cout << "fail "<<method<<endl;
    //}
    this->_check_counts[this->getTester(method)->getCheckStatusLevel()].first++;
    this->_check_counts[this->getTester(method)->getCheckStatusLevel()].second+=this->_chrono.getMilliSec();
    return ok;
}


bool FrontierTester::test(FrontierPtr frontier, FrontierCells::FrontierCheckStatus target_status){

    assert(_defaults.find(target_status)!=_defaults.end());//asserts the defaults are set
    FrontierCells::FrontierCheckStatus st=frontier->check_status,
            req=getDefault(target_status)->getRequiredCheckStatus();
    bool ok=frontier->status_ok[st];

    if( !frontier->isSymetric() && frontier->select_reverse){
        st=frontier->r_check_status;
        ok=frontier->r_status_ok[st];
    }

    if(!ok){
        return false;
    }
    if(target_status==st){
        //already checked
        return ok;
    }else if (st<req){
        //check required level first
        //(recursive)
        ok=test(frontier,req);
    }

    //now at required level:
    if(ok){
        //required successfull
        ok=this->test(frontier,_defaults[target_status]);
    }
    //else : required failed, return fail

    return ok;
}

HandOverTestDistance::HandOverTestDistance(vector<MultiHandOverAgent> * agents):
    _agents(agents)
{}

bool HandOverTestDistance::check(FrontierPtr frontier){
    return check(frontier->agentA_ind,frontier->agentB_ind,frontier->cellA(),frontier->cellB());
}

bool HandOverTestDistance::check(unsigned int a1, unsigned int a2, MultiAgentCell *c1, MultiAgentCell *c2){
    double d=c1->computeDist(c2);
    double r=_agents->at(a1).reach()+_agents->at(a2).reach();
    return d <= r;
}

HandOverTestObject::HandOverTestObject(AbstracHandOverObjectCheck *object_check, std::vector<MultiHandOverAgent> *agents):
    _object_check(object_check),
    _agents(agents)
{}

bool HandOverTestObject::check(FrontierPtr frontier){
    bool ok=false;
    if(frontier->check_status>=this->getCheckStatusLevel()){
        ok=frontier->status_ok[this->getCheckStatusLevel()];
    }else{
        MultiHandOverAgent *a1,*a2;
        a1=&_agents->at(frontier->agentA_ind);
        a2=&_agents->at(frontier->agentB_ind);
        _object_check->robots(a1->robot,a2->robot);
        Eigen::Vector3d s1,s2;
        s1.x()=frontier->cellOf(a1->index)->getCenter().x();
        s1.y()=frontier->cellOf(a1->index)->getCenter().y();
        s2.x()=frontier->cellOf(a2->index)->getCenter().x();
        s2.y()=frontier->cellOf(a2->index)->getCenter().y();
        s1.z()=a1->shoulderHeight();
        s2.z()=a2->shoulderHeight();
        _object_check->shoulders(s1,s2);
        _object_check->reach(a1->reach(),a2->reach());
        if(_object_check->check()){
            ok=true;
            frontier->object_position=confPtr_t(new Configuration(*_object_check->getQMiddle()));
            //cout<<"obj test:\n"<<frontier->object_position->at(6)<<' ' <<frontier->object_position->at(7)<<' '<<frontier->object_position->at(8)<<' '<<endl;
            frontier->status_ok[FrontierCells::FCS_OBJECT]=true;//object_ok=true;
            frontier->grabA=_object_check->grab1();
            frontier->grabB=_object_check->grab2();
        }
        frontier->check_status=std::max(FrontierCells::FCS_OBJECT,frontier->check_status);
    }
    return ok;
}

HandOverTestGrabs::HandOverTestGrabs(std::vector<MultiHandOverAgent> *agents, Robot **object):
    _agents(agents),_object_pt_pt(object)
{
}

bool HandOverTestGrabs::check(FrontierPtr frontier){
    bool ok=true;
    Eigen::Vector3d point;
    configPt q=0;
    if(frontier->check_status>=FrontierCells::FCS_GRAB_CONFS){
        if(!frontier->status_ok[FrontierCells::FCS_GRAB_CONFS]){
            ok=false;
        }else{
            //TODO costs
            ok=true;
        }
    }else{
        //cout<<"check conf HO begin"<<endl;
        //TODO check this out: (remove in the if/else at the begining of the for)
        //p3d_col_deactivate_robot_robot(frontier->agentA->robot()->getRobotStruct(),frontier->agentB->robot()->getRobotStruct());
        p3d_col_deactivate_robot((*this->_object_pt_pt)->getRobotStruct());
        for(unsigned int i_ag=0;i_ag<2;i_ag++){
            MultiAgentCell::AgentInCell *agent;
            MultiAgentCell *cell;
            confPtr_t frontier_conf;
            if(i_ag==0){
                agent=frontier->agentA;
                p3d_col_deactivate_robot_robot(agent->robot()->getRobotStruct(),frontier->agentB->robot()->getRobotStruct());
                cell=frontier->cellA();
                point=frontier->grabA;
                frontier->confA=confPtr_t(new Configuration(agent->robot()));
                frontier_conf=frontier->confA;
            }else{
                agent=frontier->agentB;
                //reactivate colisions to avoid confs where agents are colliding
                p3d_col_activate_robot_robot(agent->robot()->getRobotStruct(),frontier->agentA->robot()->getRobotStruct());
                cell=frontier->cellB();
                point=frontier->grabB;
                frontier->confB=confPtr_t(new Configuration(agent->robot()));
                frontier_conf=frontier->confB;
            }
            Configuration q_current_pos=*agent->robot()->getCurrentPos();
            MHO::MultiHandOver3dPlanner mhopl(agent->robot(),agent->agentInGrid->isHuman());
            //Configuration q_current_pos(*agent->robot()->getCurrentPos());
            //if(!agent->agentInGrid->isHuman && mhopl.grabConfCheck(q,cell->getCenter(),point) ){
            Eigen::Vector3d shoulder(cell->getCenter().x(),cell->getCenter().y(),(*_agents)[agent->index].shoulderHeight());
            if((shoulder-point).norm()>(*_agents)[agent->index].reach()){
                ok=false;
                //cout  << "object is too far to be grasp"<<endl;
            }else if(mhopl.grabConfCheck(q,cell->getCenter(),point)){
                Configuration q_grab(agent->robot(),p3d_copy_config(agent->robot()->getRobotStruct(),q));
                agent->robot()->setAndUpdate(q_grab);
                *frontier_conf=q_grab;
                p3d_destroy_config(agent->robot()->getRobotStruct(),q);
                q=0;
                //TODO add conf cost, handover cost...
                //handover_cost=10;
                //if(_draw_agents_holding_conf_during_search){
                //    g3d_draw_allwin_active();
                //}
            }else{
                //handover_cost=numeric_limits<double>::infinity();
                ok=false;
            }
            //TODO give the possibility not to put the agent at its previous conf when checking grab confs
            agent->robot()->setAndUpdate(q_current_pos);// (or not)
            //if(_draw_agents_holding_conf_during_search){
            //    g3d_draw_allwin_active();
            //}
            if(!ok){
                //cout << "HO find conf fail at "<<i_ag<<" -- "<<agent->robot()->getName()<<endl;
                break;//avoid computing too much
            }
        }
        frontier->check_status=max(frontier->check_status,FrontierCells::FCS_GRAB_CONFS);
        frontier->status_ok[FrontierCells::FCS_GRAB_CONFS]=ok;
    }
    p3d_col_activate_robot_robot(frontier->agentA->robot()->getRobotStruct(),frontier->agentB->robot()->getRobotStruct());
    p3d_col_activate_robot((*this->_object_pt_pt)->getRobotStruct());
    assert(q==0 && "q (double[]) have not been free'd");
    return ok;

}

const std::string HandOverTestPreDefined::_name=std::string("pre-defined");

HandOverTestPreDefined::HandOverTestPreDefined(std::vector<MultiHandOverAgent> *agents, string file_directory, double grid_discrete_step):
    _agents(agents), _file_dir(file_directory),_grid_pace(grid_discrete_step)
{
    _human_robot_min_max_dists=pair<double,double>(0.1,numeric_limits<double>::infinity());
    _robot_robot_min_max_dists=pair<double,double>(0.1,1.8);
}

HandOverTestPreDefined::_ConfGenAlias * HandOverTestPreDefined::_initGetConfGen(MultiHandOverAgent* a1, MultiHandOverAgent* a2){
    _ConfGenAlias *cga(0);
    //find the confGen if it exists
    map<pair<MultiHandOverAgent*,MultiHandOverAgent*>,_ConfGenAlias>::iterator f=
            _conf_generators_alias.find(make_pair(a1,a2));
    if(f==_conf_generators_alias.end()){
        f= _conf_generators_alias.find(make_pair(a2,a1));
    }
    if(f==_conf_generators_alias.end()){
        //create it
        cga=& _conf_generators_alias[make_pair(a1,a2)];
    }else{
        //get it
        cga=& f->second;
    }

    //if conf gen not initialized
    if(!cga->conf_generator){
        cga->alias_human=a1;
        cga->alias_robot=a2;
        if(a1->is_human && a2->is_human){
            //both human
            cga->conf_generator = new ConfGenerator(a1->robot,a2->robot);
        }else if(a1->is_human && ! a2->is_human){
            //0:human; 1:robot
            cga->conf_generator = new ConfGenerator(a2->robot,a1->robot);
        }else if(a2->is_human && ! a1->is_human){
            //0:robot; 1:human
            swap(cga->alias_human,cga->alias_robot);
            cga->conf_generator = new ConfGenerator(a1->robot,a2->robot);
        }else if(!a2->is_human && ! a1->is_human){
            //both robots
            cga->conf_generator = new ConfGenerator(a1->robot,a2->robot);
        }

        //get configuration form file
        string filename;

        bool file_found=false;
        if(_file_dir[_file_dir.size()-1]!='/'){
            _file_dir+="/";
        }
        //_file_dir+="statFiles/OtpComputing/";

        unsigned int try_nb=0;
        bool ok=true;
        //file names are in the form confHUMAN_NAME4-ROBOT-NAME1.xml
        //or confHUMAN_NAME-ROBOT_NAME.xml (name with no ending digits)
        //or confHUMAN_NAME2.xml or confHUMAN_NAME.xml
        //Note that both can be human or robots
        while(!file_found && ok){
            filename=_file_dir;
            filename+="conf";
            if(try_nb==0){
                filename+=cga->alias_human->robot->getName();
                filename+="-";
                filename+=cga->alias_robot->robot->getName();
            }else if(try_nb==1){
                string name=cga->alias_human->robot->getName();
                //remove digits at end of name
                while(isdigit(name[name.size()-1])){
                    name=name.substr(0,name.size()-1);
                }
                filename+=name;
                filename+="-";
                name=cga->alias_robot->robot->getName();
                //remove digits at end of name
                while(isdigit(name[name.size()-1])){
                    name=name.substr(0,name.size()-1);
                }
                filename+=name;

            }else if(try_nb==2){
                //try with human name only
                filename+=cga->alias_human->robot->getName();
            }else if(try_nb==3){
                //try with human name only and no ending digits
                string name=cga->alias_human->robot->getName();
                //remove digits at end of name
                while(isdigit(name[name.size()-1])){
                    name=name.substr(0,name.size()-1);
                }
                filename+=name;
            }else{
                ok=false;
                //it test for the file conf.xml as last chance
            }
            filename+=".xml";
            cout<<"try "<<filename<<endl;
            if(access(filename.c_str(),F_OK|R_OK)==0){
                cga->conf_list=cga->conf_generator->loadFromXml(filename);
                if(cga->conf_list.size()){
                    file_found=true;
                    ok=true;
                }
            }
            try_nb++;
        }
        if(!ok){
            cout<<"Warning! No pre defined hand over file found for "<<a1->robot->getName()<<" and "<<a2->robot->getName()<<endl;
            cout<<"\t\tusing the grasp conf found by IK as handover pose"<<endl;
        }else{
        }
        cga->use_dummy=!ok;
    }
    return cga;
}

HandOverTestPreDefined::~HandOverTestPreDefined(){
    for(map<pair<MultiHandOverAgent*,MultiHandOverAgent*>, _ConfGenAlias>::iterator it=_conf_generators_alias.begin();
        it!=_conf_generators_alias.end();it++){
        if(it->second.conf_generator){
            delete it->second.conf_generator;
        }
    }
}

bool HandOverTestPreDefined::check(FrontierPtr frontier){
    bool ok=true;

    if(frontier->check_status>=FrontierCells::FCS_HANDOVER){
        if(!frontier->status_ok[FrontierCells::FCS_HANDOVER]){
            ok=false;
        }else{
            //TODO costs
            ok=true;
        }
    }else if (frontier->check_status < HandOverTestPreDefined::getRequiredCheckStatus()){
        ok=false;
    }else{
        MultiHandOverAgent* a1(0),*a2(0);
        for(unsigned int i=0; (!a1 || !a2) && i<_agents->size();i++){
            if(_agents->at(i).robot==frontier->agentA->robot()){
                a1=&_agents->at(i);
            }else if(_agents->at(i).robot==frontier->agentB->robot()){
                a2=&_agents->at(i);
            }
        }
        if(!a1 || !a2){
            //critical error
            cerr<<"agents not found in HandOverTestPreDefined"<<endl;
            ok=false;
        }
        if(ok){
            _ConfGenAlias * cga=_initGetConfGen(a1,a2);
            Eigen::Vector2d posAfrt=frontier->cellA()->getCenter(), posBfrt=frontier->cellB()->getCenter();
            Eigen::Vector2d diff_frt=posBfrt-posAfrt;//the distance between the agents in the frontier
            Eigen::Vector2d diff_frt_n=diff_frt;
            diff_frt_n.normalize();
            Eigen::Vector2d x(1,0);
            double thAfrt=frontier->confA->at(6+5);
            double thBfrt=frontier->confB->at(6+5);

            bool found=false;
            if(cga->use_dummy){
                found=true;
            }
            unsigned int i=0;
            //loop
            //  place agents in free conf at the good distance
            //  test configurations
            //  if no collision, then, exit ok
            //end loop
            //for(int i=11;i<20;i++){
            //    Robot* rob=frontier->agentA->robot();
            //    cout << "joint "<<i<<" is "<<rob->getJoint(i)->getName()<<endl;
            //}
            while(!found && i < cga->conf_list.size()){
                ConfigHR & conf=cga->conf_list[i];
                Eigen::Vector2d posA,posB;
                double dist_conf,thA,thB;
                Eigen::Vector2d diff_conf_file,diff_conf_file_n,diff_conf=diff_frt;
                diff_conf.normalize();

                confPtr_t confA,confB;
                if(cga->alias_human->index==frontier->agentA_ind){
                    //A is "human" in conf_gen
                    confA=confPtr_t(new Configuration(cga->alias_human->robot,conf.getHumanConf()));
                    confB=confPtr_t(new Configuration(cga->alias_robot->robot,conf.getRobotConf()));
                }else{
                    //B is "human" in conf_gen
                    confB=confPtr_t(new Configuration(cga->alias_human->robot,conf.getHumanConf()));
                    confA=confPtr_t(new Configuration(cga->alias_robot->robot,conf.getRobotConf()));
                }

                //int intcnt=0,dof_cnt=0;
                //for(unsigned int ci=0;ci<confA->getRobot()->getRobotStruct()->nb_dof;ci++){
                //    double fl=std::floor(confA->at(ci));
                //    if (fl!=0 && fl==confA->at(ci)){
                //        intcnt++;
                //    }
                //    dof_cnt++;
                //}
                //for(unsigned int ci=0;ci<confB->getRobot()->getRobotStruct()->nb_dof;ci++){
                //    double fl=std::floor(confB->at(ci));
                //    if (fl!=0 && fl==confB->at(ci)){
                //        intcnt++;
                //    }
                //    dof_cnt++;
                //}
                //cout<<intcnt<<"int DoF over "<<dof_cnt<<endl;


                posA[0] = confA->at(6+0);
                posA[1] = confA->at(6+1);
                thA = confA->at(6+5);
                posB[0] = confB->at(6+0);
                posB[1] = confB->at(6+1);
                thB = confB->at(6+5);

                diff_conf_file=posB-posA;
                diff_conf_file_n=diff_conf_file;
                diff_conf_file_n.normalize();
                // double angle_diff_conf_file=atan2(diff_conf_file_n.y(),diff_conf_file_n.x()) - atan2(x.y(),x.x());
                // double thRelFileA=thA-angle_diff_conf_file;
                // double thRelFileB=thB-M_PI-angle_diff_conf_file;
                bool dist_ok=true;
                if(abs(diff_conf_file_n.norm()-1)>=0.000001){
                    cout<<"error in config read"<<endl;
                    dist_ok=false;
                    //assert(diff_conf_file_n.norm()-1<0.0000001);
                }
                dist_conf=diff_conf_file.norm();
                if(dist_ok && cga->alias_human->is_human && !cga->alias_robot->is_human){
                    //human/robot
                    dist_ok=(dist_conf>=_human_robot_min_max_dists.first && dist_conf<=_human_robot_min_max_dists.second);
                }else if(dist_ok && !cga->alias_human->is_human && !cga->alias_robot->is_human){
                    //2 robots
                    dist_ok=(dist_conf>=_robot_robot_min_max_dists.first && dist_conf<=_robot_robot_min_max_dists.second);
                }
                if(dist_ok && !frontier->isOverObstacle() && abs(dist_conf - diff_frt.norm()) > _grid_pace){
                    dist_ok=false;
                }
                if(dist_ok && dist_conf>0.00000001){
                    Eigen::Vector2d new_posA,new_posB;

                    double new_thA,new_thB;
                    double angle_diff;
                    angle_diff=atan2(diff_frt_n.y(),diff_frt_n.x()) - atan2(diff_conf_file_n.y(),diff_conf_file_n.x());
                    new_thA=thA+angle_diff;
                    new_thB=thB+angle_diff;
                    //cout << thA <<';'<<thB<<';'<<angle_diff<<';'<<new_thA<<';'<<new_thB<<endl;

                    new_posA=frontier->cellA()->getCenter();
                    new_posB=frontier->cellB()->getCenter();

                    diff_conf*=dist_conf;

                    new_posA += (diff_frt - diff_conf )/2;
                    new_posB -= (diff_frt - diff_conf )/2;

                    Eigen::Vector2d new_diff=new_posB-new_posA;
                    double new_dist=new_diff.norm();
                    if(abs( new_dist - dist_conf ) > 0.00001){
                        cout << "#####################  positions are wrong"<<endl;
                    }
                    if(abs(new_diff.dot(diff_conf_file_n)/new_dist - cos(angle_diff)) > 0.0001){
                        cout<< "#####################  angle is wrong : "<< acos(new_diff.dot(diff_conf_file_n)/new_dist) << " vs "<<angle_diff<<endl;
                    }


                    //<<
                    if(cga->alias_human->index==frontier->agentA_ind){
                        //A is "human" in conf_gen
                        new_posA=posAfrt;
                        new_thA=thAfrt;
                        new_posB[0]=new_posA[0]+posB[0]*cos(new_thA)-posB[1]*sin(new_thA);
                        new_posB[1]=new_posA[1]+posB[1]*cos(new_thA)+posB[0]*sin(new_thA);
                        new_thB=thB+new_thA;
                    }else{
                        //B is "human" in conf_gen
                        new_posB=posBfrt;
                        new_thB=thBfrt;
                        new_posA[0]=new_posB[0]+posA[0]*cos(new_thB)-posA[1]*sin(new_thB);
                        new_posA[1]=new_posB[1]+posA[1]*cos(new_thB)+posA[0]*sin(new_thB);
                        new_thA=thA+new_thB;
                    }
                    //new_posA=posA;
                    //new_thA=thA+
                    //<<


                    //place agents to their position for the HO
                    for(unsigned int i_c=0;i_c<2;i_c++){
                        (*confA)[6+i_c]=new_posA[i_c];
                        (*confB)[6+i_c]=new_posB[i_c];
                    }

                    //for the orientation
                    //double angleA=acos(x.dot(diff_conf) /diff_conf.norm());
                    //double angleB=angleA+M_PI;

                    //(*confA)[6+5]=angle_limit_PI(angleA);
                    //(*confB)[6+5]=angle_limit_PI(angleB);
                    (*confA)[6+5]=angle_limit_PI(new_thA);
                    (*confB)[6+5]=angle_limit_PI(new_thB);

                    frontier->agentA->robot()->setAndUpdate(*confA);
                    frontier->agentB->robot()->setAndUpdate(*confB);

                    //compute to have the object more or less at mid distance
                    //between the initial cells of the handover
                    Eigen::Vector2d eefect_pos,diff_obj,obj_pos;
                    //obj_pos[0]=frontier->object_position->at(6);
                    //obj_pos[1]=frontier->object_position->at(7);
                    obj_pos=(posAfrt+posBfrt)/2;//the middle position
                    if(frontier->agentA->agentInGrid->isHuman() && frontier->agentB->agentInGrid->isHuman()){
                        //2 humans
                        diff_obj*=0;
                    }else if(frontier->agentA->agentInGrid->isHuman()){
                        //A=human, B=robot
                        Robot* rob=frontier->agentB->robot();
                        p3d_matrix4* abs_pos=rob->getJoint(13)->getAbsPos();
                        p3d_vector3 v3pos;
                        p3d_mat4ExtractTrans(*abs_pos,v3pos);
                        eefect_pos[0]=v3pos[0];
                        eefect_pos[1]=v3pos[1];
                        diff_obj=obj_pos-eefect_pos;
                        //get the robot closer
                        Eigen::Vector2d eps=new_posA-new_posB;
                        eps.normalize();
                        eps*=0.1;//10cm
                        //diff_obj+=eps;
                    }else {
                        //A=robot, B=X
                        Robot* rob=frontier->agentA->robot();
                        p3d_matrix4* abs_pos=rob->getJoint(13)->getAbsPos();
                        p3d_vector3 v3pos;
                        p3d_mat4ExtractTrans(*abs_pos,v3pos);
                        eefect_pos[0]=v3pos[0];
                        eefect_pos[1]=v3pos[1];
                        diff_obj=obj_pos-eefect_pos;
                        if(frontier->agentB->agentInGrid->isHuman()){
                            //B is human
                            //get the robot closer
                            Eigen::Vector2d eps=new_posB-new_posA;
                            eps.normalize();
                            eps*=0.1;//10cm
                            //diff_obj+=eps;
                        }
                    }
                    //cout<<"moving HO of "<<diff_obj.norm() << " to "<<(diff_obj.dot(diff_conf)>0 ? frontier->agentB->robot()->getName() : frontier->agentA->robot()->getName())<<
                        //" away of "<< (diff_obj.dot(diff_conf)>0 ? frontier->agentA->robot()->getName() : frontier->agentB->robot()->getName()) << endl;

                    for(unsigned int i_c=0;i_c<2;i_c++){
                        (*confA)[6+i_c]=new_posA[i_c]+diff_obj[i_c];
                        (*confB)[6+i_c]=new_posB[i_c]+diff_obj[i_c];
                    }

                    // double angle_diff_conf=atan2(new_posB.y()-new_posA.y(),new_posB.x()-new_posA.x()) - atan2(x.y(),x.x());
                    // double thRelConfA=new_thA-angle_diff_conf;
                    // double thRelConfB=new_thB-M_PI-angle_diff_conf;

                    //if(abs(thRelConfA - thRelFileA) > 0.00001 || abs(thRelConfB - thRelFileB) > 0.00001){
                        //cout << "##########################  wrong relative angles"<<endl;
                    //}

                    frontier->agentA->robot()->setAndUpdate(*confA);
                    frontier->agentB->robot()->setAndUpdate(*confB);

                    //g3d_draw_allwin_active();
                    //usleep(50);
                    //pqp_set_collision_message(1);
                    pqp_activate_robot_collision(frontier->agentA->robot()->getRobotStruct());
                    pqp_deactivate_robot_selfcollision(frontier->agentA->robot()->getRobotStruct());
                    pqp_activate_robot_environment_collision(frontier->agentA->robot()->getRobotStruct());
                    pqp_activate_robot_collision(frontier->agentB->robot()->getRobotStruct());
                    pqp_deactivate_robot_selfcollision(frontier->agentB->robot()->getRobotStruct());
                    pqp_activate_robot_environment_collision(frontier->agentB->robot()->getRobotStruct());
                    pqp_deactivate_robot_robot_collision(frontier->agentA->robot()->getRobotStruct(),frontier->agentB->robot()->getRobotStruct());
                    if( frontier->agentA->robot()->isInCollisionWithOthersAndEnv()
                            || frontier->agentB->robot()->isInCollisionWithOthersAndEnv()
                            //|| !frontier->agentA->agentInGrid->isHuman() && frontier->agentA->robot()->getCurrentPos()->isInCollision()
                            //|| !frontier->agentB->agentInGrid->isHuman() && frontier->agentB->robot()->getCurrentPos()->isInCollision()
                            || pqp_robot_environment_collision_test(frontier->agentA->robot()->getRobotStruct())==1
                            || pqp_robot_environment_collision_test(frontier->agentB->robot()->getRobotStruct())==1){
                    //if(p3d_col_test_robot(frontier->agentA->robot()->getRobotStruct(),0)
                            //|| p3d_col_test_robot(frontier->agentB->robot()->getRobotStruct(),0)){
                        found=false;
                    }else{
                        found=true;
                        frontier->confA=confPtr_t(confA);
                        frontier->confB=confPtr_t(confB);
                    }
                    pqp_activate_robot_robot_collision(frontier->agentA->robot()->getRobotStruct(),frontier->agentB->robot()->getRobotStruct());
                    //pqp_set_collision_message(0);
                }

                i++;
            }
            ok=found;
        }
    }
    if(ok && (frontier->agentA->robot()->isInCollisionWithOthersAndEnv()
            || frontier->agentB->robot()->isInCollisionWithOthersAndEnv()
            || pqp_robot_environment_collision_test(frontier->agentA->robot()->getRobotStruct())
            || pqp_robot_environment_collision_test(frontier->agentB->robot()->getRobotStruct()))){
        cout << "ERROR, HandOverTestPreDefined says ok while there is a collision"<<endl;
    }
    frontier->check_status=max(frontier->check_status,this->getCheckStatusLevel());
    frontier->status_ok[this->getCheckStatusLevel()]=ok;

    return ok;
}

const std::string HandOverTestVisible::_name=std::string("visible");

HandOverTestVisible::HandOverTestVisible(std::vector<MultiHandOverAgent> *agents,Robot *object):
    _agents(agents), _object_pt(object)
{ }

bool HandOverTestVisible::check(FrontierPtr frontier){
    bool ok=true,result=true;

    //check status level not already met
    if(!frontier->isSymetric() && frontier->select_reverse){
        //not symetric and reverse
        if ( frontier->r_check_status>=this->getCheckStatusLevel()){
            //level already reach
            ok=false;
            //result is the result found and stored in the frontier attribute
            result = frontier->r_status_ok[this->getCheckStatusLevel()];
        }
    }else{
        //symetric, or not reverse
        if( frontier->check_status>=this->getCheckStatusLevel() ){
            ok=false;
            result= frontier->status_ok[this->getCheckStatusLevel()];
        }
    }

    if (ok){
        //Check required status level is met
        if (!frontier->isSymetric() && frontier->select_reverse){
            ok=(frontier->r_check_status >=this->getRequiredCheckStatus());
            result=false;
        }else{
            ok=(frontier->check_status >= this->getRequiredCheckStatus());
            result=false;
        }
    }

    if (ok){
        //check the object has a viman tag
        if( !_object_pt || !_object_pt->getObjectRob() ||
                !_object_pt->getObjectRob()->hasTags()){
            ok=false; // no viman tag is set
            result=false;
        }
    }

    // actual stuff starts here:
    if(ok){
        bool found=false;

        if (frontier->isSymetric()){
            frontier->deSymetrize();
        }

        confPtr_t obj_position;
        confPtr_t confG,confR;  // G=giver, R=receiver
        Robot *agentG,*agentR;

        if(frontier->select_reverse){
            obj_position=frontier->r_object_position;
            confR=frontier->r_confA;
            confG=frontier->r_confB;
            agentG=frontier->agentB->robot();
            agentR=frontier->agentA->robot();
        }else{
            obj_position=frontier->object_position;
            confG=frontier->confA;
            confR=frontier->confB;
            agentR=frontier->agentB->robot();
            agentG=frontier->agentA->robot();
        }

        //1. get object position

        Configuration q_obj_save = *obj_position;
        Configuration q = *obj_position;
        _object_pt->setAndUpdate(q);

        //2. orient it correctly
        //  a. try several options
        vector<Eigen::Affine3d> const & tfs = _object_pt->getObjectRob()->getTags();
        unsigned int tag_i=0;
        while (!found && tag_i < tfs.size()){
            Eigen::Affine3d const &t = tfs[tag_i];
            agentR->setAndUpdate(*confR);
            const p3d_matrix4 * campos = & agentR->getHriAgent()->perspective->camjoint->abs_pos;
            cout<< agentR->getName()<<endl;
            const p3d_matrix4 * objpos = _object_pt->getJoint(0)->getAbsPos();
            Eigen::Vector3d obj_v=m3dGeometry::p3dToEigen(m3dGeometry::getTranslation(*objpos));
            assert(obj_v == _object_pt->getJointPos(1)); // 1 and not 0, see getJointPos() doxygen

            //allign the tag with the cam.
            Eigen::Vector3d obj_eye_dir, // vector ( obj -> eye )
                    tag_n; // current direction (normal) of the tag

            obj_eye_dir = m3dGeometry::p3dToEigen(m3dGeometry::getTranslation(*campos)) - obj_v;
            // obj_eye_dir.normalize();

            // cout<<"obj:\n"<<obj_v<<endl;
            // cout<<"vect:\n"<<obj_eye_dir<<endl;
            tag_n = VimanTags::getTagNormal(t); // only the tag norm is interesting for us
            //tag_n = Eigen::Vector3d::UnitY();
            //obj_eye_dir = -Eigen::Vector3d::UnitZ() + Eigen::Vector3d::UnitX()+Eigen::Vector3d::UnitY();

            // //compute the rotation btw normal and desired
            // Eigen::Vector3d axis = tag_n.cross(obj_eye_dir);

            // assert(abs(tag_n.norm() - 1) < 0.000001 && abs(obj_eye_dir - 1) < 0.000001); //needed for assumption dot product = cos and cross pdt = sin
            // double sin_angle=axis.norm();
            // double cos_angle=(tag_n.dot(obj_eye_dir));
            // double angle=std::acos(cos_angle); //in [0,pi]
            // if(sin_angle <0){ // angle is in [-pi,0]
            //     angle = -angle;
            // }
            // //angle in [-pi,pi]

            Eigen::Quaterniond rot;
            rot.setFromTwoVectors(tag_n,obj_eye_dir);

            cout << "rot*tag_n=\n"<< rot * tag_n <<endl;

            m3dGeometry::setBaseOrientation(_object_pt,rot.matrix());

            //obj_position = confPtr_t(new Configuration(*_object_pt->getCurrentPos()));
            // set the object position (inside the frontier)
            *obj_position = Configuration(*_object_pt->getCurrentPos());


            // Robot *mark=global_Project->getActiveScene()->getRobotByName("MARK1");
            // Robot *mark2=global_Project->getActiveScene()->getRobotByName("MARK2");
            // Configuration q_mark=*mark->getCurrentPos();
            // Configuration q_mark2=*mark2->getCurrentPos();
            // for(unsigned int i=0;i<3;i++){
            //     q_mark[6+i]=frontier->object_position->at(6+i) + obj_eye_dir[i]/obj_eye_dir.norm()*0.5;
            //     q_mark2[6+i]=(*campos)[i][3];
            // }
            // mark->setAndUpdate(q_mark);
            // mark2->setAndUpdate(q_mark2);

            // frontier->agentB->robot()->setAndUpdate(*frontier->confB);
            // cout<<"shown"<<endl;
            // for(unsigned int i=0;i<100;i++){
            //     g3d_draw_allwin_active();
            //     usleep(100e3);
            // }


        //3. test grasp on the object


            //place agents at nav conf + ho 2dpos before
            agentR->setAndUpdate(*agentR->getInitialPosition());
            agentG->setAndUpdate(*agentG->getInitialPosition());

            m3dGeometry::setBaseFromConf(agentR,*confR);
            m3dGeometry::setBaseFromConf(agentG,*confG);

            cout << "computing the pick G"<<endl;
            if(agentG->getHriAgent()->type == HRI_PR2){
                TMI->setWorldState(new WorldState(global_Project->getActiveScene()));
                TMI->getWorldState()->saveAll();
                TMI->setTypeStr("pick");
                TMI->setType(pick);
                TMI->setObject("mainObject", this->_object_pt->getName());
                TMI->setAgent("mainAgent", agentG->getName());
                //TMI->addCurrentTaskToList();
                TMI->setComputeMP(false);
                cout << "call GTP"<<endl;
                TMI->callGTP();
                cout << "call end"<<endl;
                assert(TMI->getCurrentPN());
                found=TMI->getCurrentPN()->hasSucceeded();
                p3d_destroy_all_grasps();
                if(found){
                    assert(TMI->getCurrentPN()->getTaskSolution());
                    //set the giver configuration (inside the forntier struct)
                    //*confG = Configuration(*TMI->getCurrentPN()->getTaskSolution()->getStopWS()->getRobotConf(agentG));
                    *confG = Configuration(*TMI->getCurrentPN()->getTaskSolution()->getSubSolutionWithId(2)->getEndState()->getRobotConf(agentG));

                }
            }else{
                cout<< "pick G :"<<agentG->getName() <<" not a PR2, cowardly giving up (but lets say its ok, right? ;) )"<<endl;
                found=true;
            }

            cout << "done: "<<found<<endl;

            if(found){
                cout << "computing the pick R"<<endl;
                if(agentR->getHriAgent()->type == HRI_PR2){
                    //for the receiver
                    TMI->setWorldState(new WorldState(global_Project->getActiveScene()));
                    TMI->getWorldState()->saveAll();
                    TMI->setTypeStr("pick");
                    TMI->setType(pick);
                    TMI->setObject("mainObject", this->_object_pt->getName());
                    TMI->setAgent("mainAgent", agentR->getName());
                    TMI->addCurrentTaskToList();
                    TMI->setComputeMP(false);
                    TMI->callGTP();
                    assert(TMI->getCurrentPN());
                    found=TMI->getCurrentPN()->hasSucceeded();
                    p3d_destroy_all_grasps();
                    if(found){
                        //set the giver configuration (inside the frontier struct)
                        assert(TMI->getCurrentPN()->getTaskSolution());
                        //*confR = Configuration(*TMI->getCurrentPN()->getTaskSolution()->getStopWS()->getRobotConf(agentR));
                        *confR = Configuration(*TMI->getCurrentPN()->getTaskSolution()->getSubSolutionWithId(2)->getEndState()->getRobotConf(agentR));
                    }
                }else{
                    cout<< "pick R :"<<agentR->getName() <<" not a PR2, cowardly giving up (but lets say its ok, right? ;) )"<<endl;
                    found=true;
                }
                cout << "done: "<<found<<endl;
            }




            tag_i++;
        }


        //4. on failure, try some random positions in (reachable(A) inter reachable(B) )
        _object_pt->setAndUpdate(q_obj_save);
        result=found;
    }

    frontier->check_status=max(frontier->check_status,this->getCheckStatusLevel());
    frontier->status_ok[this->getCheckStatusLevel()]=result;
    cout << "end visual test"<<endl;
    return result;
}


const std::string HandOverTestSelection::_name=std::string("selection");

HandOverTestSelection::HandOverTestSelection(std::vector<MultiHandOverAgent> *agents, Robot *object, string file_directory, double grid_discrete_step):
    _agents(agents),_object_pt(object) , _test_pre(new HandOverTestPreDefined(_agents,file_directory,grid_discrete_step)),_test_vis(new HandOverTestVisible(_agents,object)),
    _is_owner(0)
{ }

HandOverTestSelection::HandOverTestSelection(std::vector<MultiHandOverAgent> *agents, Robot *object, HandOverTestPreDefined *predef, HandOverTestVisible *visib, bool own_tester):
    _agents(agents),_object_pt(object) , _test_pre(predef),_test_vis(visib),
    _is_owner(own_tester)
{}


HandOverTestSelection::~HandOverTestSelection(){
    if(_is_owner){
        delete _test_pre;
        delete _test_vis;
    }
}

bool HandOverTestSelection::check(FrontierPtr frontier){
    bool use_vis=false;
    if(!_agents->at(frontier->agentA_ind).is_human && !_agents->at(frontier->agentB_ind).is_human){
        //both robots
        use_vis = true;
    }
    if( use_vis
            && ! ( _object_pt->getObjectRob() && _object_pt->getObjectRob()->hasTags() ) ){
        use_vis=false;
    }

    if(use_vis){
        return _test_vis->check(frontier);
    }else{
        return _test_pre->check(frontier);
    }
}

const std::string HandOverTestTrajectory::_name=std::string("trajectories");

HandOverTestTrajectory::HandOverTestTrajectory(std::vector<MultiHandOverAgent> *agents):
    _agents(agents)
{
}

bool HandOverTestTrajectory::check(FrontierPtr frontier){
    bool ok=true;

    //TODO simple check againts polymorphism issues
    assert(this->getRequiredCheckStatus() == FrontierCells::FCS_HANDOVER);
    assert(this->getCheckStatusLevel() == FrontierCells::FCS_TRAJECTORIES);
    if(frontier->check_status>=this->getCheckStatusLevel()){
        if(!frontier->status_ok[this->getCheckStatusLevel()]){
            ok=false;
        }else{
            //TODO costs
            ok=true;
        }
    }else if (frontier->check_status < this->getRequiredCheckStatus()){
        ok=false;
    }else{
        MultiHandOverAgent* a1(0),*a2(0);
        for(unsigned int i=0; (!a1 || !a2) && i<_agents->size();i++){
            if(_agents->at(i).robot==frontier->agentA->robot()){
                a1=&_agents->at(i);
            }else if(_agents->at(i).robot==frontier->agentB->robot()){
                a2=&_agents->at(i);
            }
        }
        if(!a1 || !a2){
            //critical error
            cerr<<"agents not found in HandOverTestTrajectory"<<endl;
            ok=false;
        }
        if(ok){
            confPtr_t initconf;
            //init conf is the robot initial conf moved at the x y theta of the handover position
            initconf = confPtr_t(new Configuration(*a1->robot->getInitialPosition()));
            for (unsigned int i = 0; i < 6; i++){
                initconf->at(6+i) = frontier->confA->at(6+i);
            }
            API::Trajectory *traj(0);
            ok=this->computeTraj(a1->robot,initconf,frontier->confA,traj);
            if (ok){
                assert(traj);
                frontier->trajA=traj;
            }
        }
        if(ok){
            //TODO put agent A at its HO conf?
            confPtr_t initconf;
            //init conf is the robot initial conf moved at the x y theta of the handover position
            initconf = confPtr_t(new Configuration(*a2->robot->getInitialPosition()));
            for (unsigned int i = 0; i < 6; i++){
                initconf->at(6+i) = frontier->confA->at(6+i);
            }
            API::Trajectory *traj(0);
            ok=this->computeTraj(a2->robot,initconf,frontier->confB,traj);
            if (ok){
                assert(traj);
                frontier->trajB=traj;
            }

        }

    }
    frontier->check_status=max(frontier->check_status,this->getCheckStatusLevel());
    frontier->status_ok[this->getCheckStatusLevel()]=ok;
    return ok;
}

bool HandOverTestTrajectory::computeTraj(Robot* r,confPtr_t start, confPtr_t target, API::Trajectory *traj){
    bool ok=true;
    MultiHandOver3dPlanner mp(r,r->getHriAgent()->is_human);
    vector<Eigen::Vector2d> path;

    assert(r == start->getRobot() && r == target->getRobot());

    path.push_back(Eigen::Vector2d(start->at(6+0),start->at(6+1)));
    path.push_back(Eigen::Vector2d(start->at(6+0),start->at(6+1)));

    ok = mp.computeHandOverTrajectory(start,target);

    traj = new API::Trajectory(r->getCurrentTraj());

    return ok;
}

}
}
